(ns pdengage.core
  (:require
    [clojure.string :as string]

    [om.core :as om :include-macros true]
    [om.dom :as dom :include-macros true]
    ;; Bootstrap
    [om-bootstrap.button :as b]
    [om-bootstrap.grid :as g]
    [om-bootstrap.nav :as n]
    [om-bootstrap.table :refer [table]]
    [om-bootstrap.input :as b-input]

    [om-tools.core :refer-macros [defcomponent]]
    [om-tools.dom :as odom :include-macros true]

    [cljs-time.core :as time-core]
    [cljs-time.coerce :as time-coerce]
    [cljs-time.format :as time-format]

    [goog.string :as gstring]
    [goog.string.StringBuffer]
    [goog.date :as gdate]
    [goog.date.DateTime :as gdatetime]

    ;; Schema Support
    [schema.core :as sce :include-macros true]

    [goog.i18n.DateTimeFormat :as dtformat]
    [goog.i18n.DateTimeParse :as dtparse]
    [goog.Uri]
    [goog.date]
    
    ;; Ajax
    [ajax.core :refer [GET POST]])
  (:require-macros [pdengage.macros :refer [defstat]])
  (:import goog.History))

;; without this, we can get crashes when using println
(enable-console-print!)


;; http://www.keenthemes.com/preview/metronic/theme/templates/admin4/index.html
;; https://groups.google.com/forum/#!topic/clojurescript/Bd4MthunzkQ
;; http://ironsummitmedia.github.io/startbootstrap-sb-admin/

;;--------------------------------------------------
;; ObjC/JS Interop
;;--------------------------------------------------

(defn- loc [term]
  (if js/proxy_object
    (if-let [t (.localizationFor_ js/proxy_object term)]
      t
      term)
    term))

(def translation-table
  {
    :tagged_users (loc "Tagged Users")
    :low_resolution (loc "Image")
    :created (loc "Created")
    :image (loc "Image")
    :name (loc "Name")
    :count (loc "Count")
    :filter (loc "Filter")
    :likes (loc "Likes")
    :comments (loc "Comments")
    :score (loc "Score")
    :video (loc "Video")
    :type (loc "Image-Type")
    :image_count (loc "Posts")
    :follow_count (loc "Follows")
    :follower_count (loc "Followers")
    :commenter_name (loc "Username")
    :commenter_bio (loc "Bio")
    :commenter_website (loc "Website")
    :commenter_followers (loc "Followers")
    :commenter_follows (loc "Follows")
    :commenter_media (loc "Image Count")
    :created_time (loc "Created")
    :comment_text (loc "Comment")
    :commenter_image (loc "Profile")
    :liker_name (loc "Username")
    :liker_bio (loc "Bio")
    :liker_website (loc "Website")
    :like_time (loc "Created")
    :like_type (loc "Type")
    :like_count (loc "Likes")
    :comment_count (loc "Comments")
    :full_name (loc "Name")
    :profile_picture (loc "Profile")
    :average_like_count (loc "Average Like Count")
    :average_comment_count (loc "Average Comment Count")
    })

(defn trnl [k]
  "translate from the dabase name of a key to a proper key. return camcel-cased variant, if not in there"
  (if-let [x ((keyword k) translation-table)] x (-> k name clojure.string/capitalize)))

(defn- osxaction-get [call]
  (GET call
    {;:handler #(println "done gauge store", %)
                                        ;:response-format :json ; TODO transit
                                        ;:format :json
                                        ;:keywords? true
      :error-handler (fn [{:keys [status response status-text] :as e}]
                       (if (= 0 status)
                         (println "Error" status status-text)
                         ))})
  )

;; in order to be able to bundle up statistics / json requests during loading
;; we need to define each page's statistic request paths in here.
(def pages
  [{:key :performance :name "Performance" :icon "fa-bar" :route "/engage/performance"
     :stats [:account-image-upload-growth :performance-metrics :relative-comment-growth :relative-follower-growth :upload-follower-change ]
     } 
    {:key :content :name "Content" :icon "fa-like" :route "/engage/content"
      :stats [
               :most-important-images :most-used-my-images :best-performing-images 
               :grouped-likes :grouped-comments :most-active-likers :account-new-likes :image-content-types :most-active-commenters :filter-metrics]
      }
    {:key :audience :name "Audience" :icon "fa-network" :route "/engage/content"
      :stats [:important-commenters :important-followers :mostused-followers :mostused-commenters :filter-metrics :most-important-follower-images :image-content-types]
      }])


(def StatisticsEntry
  {sce/Keyword {(sce/required-key :call) sce/Any ;; Call can be any data, i.e. {:res "likes" :id 5 :cid 7}
                 (sce/required-key :date) sce/Any ; goog.date.UtcDateTime last time we fetched it from the server
                 (sce/required-key :created) sce/Int ; timestamp when the data was created
                 (sce/required-key :type) (sce/enum :daily :overview) ; what kind of statistic this is
                 (sce/required-key :data) sce/Any ; the actual data
                 }}) 

(def
  app-state
  (atom
    {:user {:targets {:daily-like-target {:num 11 :dim :day}
                       :daily-follow-target {:num 12 :dim :day}
                       :weekly-upload-target {:num 13 :dim :day}
                       :daily-comment-target {:num 14 :dim :week}}}
      :current {:page :dashboard
                 :account-id nil}
      :statistics
      {:important-commenters (defstat "important-commenters") 
        :important-followers (defstat "important-followers")
        :mostused-followers (defstat "mostused-followers")
        :mostused-commenters (defstat "mostused-commenters")
        :most-important-images (defstat "most-important-images")
        :most-important-follower-images (defstat "most-important-follower-images")
        :performance-metrics (defstat "performance-metrics")
        :filter-metrics (defstat "filter-metrics")
        :best-performing-images (defstat "best-performing-images")
        :relative-comment-growth (defstat "relative-comment-growth")
        :relative-follower-growth (defstat "relative-follower-growth")
        :upload-follower-change (defstat "upload-follower-change")
        :account-image-upload-growth (defstat "account-image-upload-growth")
        :grouped-likes (defstat "grouped-likes")
        :grouped-comments (defstat "grouped-comments")
        :most-active-likers (defstat "most-active-likers")
        :most-active-commenters (defstat "most-active-commenters")
        :most-used-my-images (defstat "most-used-my-images")
        :account-new-likes (defstat "account-new-likes")
        :image-content-types (defstat "image-content-types")
        }
      }))

;; ---------------------------------------------------------------
;; Server Communication
;; ---------------------------------------------------------------


(defn- flatten-pages [xs]
  (reduce #(if (:children %2)
             (concat %1 (list %2) (flatten-pages (:children %2)))
             (concat %1 (list %2)))
    '() xs))

(defn- map-pages []
  ((memoize (fn []
              (let [c (flatten-pages pages)
                     a (map :key c)]
                (zipmap a c))))))

(defn- calls-for-page [page]
  (let [all-stat-keys (-> (map-pages) page :stats)
         all-stats (select-keys (-> @app-state :statistics) all-stat-keys)
         required-stats
         (filter (fn [[key stat]]
                   (cond
                     (not (:data stat)) true ; Didn't load the data yet
                     (not (:date stat)) true
                     (< 180 (time-core/in-minutes (time-core/interval (:date stat) (time-core/now)))) true ; three hours old
                     :else nil
                     ) 
                   ) all-stats)]
                                        ;(println "required" (map #(-> % last :call) required-stats))
                                        ;(println page "required" required-stats)
                                        ;(into {} #(->> % last (select-keys [:key :call])))
    (into [] (map (fn [x] (-> x last :call)) required-stats))
                                        ;(into {} (map (fn [x] [(first x) (-> x last :call)]) required-stats)
    ))

(defn- process-calls-response
  "Process the response of a analytics call. The response should look like this
  {:keyname data :keyname2 data}
  data is a block like {:hour int, :key 'name again' :humantime 'timestring' :time int :data [...data]}"
  [response]
  ;;(println "type" (type response) (keys response) (type (-> response keys first)))
  ;;(println "got response" response)
  (swap! app-state update-in [:statistics]
    (fn [stats]
      (reduce (fn [s [akey {:keys [hour key time data]}]]
                (-> s
                  (assoc-in [akey :data] data)
                  (assoc-in [akey :created] (time-coerce/from-long time))
                  (assoc-in [akey :date] (time-core/now)))
                ) stats response))))

(defn bundle-calls
  "Determine which data is needed, bundle up the calls, send them to the server,
  and put the results back into the data structure.
  The request is structured as follows
  POST /analytics/uid
  BODY: (json) {:keyname 'call1' :keyname {call2info}}"
  [page]
  (let [calls (calls-for-page page)
         cid (:account-id (:current @app-state))
         path (str "/query-analytics/" cid "/bundle?keys=" (string/join "," calls))]
                                        ;(println @app-state)
                                        ;(.log js/console "send" (clj->js calls))
    (GET path
      {:handler process-calls-response
        :response-format :json ; TODO transit
        :format :json
        :keywords? true
        :error-handler (fn [{:keys [status response status-text] :as e}]
                         (if (= 0 status)
                           (process-calls-response response)
                           (println "Error" status status-text)
                           ))})))

(defn store-gauge-target [target value dimension]
  (GET (str "/gauge-target/" (:account-id (:current @app-state)) "/" (str target) "/" (str value) "/" (str dimension))
    {:handler #(println "done gauge store", %)
      :response-format :json ; TODO transit
      :format :json
      :keywords? true
      :error-handler (fn [{:keys [status response status-text] :as e}]
                       (if (= 0 status)
                         (println "Error" status status-text)
                         ))}))

(defn ^:export proofjs []
  (.log js/console (clj->js @app-state)))

(defn ^:export proof []
  (println @app-state))

(defn request-from-server [action params handler]
  "Requests from the server by creating a URL from the params.
   also parses the response and returns the following result
   to the handler:
   {:time time when this data was rendered (js date)
    :key concrete name of this data
    :hour the hour id when this was created
    :data the actual rendered data}"
  (let [user (:user params)
         account (:account params)
         params (dissoc (dissoc params :account) :user )
         param-map (string/join "&" (map (fn [x] (str (name (first x)) "=" (last x))) (into [] params)))
         con (if (> (count params) 0) "?" "")
         url (str "/analytics/" user "/" account "/" action con param-map)]
    ;(println "calling" url)
    (GET url {:keywords? true
               :handler (fn [result]
                          ;; transform into the proper struct
                          (handler {:time (time-coerce/from-long (:time result))
                                     :key (:key result)
                                     :hour (:hour result)
                                     :data (:data result)}))
               :response-format :transit})))



(defn osxaction [type action]
  "perform an action in photodesk. valid types: :tag :username :photo :user :dl-csv :dl-img"
  ;(println type action)
  (let [ ep "analytics-action"
         ac (:account-id (:current @app-state))]
    (when (and ac type action)
      (GET (str "/" ep "/" ac "/" (name type) "/" action)))))


(defn ^:export setaccount [ac]
                                        ;(println app-state)
  (swap! app-state (fn [s] (assoc-in s [:current :account-id] ac)))
                                        ;(println app-state)
  )

(defn ^:export setroute [route]
  "set the route to be selected"
  (swap! app-state (fn [s] (assoc-in s [:current :page] (keyword route))))
  (bundle-calls (keyword route)))

(defn ^:export settarget [target dimension value]
  "set the value of a target"
                                        ;(println app-state)
  (swap! app-state (fn [s] (assoc-in s [:user :targets (keyword target)] {:num (int value) :dim (keyword dimension)})))
                                        ;(println app-state)
  )

;; ---------------------------------------------------------------
;; Helpers
;; ---------------------------------------------------------------

(defn- stat-time
  "Parse the 'time created' information from the statistics and return
   a string that shows at which datetime this statistics was created"
  [cursor key]
  (let [c (-> cursor key :created)
         f (goog.i18n.DateTimeFormat. "dd/MM hh:mm" )]
    (if c
      (.format f c)
      ""
      )))

(defn- parse-url-hostname [url]
  "Parse an URL an return the hostname"
  (let  [parser (.createElement js/document "a")]
    (aset parser "href" url)
    (aget parser "hostname")))

(defn- js-export-csv [data]
  (let [header (->> data first keys (map name))
         values (->> data rest (map vals))
         csv-content "data:text/csv;charset=utf-8,"
         lines (map (fn [row] (clojure.string/join "," (map #(clojure.string/replace (str %) "," ".") row))) values)
         content (str csv-content (clojure.string/join "," header) "\n"
                   (clojure.string/join "\n" lines) "\n")
         output (.encodeURI js/window content)]
    (osxaction :dl-csv output)))

(defn- js-export-pdf [element format]
  (let [image (.toBase64Image element)]
    (osxaction :dl-img image))) 

;; https://github.com/davesann/cljs-uuid/blob/master/src/cljs_uuid/core.cljs
(defn make-random
  "Returns a new randomly generated (version 4) cljs.core/UUID,
  like: xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx
  as per http://www.ietf.org/rfc/rfc4122.txt.
  Usage:
  (make-random)  =>  #uuid \"305e764d-b451-47ae-a90d-5db782ac1f2e\"
  (type (make-random)) => cljs.core/UUID"
  []
  (letfn [(f [] (.toString (rand-int 16) 16))
           (g [] (.toString  (bit-or 0x8 (bit-and 0x3 (rand-int 15))) 16))]
    (UUID. (.toString
             (goog.string.StringBuffer.
               (f) (f) (f) (f) (f) (f) (f) (f) "-" (f) (f) (f) (f) 
               "-4" (f) (f) (f) "-" (g) (f) (f) (f) "-"
               (f) (f) (f) (f) (f) (f) (f) (f) (f) (f) (f) (f))))))


(def wrapper-class "slimScrollDiv")
(defn- enable-slim-scroll [node height]
  "Enable a slim scroll for the given element."
  (let [n (js/jQuery node)]
    (when (= "1" (.attr node "data-initialized"))
      ;; remove the old slim scroll
      (.removeAttr node "data-initialized")
      (.removeAttr node "style")
      (.slimScroll node #js {:wrapperClass wrapper-class :destroy true}))
    (.slimScroll node #js {:wrapperClass wrapper-class :height height})
    (.attr node "data-initialized" "1")))

(defcomponent textbox [text owner opts]
  (init-state [_]
    {:content text})
  (render [_]
    (let [finish-action (:finish-action opts)
           button-title (:button-title opts)
           button-placeholder (:placeholder opts)]
      (dom/div #js {:className "input-group input-group-sm"}
        (dom/input #js {:type "text" :className "form-control" :placeholder button-placeholder
                         :value (om/get-state owner :content)
                         :onChange #(om/set-state! owner :content (.. % -target -value))})
        (dom/span #js {:className "input-group-btn"}
          (dom/button #js {:onClick (fn [e]
                                      (finish-action (om/get-state owner :content))
                                      ) :className "btn btn-default" :type "button"}
            button-title))
        ))))

(defcomponent selectbox [{:keys [ranges active]} owner opts]
  (init-state [_]
    {:active active})
  (render-state
    [_ state]
    (apply dom/div #js {:className "btn-group btn-group-sm" :role "group"}
      (for [{:keys [title key]} ranges]
        (dom/div #js {:type "button"
                       :onClick (fn [e]
                                  (om/set-state! owner :active key)
                                  ((:click-fn opts) key)
                                  )
                       :className (str "btn btn-default"
                                    (when (= (:active state) key) " active"))} title)))))


;; ---------------------------------------------------------------
;; Font Awesome Icon Composition
;; ---------------------------------------------------------------

(defn fawesome-classes [icon & items]
  "Return a Bootstrap icon class string. The keys map below
   displays all possible options"
  (let [keys [:lg ; 33% size increase
               :2x :3x :4x :5x
               :fixed-width
               :bullet ; bullet in list. parent <ul> needs "fa-ul"
               :border
               :pull-right :pull-left ; for easy pull quotes or article icons
               :spin ; rotate
               :pulse ; rotate with 8 steps
               :90 :180 :270 ; rotate n degrees
               :horizontal :vertical ; flip in the appropriate ways
               ;; stacking needs the parent element (i.e. <span>) to have "fa-stack"
               :stack-1x ; regular icon size stacking
               :stack-2x
               :inverse]
         defs (map vector keys (repeat false))
         user (map vector items (repeat true))
         sets (merge (into {} defs) (into {} user))]
    (->> (for [setting sets
                :when (last setting)]
           (case (first setting)
             :fixed-width "fa-fw"
             :bullet "fa-li"
             :pull-left "pull-left" :pull-right "pull-right"
             :90 "fa-rotate-90" :180 "fa-rotate-180" :270 "fa-rotate-270"
             :horizontal "fa-flip-horizontal"
             :vertical "fa-flip-vertical"
             (str "fa-" (name (first setting)))))
      (cons icon)
      (cons "fa")
      (clojure.string/join " "))))

(defn fawesome-icon [icon text params & items]
  "Generate a font awesome i tag"
  (dom/i (clj->js (merge params {:className (str (:className params) " "
                                              (apply fawesome-classes (cons icon items)))})) (if text text " ")))

(defn fawesome-spinner [text params class]
  "Generate a font awesome spinner"
  (dom/i (clj->js (merge params {:className (fawesome-classes "fa-spinner" :spin)})) (if text text " ")))

(defn fawesome-pulser [text params class]
  "Generate a font awesome pulser"
  (dom/i (clj->js (merge params {:className (fawesome-classes "fa-spinner" :pulse)})) (if text text " ")))

(defn fawesome-icon-simple [icon]
  "Generate just the icon"
  (dom/i (clj->js {:className (fawesome-classes icon)})))



;; ---------------------------------------------------------------
;; Helper Components
;; ---------------------------------------------------------------

"Define a sparkline component. {:opts ...} defines the settings:
    :type -> :line :bar :tristate :discrete :bullet :pie :box
    :width -> ..
    :height -> ..
    rest is documented here:
    http://omnipotent.net/jquery.sparkline/#bar "
(defcomponent sparkline-component [data owner opts]
  (init-state
    [_]
    {:update-fn
      (fn [s]
        (let [n (om.core/get-node owner)
               d {:disableTooltips true :disableInteraction :true :disableHighlight :true :barColor "#7799FF" :lineColor "#7799FF"}
               m (merge d opts)]
          (.sparkline
            (js/jQuery n) (clj->js s) (clj->js m))))})

  ;; install the sparkline into the dom
  (did-mount [_] ((om/get-state owner [:update-fn]) data))

  ;; update the component
  (will-update [_ next-props next-state] 
    ((om/get-state owner [:update-fn]) next-props))

  (render [_]
    (dom/span nil "c")))

(defcomponent testtest [data owner opts]
  (init-state
    [_]
    {:update-fn
      (fn [s]
        (let [n (om.core/get-node owner)
               d {:disableTooltips true :disableInteraction :true :disableHighlight :true :barColor "#7799FF" :lineColor "#7799FF"}
               m (merge d opts)]
          (.sparkline
            (js/jQuery n) (clj->js s) (clj->js m))))})

  ;; install the sparkline into the dom
  (did-mount [_] ((om/get-state owner [:update-fn]) data))

  ;; update the component
  (will-update [_ next-props next-state] 
    ((om/get-state owner [:update-fn]) next-props))

  (render [_]
    (dom/span nil "c")))


(defcomponent timerange-buttongroup [{:keys [ranges active]} owner opts]
  (render
    [_]
    (apply dom/div #js {:className "btn-group btn-group-sm" :role "group"}
      (for [{:keys [title key]} ranges]
        (dom/div #js {:type "button"
                       :onClick (partial (:click-fn opts) key)
                       :className (str "btn btn-default"
                                    (when (= active key) " active"))} title)))))



;; ---------------------------------------------------------------
;; Gauges (Mini Data Capsules)
;; ---------------------------------------------------------------

(def GaugeColumnOptions
  {(sce/required-key :title) sce/Str
    (sce/optional-key :transformer) sce/Any ; format / transform the value
    })

(def GaugeOptions
  {(sce/required-key :columns) {sce/Keyword GaugeColumnOptions}
    (sce/required-key :icon) sce/Str
    (sce/required-key :label) sce/Str
    (sce/required-key :target) {(sce/required-key :value) {(sce/required-key :num) sce/Int (sce/required-key :dim) sce/Str}
                                 (sce/required-key :field) sce/Keyword}
    (sce/required-key :color) sce/Str}) 

(defcomponent gauge-column [data owner opts]
  (render [_]
    (let [color (:color opts)
           trans? (:transformer opts)]

      (odom/div {:class "col-md-6 col-sm-6"}
        (odom/div {:class (str "font-sm font-grey-mint uppercase")} (odom/small nil (:title opts)))
        (odom/div {:class (str "font-md font-" color)}
          (odom/small nil (if (and data trans?) (trans? data) data)))))))

(defn- gauge-data [state stat-key]
  (-> state :statistics stat-key :data))

(defn- gauge-target [state stat-key value-key target-key]
  (let [target (-> state :user :targets target-key)
         performance (-> (gauge-data state stat-key) value-key)]
    (if (and target performance (> target 0) (> performance 0))
      (/ performance target)
      0)))

(defcomponent gauge [data owner opts]
  (init-state [_]
    (let [{target-num :num target-dim :dim} (-> opts :target :value)]
      {:target-edit false
        :target-dim target-dim
        :target-num target-num
        }
      )
    )
  (render-state
    [_ state]
    (let [color (:color opts)
           {target-num :num target-dim :dim} (-> opts :target :value)
           target-field (-> opts :target :field)
           target-title (-> opts :target :title)
           target-val (if (and data target-field) (-> data target-field) 0)
           target-calc (if (and target-val target-num) (* 100 (/ target-val target-num)) 0)
           target-action (fn []
                           ((:target-action opts) (om/get-state owner :target-num) (om/get-state owner :target-dim)))
           all-cols (keys (:columns opts))
           all-vals (map #(% data) all-cols)
           change-target #(om/set-state! owner :target-edit (not (:target-edit state)))
           ;; change the order if the first is 0, so that there's always an attractive value up
           [col & cols] (if (< (first all-vals) 1) (cons (second all-cols) (cons (first all-cols) (drop 2 all-cols))) all-cols)
           {:keys [type transformer title]} (if (and data col) (col data) {})
           [val & vals] (if (and data col cols) (map #(% data) (cons col cols)) [])
           ]
      (odom/div
        {:class "dashboard-stat2"}
        (odom/span {:class "caption-subject bold uppercase"} (:label opts))
        (odom/div {:class "display"}
          (odom/div {:class "number"} 
            (odom/h3 {:class (str "font-" color)} (if (and val transformer) (transformer val) val))
            (odom/small {:class "uppercase"} title))
          (odom/div {:class "icon"}
            (fawesome-icon-simple (:icon opts))))
        (if (:target-edit state)
          (odom/div {:class "row targets gauge-edit-box"}
            (odom/div {:class "col-md-12 col-sm-12"}
              (om/build textbox target-num {:opts {:finish-action
                                                    (fn [new-target]
                                                      (om/set-state! owner :target-num new-target)
                                                      (target-action)
                                                      (change-target))
                                                    :button-title (loc "Set Target")
                                                    :placeholder "0"}})
              )

            (om/build selectbox {:ranges [{:title (loc "Day") :key :day}
                                           {:title (loc "Week") :key :week}
                                           {:title (loc "Month") :key :month}] :active target-dim}
              {:opts {:click-fn (fn [new-dim] (om/set-state! owner :target-dim new-dim)
                                  (target-action))}}))

          (if target-val
;;            (comment odom/div {:class "row"}
;;              (odom/div {:class "col-md-12 col-sm-12 progress-info"}
;;                (odom/div {:class "display"}
;;                  (odom/div {:class "number"} 
;;                    (odom/h3 {:class (str "font-" color)} target-val)
;;                    (odom/small {:class "uppercase"} target-title))
;;                  )))
            (odom/div {:class "row"})
            (odom/div {:class "row"}))

;;          (comment odom/div {:class "row targets"}
;;            (odom/div {:class "col-md-12 col-sm-12 progress-info"}
;;              (odom/div {:class "progress"}
;;                (odom/span {:style {:width (str target-calc "%")} :class (str "progress-bar progress-bar-success " color)}
          ;;                  (odom/span {:class "sr-only"} (str (loc "of target ") target-dim))))
;;              (odom/div {:class "status"}
;;                (odom/div {:class "status-title"} (goog.string/format "%d%%" target-calc) ) 
          ;;                (odom/div {:on-click change-target :class "status-button"} (loc "SET TARGET"))
          ;;                (odom/div {:class "status-number"} (str (loc "target ") target-dim))
;;                )))
          )

        (odom/div {:class "row"}
          (for [c cols]
            (om/build gauge-column (c data) {:opts (merge {:color color} (c (:columns opts)))})))
        ))))

;; Basic Settings for a chart
(def ChartOpts
  "The possible chart options"
  {(sce/required-key :conf) sce/Any ; the json that defines the amchart config
    (sce/required-key :format) {sce/Keyword ; the column name. only columns listed here will be displayed
                                 {(sce/optional-key :transformer) sce/Any}}  ; for each column, a transformer to transform the value
    (sce/optional-key :time-ranges) [{:title sce/Str :key sce/Keyword}] ; the different time ranges, if there are
    (sce/optional-key :settings) {(sce/required-key :height) sce/Int
                                   (sce/optional-key :class) sce/Str
                                   (sce/required-key :title) sce/Str
                                   }})

(defn- chartjs-pie-conf [titlefield valuefield conf]
  {:ylabelkey titlefield :yvaluekey valuefield})

(defcomponent chart-view [data owner opts] ;;  :- ChartOpts
  (init-state [_]
    {:current-range nil
      :chart-js nil ; the js var that is the chart. will be set via a fn. very dirty. very dirty.
      })
  (render-state
    [_ state]   
    (let [
           format (:format opts)
           time (:time opts)
           subtitle (:subtitle opts)
           ;; Apply the current range, if one exists
           time-ranges (:time-ranges opts)
           current-range (if-let [r (om/get-state owner :current-range)] r (-> time-ranges first :key))
           rows (if current-range (current-range data) data)
           rows (if-let [t (:row-transformer format)] (t rows) rows) 
           ;; Limit to the valid columns
           valid-columns (-> opts :format keys)
           rows (if (and valid-columns (count valid-columns)) (map #(select-keys % valid-columns) rows) rows)
           ;; Apply value transformer to rows
           rows (map (fn [r] (into {} (map (fn [[k v]] (if-let [t (-> format k :transformer)] [k (t v)] [k v]) ) r)) ) rows)
           ;; Get display settinsg
           settings (:settings opts)
           height (:height settings)
           class (:class settings)
           title (:title settings)
           conf (:conf opts)]

      (odom/div {:class "portlet light bordered"}
        (odom/div {:class "portlet-title"}
          (odom/div {:class "caption font-green-sharp"} (odom/div {:class "caption-subject bold"} title))
          (when subtitle
            (odom/div {:class "actions"}
              (odom/span {:class "label label-primary"} subtitle))))

        (odom/div {:class "portlet-body"}
          (g/row nil
            (g/col {:xs 12}
              (om/build chart-view-graph rows {:opts {:class class :height height :conf conf
                                                       :chart-js-setter #(om/set-state! owner :chart-js %)}}))))
        (odom/div {:class "portlet-footer"}
          (g/row nil
            (g/col {:xs 10 :md 10}
              (om/build timerange-buttongroup {:ranges time-ranges :active current-range}
                {:opts {:click-fn #(om/set-state! owner :current-range %)}}))
            (g/col
              {:xs 2 :md 2}
              (om/build chartview-options-button {:data rows :chart-js (om/get-state owner :chart-js)})
              )))
        ))))

(defcomponent chartjs-area-graph [data owner opts]
  (init-state [_]
    {:uid (str (make-random))
      :update-fn (fn [s]
                   (let [u (om/get-state owner :uid) 
                          ue (. js/document (getElementById u))
                          gradient (.createLinearGradient (.getContext ue "2d") 0 0 0 (:height opts))
                          _ (.addColorStop gradient 0 "rgba(132, 184, 229, 0.65)")
                          _ (.addColorStop gradient 1 "rgba(79, 145, 202, 0.15)")
                          conf (:conf opts)
                          ;; conf settings
                          ylabelkey (:ylabelkey conf)
                          titlename (:title conf)
                          ytitle (:ytitle conf)
                          yvaluekey (:yvaluekey conf)
                          dateformat (:dateFormat conf)
                          tooltip (:tooltiptitle conf)
                          ;; 
                          type "line"
                          labels (map ylabelkey s)
                          dataset [{:label tooltip
                                     :backgroundColor gradient  
                                     :pointStyle "circle"
                                     :lineTension 0.00001 ; 0 for straight lines 100 for soft lines (or 50 or so)
                                     :tension 0.00001 ; 0 for straight lines 100 for soft lines (or 50 or so)
                                     :pointHitRadius 8
                                     :pointRadius 5
                                     :pointBorderColor "rgb(46, 110, 165)"
                                     :pointBackgroundColor "#ffffff"
                                     :pointBorderWidth 1
                                     :borderWidth 1
                                     :borderColor "rgb(70, 140, 200)"
                                    :data (map yvaluekey s)}]
                          options {:responsive true
                                    :title {:display false}
                                    :legend {:display false}
                                    :tooltips {:mode "single"}
                                    :scales
                                    {:yAxes
                                      [{:scaleLabel
                                         {:labelString ytitle :display true}
                                         :ticks {:beginAtZero true}}]
                                      :xAxes [{:display true
                                                :ticks {:autoSkip true}
                                                :type "time"
                                                :time {:parser dateformat
                                                        :round true
                                                        :tooltipFormat "YYYY-MM-DD"
                                                        :displayFormats {"month" "YYYY-MM"}}
                                                ;:ticks {:fontColor "rgb(34, 88, 133)"}
                                                }]
                                      }}
                          opt {:type type :data {:datasets dataset :labels labels} :options options}
                          a (js/Chart. ue (clj->js opt))]
                     ((:chart-js-setter opts) a)
                     ))})

  ;; install the chart into the dom
  (did-mount [_] ((om/get-state owner [:update-fn]) data))

  ;; update the component
  (will-update [_ next-props next-state] 
    ((om/get-state owner [:update-fn]) next-props))

  (render [_]
    (odom/canvas {:id (om/get-state owner :uid) :class (str "chart " (:class opts)) :style {:height (str (:height opts) "px")}})))

(defcomponent chartjs-bar-graph [data owner opts]
  (init-state [_]
    {:uid (str (make-random))
      :update-fn (fn [s]
                   (let [u (om/get-state owner :uid) 
                          ue (. js/document (getElementById u))
                          conf (:conf opts)
                          gradient (.createLinearGradient (.getContext ue "2d") 0 0 0 (:height opts))
                          _ (.addColorStop gradient 0 "rgba(132, 184, 229, 0.65)")
                          _ (.addColorStop gradient 1 "rgba(79, 145, 202, 0.15)")
                          ;; conf settings
                          ylabelkey (:ylabelkey conf)
                          titlename (:title conf)
                          ytitle (:ytitle conf)
                          yvaluekey (:yvaluekey conf)
                          dateformat (:dateFormat conf)
                          tooltip (:tooltiptitle conf)
                          type "bar"
                          labels (map ylabelkey s)
                          dataa (map yvaluekey s)
                          tooltipcallback (:tooltipcallback conf)
                          tooltipcaler (fn [x] (let [row (first (drop (get (js->clj x) "index") s))]
                                                 (tooltipcallback row)))
                          dataset [{:label tooltip
                                     :backgroundColor gradient  
                                     :pointStyle "circle"
                                     :pointRadius 8
                                     :lineTension 0 ; 0 for straight lines 100 for soft lines (or 50 or so)
                                     :pointHitRadius 16
                                     :pointBorderColor "rgb(46, 110, 165)"
                                     :pointBackgroundColor "#ffffff"
                                     :pointBorderWidth 1
                                     :borderWidth 1
                                     :borderColor "rgba(70, 140, 200, 0.8)"
                                    :data dataa}]
                          options {:responsive true
                                    :title {:display false}
                                    :categoryPercentage 1.0
                                    :legend {:display false}
                                    :tooltips {:mode "single"}
                                    ;:animation {:onComplete js/addChartJSLabels} ;
                                    :scales
                                    {:yAxes
                                      [{:scaleLabel {:labelString ytitle :display true}
                                         :ticks {:beginAtZero true :min 0 :suggestedMin 0}
                                         }]
                                      :xAxes []
                                      }}
                          options (if dateformat (assoc-in options [:scales :xAxes 0]
                                                     {
                                                       :display true
                                                       :type "time"
                                                       :time {:parser dateformat
                                                               :round true :tooltipFormat "YYYY-MM-DD"
                                                               :displayFormats {"month" "YYYY-MM"}
                                                               }
                                                       :ticks {:beginAtZero true :min 0 :suggestedMin 0 :autoSkip true}
                                                       })
                                    options)
                     options (if tooltipcallback (assoc options :tooltips {:callbacks {:label tooltipcaler}}) options)

                     opt {:type type :data {:datasets dataset :labels labels} :options options}
                          ue (. js/document (getElementById u))
                          a (js/Chart. ue (clj->js opt))]
                     ((:chart-js-setter opts) a)
                     ))})

  ;; install the chart into the dom
  (did-mount [_] ((om/get-state owner [:update-fn]) data))

  ;; update the component
  (will-update [_ next-props next-state] 
    ((om/get-state owner [:update-fn]) next-props))

  (render [_]
    (odom/canvas {:id (om/get-state owner :uid) :class (str "chart " (:class opts)) :style {:height (str (:height opts) "px")}})))

(defn- backgroundcolors []
  ["#00FFFF" "#0000FF" "#FF00FF" "#808080" "#808080" "#008000" "#00FF00" "#800000" "#000080" "#808000" "#FFA500" "#800080" "#FF0000" "#C0C0C0" "#008080" "#FFFF00" "#8A2BE2" "#A52A2A" "#DEB887" "#5F9EA0" "#DAA520" "#ADFF2F" "#F0FFF0" "#FF69B4" "#CD5C5C" "#4B0082"])

(defcomponent chartview-options-button [{:keys [data chart-js]} owner opts]
  (render
    [_]
    (dom/div
      #js {:className "input-group input-group-sm" :style #js {:textAlign "right"}}
      (dom/span
        #js {:className "input-group-btn"}
        (dom/button
          #js {:className "btn btn-default dropdown-toggle" :data-toggle "dropdown" :aria-expanded "false" :type "button"}
          (fawesome-icon "fa-cog" nil nil :lg))
        (apply dom/ul #js {:className "dropdown-menu dropdown-menu-right" :role "menu"}
          [(dom/li nil (dom/a #js {:onClick
                                    (fn [e]
                                      (js-export-pdf
                                        chart-js
                                        "png"))
                                    :className "fa fa-image"} (loc " Image")))
            (when data
              (dom/li nil (dom/a #js {:onClick #(js-export-csv data)
                                       :className "fa fa-download"} " CSV")) )])))))

(defcomponent chartjs-pie-graph [data owner opts]
  (init-state [_]
    {:uid (str (make-random))
      :update-fn (fn [s]
                   (let [u (om/get-state owner :uid) 
                          conf (:conf opts)
                          ;; conf settings
                          ylabelkey (:ylabelkey conf)
                          titlename (:title conf)
                          ytitle (:ytitle conf)
                          yvaluekey (:yvaluekey conf)
                          tooltip (:tooltiptitle conf)
                          type "pie"
                          ;; pie charts are limited to the first 20
                          labels (map ylabelkey (take 20 s))
                          dataa (map yvaluekey (take 20 s))
                          ;tooltipcallback (:tooltipcallback conf)
                          ;tooltipcaler (fn [x] (let [row (first (drop (get (js->clj x) "index") s))]
                          ;                       (tooltipcallback row)))
                          dataset [{:data dataa :backgroundColor (take (count dataa) (backgroundcolors))}]
                          options {:responsive true}
                     ;options (when tooltipcallback (assoc options :tooltips {:callbacks {:label tooltipcaler}}))

                          opt {:type type :data {:datasets dataset :labels labels} :options options}
                          ;_ (println opt)
                          ue (. js/document (getElementById u))
                          a (js/Chart. ue (clj->js opt))]
                     ((:chart-js-setter opts) a)
                     ))})

  ;; install the chart into the dom
  (did-mount [_] ((om/get-state owner [:update-fn]) data))

  ;; update the component
  (will-update [_ next-props next-state] 
    ((om/get-state owner [:update-fn]) next-props))

  (render [_]
    (odom/canvas {:id (om/get-state owner :uid) :class (str "chart " (:class opts)) :style {:height (str (:height opts) "px")}})))

(defcomponent chartjs-view [data owner opts] ;;  :- ChartOpts
  (init-state [_]
    {:current-range nil
      :chart-js nil ; the js var that is the chart. will be set via a fn. very dirty. very dirty.
      :update-chart (fn [] (let [blck (om/get-state owner [:chart-js])]
                             (.destroy blck)
                             ))
      })

  (render-state
    [_ state]
    (let [
           format (:format opts)
           time (:time opts)
           subtitle (:subtitle opts)
           ;; Apply the current range, if one exists
           time-ranges (:time-ranges opts)
           current-range (if-let [r (om/get-state owner :current-range)] r (-> time-ranges first :key))
           rows (if current-range (current-range data) data)
           rows (if-let [t (:row-transformer format)] (t rows) rows) 
           ;; Limit to the valid columns
           valid-columns (-> opts :format keys)
           rows (if (and valid-columns (count valid-columns)) (map #(select-keys % valid-columns) rows) rows)
           ;; Apply value transformer to rows
           rows (map (fn [r] (into {} (map (fn [[k v]] (if-let [t (-> format k :transformer)] [k (t v)] [k v]) ) r)) ) rows)
           ;; Get display settinsg
           settings (:settings opts)
           height (:height settings)
           class (:class settings)
           title (:title settings)
           conf (:conf opts)]
      
      (odom/div {:class "portlet light bordered"}
        (odom/div {:class "portlet-title"}
          (odom/div {:class "caption font-green-sharp"} (odom/div {:class "caption-subject bold"} title))
          (when subtitle
            (odom/div {:class "actions"}
              (odom/span {:class "label label-primary"} subtitle))))

        (odom/div {:class "portlet-body"}
          (g/row nil
            (g/col {:xs 12}
              (case (:type opts)
                :bar  (om/build chartjs-bar-graph rows {:opts {:class class :height height :conf conf
                                                                 :chart-js-setter #(om/set-state! owner :chart-js %)}})
                :pie (om/build chartjs-pie-graph rows {:opts {:class class :height height :conf conf
                                                               :chart-js-setter #(om/set-state! owner :chart-js %)}})
                :line (om/build chartjs-area-graph rows {:opts {:class class :height height :conf conf
                                                                 :chart-js-setter #(om/set-state! owner :chart-js %)}})))))
        (odom/div {:class "portlet-footer"}
          (g/row nil
            (g/col {:xs 10 :md 10}
              (om/build timerange-buttongroup {:ranges time-ranges :active current-range}
                {:opts {:click-fn
                         (fn [x]
                           ((:update-chart state))
                           (om/set-state! owner :current-range x)
                           )}}))
            (g/col
              {:xs 2 :md 2}
              (om/build chartview-options-button {:data rows :chart-js (om/get-state owner :chart-js)})
              )))
        ))))

(defcomponent not-enough-data-view [data owner opts]
  (render
    [_]
    (let [settings (:settings opts)
           title (:title settings)
           subtitle (:subtitle opts)]
      (odom/div {:class "portlet light bordered"}
        (odom/div {:class "portlet-title"}
          (odom/div {:class "caption font-green-sharp"} (odom/div {:class "caption-subject bold"} title))
          (when subtitle
            (odom/div {:class "actions"}
              (odom/span {:class "label label-primary"} subtitle))))

        (odom/div {:class "portlet-body"}
          (g/row nil
            (g/col {:xs 12}
              data
              )))))))

;; ---------------------------------------------------------------
;; Table Component
;; ---------------------------------------------------------------

(defcomponent table-view-header [{:keys [header sort-fn]} _] 
  (render
    [_]
    (odom/th {:on-click sort-fn :style {:cursor "pointer"}}
      (trnl header) ;(-> header name clojure.string/capitalize)
      (fawesome-icon "fa-arrows-v" nil {:className "table-arrow"
                                         } :border :fixed-width))))

(defn- table-row-progress [value]
  "return value as progress field. the value has to
   be an integer in the 0-100 range"
  ;; FIXME: Move to CSS
  (dom/div #js {:title (str value "%")  :className "progress" :style #js {:height "14px" :marginTop "3px" :marginBottom "3px"}}
    (dom/div #js {:className "progress-bar progress-bar-info"
                   :role "progressbar"
                   :aria-valuenow (str value) :aria-valuemin "0" :aria-valuemax "100"
                   :style #js {:width (str value "%") :fontSize "9px" :lineHeight "14px"}} (str value "%"))))

(defn- table-row-image [value scaled? scale-fn vk row]
  "return value as link to an image"
  (dom/img
    #js {:onClick (fn []
                    ;(println "vk" vk)
                    ;(println "row" (into {} row))
                    (if-let [uid (vk (into {} row))]
                      (osxaction :photo uid)
                      (scale-fn)
                      )
                    ) :src value :className (str "img-thumbnail" (when scaled? " img-touchsizable"))}))

(defn- table-row-url [value]
  (let [hostname (parse-url-hostname value)]
    (if (and (not (nil? value)) (> (count value) 0))
      (dom/a #js {:onClick (fn [] (osxaction :link value)) :className "btn btn-xs btn-success" :title value}
        (dom/span #js {:className "glyphicon"})
        (str " " hostname))
      (dom/span nil "")
      )))

(defn- table-row-iguser [value vk row]
  (dom/a #js {:onClick (fn []
                         (when-let [uid (vk (into {} row))]
                           (osxaction :user uid)
                           )) :className "btn btn-xs btn-success" :title value}
    (dom/span #js {:className "glyphicon"})
    (str "@" value)))

(defn- table-row-tag [value]
  (dom/a #js {:onClick (fn []
                         (osxaction :tag value)
                         ) :className "btn btn-xs btn-success" :title value}
    (dom/span #js {:className "glyphicon"})
    (str "#" value)))

(defn- table-row-score [value max]
  "a button indicating a score on a scale from 0 to max"
  (odom/span {:class "badge"} (str (/ (int value) (int max)))))

(defn- table-row-percent [value]
  "a percentage-value"
  (odom/span {:class "badge"} (str (int (* value 100)))))

(defn- table-row-growth [value]
  "a growth button"
  (odom/span {:class "badge"} (str (when (> value 0) "+") value)))

(defn- table-row-date [value]
  (.toDateString value))

(defn- table-row-time [value]
  (.toTimeString value))

(defn- table-row-timestamp [value]
   (.format (goog.i18n.DateTimeFormat. "dd/MM hh:mm") (goog.date.DateTime/fromTimestamp value)))

(defn- table-row-isodatetime [value]
  (let [parsedDate (goog.date/fromIsoString value)]
    (.format (goog.i18n.DateTimeFormat. "dd/MM hh:mm") parsedDate)))

(defn- table-row-datetime [value]
  (.toLocaleString value))

(defn index-of [coll value]
  (some (fn [[idx item]] (if (= value item) idx))
    (map-indexed vector coll)))

(defcomponent table-row [row owner opts]
  (init-state [_] {:image-scaled true})
  (render-state
    [_ state]
    (odom/tr
      (let [valid-columns (:columns opts)
                                        ;row (if (and valid-columns (count valid-columns)) (map #(select-keys % valid-columns) row) rows)
             ;_ (println "valrow" valid-columns)
             ;_ (println "row" row)
             limrow (if (and valid-columns (count valid-columns)) (select-keys (into {} row) valid-columns) row)
             ;_ (println "limrow" limrow)
             ]
        (for [c limrow] ; (sort-by #(index-of (:columns opts) %) row) 
          (let [k (first c)
                 v (last c)
                 f (k opts)
                 a (if-let [al (:align (k opts))] al "left")
                 t (:type f)
                 vk (:vkey f)
                 m (:maxscore f)
                 m (if (nil? m) 1 m)
                 ;; Mix in all the params we need
                 p {:style {:textAlign a :verticalAlign "middle"}}
                 p (if (:action f) (assoc p :onClick (partial (:action f) row c)) p)
                 p (if (:class f) (assoc p :className (:class f)) p)]
            (dom/td (clj->js p)
              (case t
                nil v
                :date (table-row-date v)
                :time (table-row-time v)
                :tag (table-row-tag v)
                :datetime (table-row-datetime v)
                :timestamp (table-row-timestamp v)
                :isodatetime (table-row-isodatetime v)
                :url (table-row-url v)
                :percent (table-row-percent v)
                :iguser (table-row-iguser v vk row)
                :growth (table-row-growth v)
                :score (table-row-score v m)
                :sparkline (om/build sparkline-component v {:opts (:ctx f)})
                :progress (table-row-progress v)
                :image-url (table-row-image v (:image-scaled state) #(om/set-state! owner :image-scaled (not (:image-scaled state))) vk row))
              )))
        )
      )))

(defn filtered-data [afilter data]
  "filter each row (all fields) by search term afilter.
   But only if the contents of the field is indeed a string"
  (if (and afilter (> (count afilter) 0))
    (filter (fn [row]
              (some #(= % true)
                (map (fn [cell]
                       (and (string? (last cell))
                         (boolean (re-find (re-pattern (str "(?i)" afilter)) (last cell)))))
                  row))) data)
    data))

(defn sorted-data [key direction data]
  (if key
    ((if direction reverse identity)
      (sort-by key data))
    data))

(defn change-sorting [owner key]
  ;; invert direction on second click
  (when (= (om/get-state owner :sort) key)
    (om/set-state! owner :sort-dir (not (om/get-state owner :sort-dir))))
  (om/set-state! owner :sort key))

(defn header-types [data titles sort-fn]
  ;(println "htypes" data titles)
  (map #(hash-map :header (if-let [t (% titles)] t %) :sort-fn (partial sort-fn %))
    (-> data
      first
      keys)))

(def TableOpts
  "The possible table options"
  {(sce/optional-key :format) {sce/Keyword ; the name of the column 
                                {(sce/optional-key :type) (sce/enum :image-url :url :progress :sparkline :growth :iguser :date :time :datetime)
                                  (sce/optional-key :align) sce/Str ; alignment
                                  (sce/optional-key :ctx) sce/Any ; context for the row type. will be handed over
                                  (sce/optional-key :class) sce/Str ; css class for td
                                  (sce/optional-key :action) sce/Any ; fn (row field) -> nil
                                  (sce/optional-key :custom-sort) sce/Any} ; fn (any) -> Num
                                } 
    (sce/optional-key :columns) [sce/Keyword] ; visible columns
    (sce/optional-key :time-ranges) [{:title sce/Str :key sce/Keyword}] ; the different time ranges, if there are
    (sce/optional-key :settings) {(sce/optional-key :export) sce/Bool
                                   (sce/required-key :title) sce/Str} ; any settings.
    })

(defn table-footer [time-ranges current-range owner rows export-fn]
  (odom/div {:class "portlet-footer"}
    (g/row nil
      (g/col {:xs 7 :md 7}
        (om/build timerange-buttongroup {:ranges time-ranges :active current-range}
          {:opts {:click-fn #(om/set-state! owner :current-range %)}}))
      (g/col
        {:xs 5 :md 5}
        (dom/div #js {:className "input-group input-group-sm"}
          (dom/div #js {:className "input-group-addon"}
            (fawesome-icon "fa-search" nil nil ""))
          (dom/input #js {:type "text" :className "form-control" :placeholder (loc "Filter")
                           :value (om/get-state owner :filter)
                           :onChange #(om/set-state! owner :filter (.. % -target -value))})
          (dom/span #js {:className "input-group-btn"}
            (dom/button #js {:className "btn btn-default dropdown-toggle" :data-toggle "dropdown" :aria-expanded "false" :type "button"}
              (fawesome-icon "fa-cog" nil nil :lg))
            (apply dom/ul #js {:className "dropdown-menu dropdown-menu-right" :role "menu"}
              (flatten
                [(when (and (> (count (om/get-state owner :filter)) 0)
                         (not (some #(= % (om/get-state owner :filter))
                                (om/get-state owner :recent-searches))))
                   [(dom/li nil (dom/a #js {:onClick #(om/update-state! owner :recent-searches (partial cons (om/get-state owner :filter)))}
                                  (loc "Save Search")))
                     (dom/li #js {:className "divider"})])
                  (when (> (count (om/get-state owner :recent-searches)) 0)
                    [(dom/li nil (dom/a #js {:onClick #()} (dom/strong nil (loc "Saved Searches"))))
                      (for [search (om/get-state owner :recent-searches)]
                        (dom/li nil (dom/a #js {:onClick #(om/set-state! owner :filter search)} search)))
                      (dom/li #js {:className "divider"})])
                  (dom/li nil (dom/a #js {:onClick #(js-export-csv rows)
                                           :className "fa fa-download"} (loc " Export as CSV") ))]))))))))

(defn- wrap-in-scroller [scroller height & more]
  (if scroller
    (odom/div
      {:class "portlet-body"}
      (apply odom/div {:class "scroller" :style {:height (str height "px") :overflow "hidden" :width "auto"}
                        :data-always-visible 1 :data-rail-visible1 0 :data-handle-color "#D7DCE2"}
        more))
    (apply odom/div {:class "portlet-body"} more)
    ))

(defcomponent table-view [data owner opts :- TableOpts]
  (init-state [_] {:sort :nil
                    :sort-dir nil
                    :filter nil
                    :recent-searches []
                    :current-range nil
                    :update-fn (fn []
                                 (let [n (om.core/get-node owner)
                                        h (str (-> opts :format :height) "px")
                                        e (.find (js/jQuery n) ".general-item-list")]
                                   (enable-slim-scroll e h)))})

  ;; activate the slimscroll
  (did-mount [_] ((om/get-state owner [:update-fn])))

  ;; activate the slimscroll
  (will-update [_ _ _] ((om/get-state owner [:update-fn])))

  (render-state
    [_ state]   
    (let [valid-columns (:columns opts)
           time-ranges (:time-ranges opts)
           time (:time opts)
           subtitle (:subtitle opts)
           current-range (if-let [r (om/get-state owner :current-range)] r ; use the user-selected-one
                           (-> time-ranges first :key)) ; or the first one containing items
           format (:format opts)
           rowtitles (into {} (filter #(last %) (map (fn [[k v]] [k (:title v)]) format)))
           export-fn (-> opts :settings :export)
           title (-> opts :settings :title)
           caption (-> opts :settings :caption)
           scroll-height (-> opts :format :height)
           scroll (> scroll-height 0)
           rows (if current-range (current-range data) data)
           rows (if-let [t (:row-transformer format)] (t rows) rows) 
           rows (map (fn [row] (sort-by #(index-of (:columns opts) (first %)) row)) rows)
           ;rows (if (and valid-columns (count valid-columns)) (map #(select-keys % valid-columns) rows) rows)
           onerow (if (and valid-columns (count valid-columns))
                    (map #(select-keys (into {} %) valid-columns) rows)
                     rows)
           ;_ (println "valid-cols" valid-columns)
           ;_ (println "row" (first rows))
           ;_ (println "onerow" onerow)
           current-sort (:sort state)]
      (odom/div {:class "portlet light"}
        (odom/div
          {:class "portlet-title"}
          (odom/div {:class "caption font-green-sharp"} (odom/div {:class "caption-subject bold"} title))
          (when subtitle
            (odom/div {:class "actions"}
              (odom/span {:class "label label-primary"} subtitle)
              )))

        (when caption
          (odom/div {:class "row"} (odom/p {:class "table-caption"} caption)))

        (table-footer time-ranges current-range owner rows export-fn)

        (wrap-in-scroller scroll scroll-height
          (odom/div {:class "general-item-list"}
            (odom/table {:class "table table-hover table-light table-for-scrolling"}
              (dom/thead nil
                (apply dom/tr nil
                  (om/build-all table-view-header (header-types onerow
                                                    rowtitles (partial change-sorting owner)))))
              (if (empty? rows)
                (om/build-all table-row [{:description (loc "No Data")}] {:opts {:description {:align "center"}}})
                (odom/tbody
                  (om/build-all table-row
                    (sorted-data (if-let [cs (-> format current-sort :custom-sort)] cs current-sort)
                      (:sort-dir state)
                      (filtered-data (:filter state)
                        rows))
                    {:opts (assoc format :columns (:columns opts))}))))))

        (table-footer time-ranges current-range owner rows export-fn))
      )))

(def Comment
  "Structure of a comment, mainly for docs. Only the fields necessary for the table"
  {(sce/required-key :image_id) sce/Str
    (sce/required-key :comment_id) sce/Str
    (sce/required-key :commenter_name) sce/Str
    (sce/required-key :commenter_image) sce/Str
    (sce/required-key :commenter_website) sce/Str
    (sce/required-key :created_time) sce/Any
    (sce/required-key :comment_text) sce/Str})

(defcomponent comment-table-item [data :- Comment owner opts]
  (render [_]
    (odom/div
      {:class "item"}
      (odom/div
        {:class "item-head"}
        (odom/div
          {:class "item-details"}
          (odom/img {:class "item-pic" :src (:commenter_image data)})
          (odom/a {:class "item-name primary-link" :href (str "http://www.instagram.com/" (:commenter_name data))} (:commenter_name data))
          (odom/span {:class "item-label"} (:created_time data)))
        (odom/span {:class "item-status"} (odom/span {:class "badge badge-empty badge-success"}) (loc " New") ))
      (odom/div {:class "item-body"}
        (:comment_text data))
      )))

(def CommentTableOpts
  "The possible table options"
  {(sce/optional-key :time-ranges) [{:title sce/Str :key sce/Keyword}] ; the different time ranges, if there are
    (sce/required-key :format) {(sce/required-key :height) sce/Str}
    (sce/optional-key :settings) {(sce/optional-key :export) sce/Bool
                                   (sce/required-key :title) sce/Str} ; any settings.
    })

(defcomponent comment-table-view [data :- [Comment] owner opts :- CommentTableOpts]
  (init-state [_] {:sort :nil
                    :sort-dir nil
                    :filter nil
                    :recent-searches []
                    :current-range nil
                    :update-fn (fn []
                                 (let [n (om.core/get-node owner)
                                        h (str (-> opts :format :height) "px")
                                        e (.find (js/jQuery n) ".general-item-list")]
                                   (enable-slim-scroll e h)))})

  ;; activate the slimscroll
  (did-mount [_] ((om/get-state owner [:update-fn])))

  ;; activate the slimscroll
  (will-update [_ _ _] ((om/get-state owner [:update-fn])))

  (render-state
    [_ state]   
    (let [valid-columns (:columns opts)
           time-ranges (:time-ranges opts)
           current-range (if-let [r (om/get-state owner :current-range)] r (-> time-ranges first :key))
           format (:format opts)
           export-fn (-> opts :settings :export)
           time (:time opts)
           title (-> opts :settings :title)
           rows (if current-range (current-range data) data)
           rows (if (and valid-columns (count valid-columns)) (map #(select-keys % valid-columns) rows) rows)
           current-sort (:sort state)]
      (odom/div {:class "portlet light"}
        (odom/div
          {:class "portlet-title"}
          (odom/div {:class "caption font-green-sharp"} (odom/div {:class "caption-subject bold"} title))
          (odom/div {:class "actions"} (odom/span {:class "label label-info"} time)))

        (odom/div
          {:class "portlet-body"}
          (wrap-in-scroller true 305 (odom/div {:class "general-item-list"}
                                       (om/build-all comment-table-item
                                         (filtered-data (:filter state)
                                           data)))))

        (table-footer time-ranges current-range owner rows export-fn))
      )))

;; ---------------------------------------------------------------
;; Layout Helpers
;; ---------------------------------------------------------------

(defn- grid-full-width [& items]
  (g/row nil (g/col {:xs 12 :md 12}
               items)))

(defn- grid-half-width [item1 item2]
  (g/row nil
    (g/col {:xs 12 :md 6} item1)
    (g/col {:xs 12 :md 6} item2)))

(defn- grid-three [i1 i2 i3]
  (g/row nil
    (g/col {:xs 4 :md 4} i1)
    (g/col {:xs 4 :md 4} i2)
    (g/col {:xs 4 :md 4} i3)))

(defn- grid-four [i1 i2 i3 i4]
  (g/row nil
    (g/col {:xs 6 :md 3} i1)
    (g/col {:xs 6 :md 3} i2)
    (g/col {:xs 6 :md 3} i3)
    (g/col {:xs 6 :md 3} i4)))

;; ---------------------------------------------------------------
;; Pages.
;; ---------------------------------------------------------------

(defmulti application-views (fn [cursor owner opts]
                              (-> cursor :current :page)))

(defn- pageinfo-from-cursor [cursor]
  (let [current (-> cursor :current)
         page (->> pages (filter #(= (:key %) (:page current))) first)
         params (dissoc current :page)]
    {:name (:name page)
      :icon (:icon page)
      :route (:route page)
      :params params}))


(defmethod application-views :dashboard [cursor owner opts]
  (reify
    om/IRender
    (render [_]
      (let [page (pageinfo-from-cursor cursor)
             stats (:statistics cursor)]
        (grid-full-width
          (odom/div {:align "center" :style {:margin-top "150px"}}
            (dom/h3 #js {:style {:marginTop "100px" :textAlign "center"}}
              (loc "Loading Data..."))))))))

(defmethod application-views :performance [cursor owner opts]
  (reify
    om/IRender
    (render [_]
      (let [page (pageinfo-from-cursor cursor)
             stats (:statistics cursor)]
        (odom/div
          (grid-four
            (om/build gauge (gauge-data cursor :performance-metrics)
              {:opts {:color "green"
                       :icon "fa-image"
                       :label (loc "Uploads")
                       ;:target {:field :images-last-week :value (-> cursor :user :targets :weekly-upload-target)
                       ;          :title "This week"}
                       :target-action
                       (fn [new-target new-dim]
                         ;; store in the app-state
                         (om/update! cursor [:user :targets :weekly-upload-target] {:num new-target :dim new-dim})
                         ;; and store in the mac app.
                         (store-gauge-target :weekly-upload-target new-target new-dim)
                         )
                       :columns
                       {;:images-last-week {:title "Last Week"}
                         :images {:title "Total" :transformer #(goog.string/format "%d" %)}
                         :images-last-week {:title (loc "This Week")}
                         :images-months {:title (loc "Month") :transformer #(goog.string/format "%.1f" %)}
                         :images-weeks {:title (loc "Week") :transformer #(goog.string/format "%.1f" %)}
                         :images-days {:title (loc "Day") :transformer #(goog.string/format "%.1f" %)}
                         }}})
            (om/build gauge (gauge-data cursor :performance-metrics)
              {:opts {:color "red-flamingo"
                       :icon "fa-heart"
                       :label (loc "Likes")
                       ;:target {:field :new-likes :value (-> cursor :user :targets :daily-like-target)}
                       :columns ;; if we have zero new likes, we make all likes first
                       {:new-likes {:title (loc "New") :target (gauge-target cursor :performance-metrics :new-likes :daily-like-target)}
                         :all-likes {:title (loc "All")}
                         :likes-months {:title (loc "Per Month") :transformer #(goog.string/format "%.1f" %)}
                         :likes-weeks {:title (loc "Week") :transformer #(goog.string/format "%.1f" %)}
                         :likes-days {:title (loc "Day") :transformer #(goog.string/format "%.1f" %)}
                         }
                       }})
            (om/build gauge (gauge-data cursor :performance-metrics) ; :daily-comment-target
              {:opts {:color "blue-sharp"
                       :icon "fa-comment"
                       ;;:target {:field :new-comments :value (-> cursor :user :targets :daily-comment-target)}
                       :label (loc "Comments")
                       :columns
                       {:new-comments {:title (loc "New")}
                         :all-comments {:title (loc "All")}

                         :comments-months {:title (loc "Per Month") :transformer #(goog.string/format "%.1f" %)}
                         :comments-weeks {:title (loc "Week") :transformer #(goog.string/format "%.1f" %)}
                         :comments-days {:title (loc "Day") :transformer #(goog.string/format "%.1f" %)}

                         }}}
              )
            (om/build gauge (gauge-data cursor :performance-metrics) ;  :daily-follow-target
              {:opts {:color "blue-madison"
                       :icon "fa-user-plus"
                       :label (loc "Followers")
                       ;;:target {:field :new-followers :value (-> cursor :user :targets :daily-follow-target)}
                       :columns
                       {:new-followers {:title (loc "New")}
                         :followers {:title (loc "All")}
                         :followers-months {:title (loc "Per Month") :transformer #(goog.string/format "%.1f" %)}
                         :followers-weeks {:title (loc "Week") :transformer #(goog.string/format "%.1f" %)}
                         :followers-days {:title (loc "Day") :transformer #(goog.string/format "%.1f" %)}
                         }}}))

          (grid-full-width
            (om/build chartjs-view (-> stats :relative-comment-growth :data)
              {:opts {:conf {:ylabelkey :ce :yvaluekey :count :ytitle (loc "Comments per day") :title (loc "Comments") :dateFormat "YYYY-MM-DD" :tooltiptitle (loc "# of comments")}
                       :type :line
                       :time (stat-time stats :relative-comment-growth)
                       :format {:count nil :ce nil}
                       :time-ranges [{:title (loc "Standard") :key :standard} {:title (loc "Accumulation") :key :increasing}]
                       :settings {:height 400
                                   :title (loc "Comment Growth")}}}))

          (grid-full-width
            (om/build chartjs-view (-> stats :account-image-upload-growth :data)
              {:opts {:conf {:ylabelkey :date :yvaluekey :count :ytitle (loc "Upload Growth") :title (loc "Upload Growth") :dateFormat "YYYY-MM" :tooltiptitle (loc "# of uploads")}
                       :type :line
                       :time (stat-time stats :account-image-upload-growth)
                       :subtitle (loc "Amount of items you uploaded per day")
                       :format {:count nil :date nil :dates nil}
                       :time-ranges [{:title (loc "Standard") :key :standard} {:title (loc "Accumulation") :key :increasing}]
                       :settings {:height 400
                                   :title (loc "Upload Growth")}}}))
          (grid-full-width
            (if (> 4 (count (-> stats :relative-follower-growth :data)))
              (om/build not-enough-data-view (loc "Photodesk didn't collect enough data yet to display anything here.")
                {:opts {:subtitle (loc "The amount of new followers per day")
                         :settings {:height 200 :title (loc "Follower Growth")}}})
              (om/build chartjs-view (-> stats :relative-follower-growth :data)
                {:opts {:conf
                         {:ylabelkey :date :yvaluekey :count :ytitle (loc "Followers") :title (loc "New Followers") :dateFormat "YYYY-MM-DD" :tooltiptitle (loc "# of followers")}
                         :type :line
                         :time (stat-time stats :relative-follower-growth)
                         :subtitle (loc "The amount of new followers per day")
                         :format {:count nil :date nil}
                         :settings {:height 400
                                     :title (loc "Follower Growth")}}})
              ))



          (grid-half-width
            (if (>= 1 (-> stats :upload-follower-change :data :follows count))
              (om/build not-enough-data-view (loc "Photodesk didn't collect enough data yet to display anything here.")
                {:opts {:subtitle ""
                         :settings {:height 200 :title (loc "Followers by Uploads")}}})

              (om/build table-view (-> stats :upload-follower-change :data :follows)
                {:opts {:settings {:title (loc "Followers by Upload")}
                         :columns [:fol-change :created :low_resolution ] ; :link
                         :time (stat-time stats :upload-follower-change)
                         :format {:height 300
                                        ;:link {:type :url}
                                   :created {:type :timestamp}
                                   :fol-change {:type :growth :title (loc "Change")}
                                   :low_resolution {:type :image-url
                                                     :vkey :id}}}}))


            (if (>= 1 (-> stats :upload-follower-change :data :unfollows count))
              (om/build not-enough-data-view (loc "Photodesk didn't collect enough data yet to display anything here.")
                {:opts {:subtitle ""
                         :settings {:height 200 :title (loc "Unfollowers by Upload")}}})

              (om/build table-view (-> stats :upload-follower-change :data :unfollows)
                {:opts {:settings {:title (loc "Unfollowers by Upload")}
                         :columns [:fol-change :created :low_resolution ] ;:link
                         :time (stat-time stats :upload-follower-change)
                         :format {:height 300
                                        ;:link {:type :url}
                                   :created {:type :timestamp}
                                   :fol-change {:type :growth :title (loc "Change")}
                                   :low_resolution {:type :image-url
                                                     :vkey :id
                                                     :action (fn [row field] )}}}}))

          ))))
    om/IDidMount
    (did-mount [_]
      (bundle-calls :performance))))

(defmethod application-views :content [cursor owner opts]
  (reify
    om/IRender
    (render [_]
      (let [page (pageinfo-from-cursor cursor)
             stats (:statistics cursor)]
        (odom/div

          (grid-full-width
            (om/build chartjs-view (-> stats :grouped-likes :data)
              {:opts {:conf
                       {:ylabelkey :key :yvaluekey :total :ytitle (loc "Likes per Image") :title (loc "Likes per Image") :dateFormat nil :tooltiptitle (loc "# likes")
                         :tooltipcallback (fn [val] (str (goog.string/format "%d%%" (:total val)) ": " (:likes val) (loc " likes on ") (:count val) (loc " images for ") (:key val)))}
                       :type :bar
                       :time (stat-time stats :grouped-likes)
                       :subtitle (loc "Aggregated likes for different categories")
                       :format {:height 500 :count nil :total nil :likes nil :key nil}
                       :time-ranges [{:title (loc "Tags") :key :tags} {:title (loc "Mentions") :key :mentions} {:title (loc "Filters") :key :filters} {:title (loc "Type") :key :types}]
                       :settings {:height 400
                                   :title (loc "Likes per category")}}}))

          (om/build chartjs-view (-> stats :grouped-comments :data)
            {:opts {:conf
                     {:ylabelkey :key :yvaluekey :total :ytitle (loc "Comments per Image") :title (loc "Comments per Image") :dateFormat nil :tooltiptitle (loc "# comments")
                       :tooltipcallback (fn [val] (str (:likes val) (loc " comments on ") (:count val) (loc " images for ") (:key val)))}
                     :type :bar
                     :time (stat-time stats :grouped-comments)
                     :subtitle (loc "Aggregated comments for different categories")
                     :format {:height 500 :count nil :total nil :likes nil :key nil}
                     :time-ranges [{:title (loc "Tags") :key :tags} {:title (loc "Mentions") :key :mentions} {:title (loc "Filters") :key :filters} {:title (loc "Type") :key :types}]
                     :settings {:height 400
                                 :title (loc "Comments per category")}}})

          (grid-half-width
            (om/build table-view (-> stats :most-active-commenters :data)
              {:opts {:settings {:title (loc "Most active Commenters")}
                       :columns [:name :count]
                       :format {:height 500 :name {:type :iguser :vkey :id}
                                 :tag {:align "right"}}}})

            (om/build table-view (-> stats :most-active-likers :data)
              {:opts {:settings {:title (loc "Most active Likers")}
                       :columns [:name :count]
                       :format {:height 500 :name {:type :iguser :vkey :id}
                                 :tag {:align "right"}}}}))

;;          (grid-half-width
;;            (om/build table-view (-> stats :filter-metrics :data)
;;              {:opts {:settings {:title (loc "Filters on our Pictures")}
;;                       :columns [:filter :count]
;;                       :format {:height 500
;;                                 :row-transformer #(:account %)
;;                                 :filter nil
;;                                 :count {:type :score}}
;;                       :time-ranges [{:title (loc "Total") :key :total}
;;                                      {:title (loc "Yearly") :key :yearly}
;;                                      {:title (loc "Monthly") :key :monthly}
;;                                      {:title (loc "Weekly") :key :weekly}
;;                                      {:title (loc "Daily") :key :daily}
;;                                      ]
;;                       :time (stat-time stats :filter-metrics)
;;                       }})
;;
;;
;;            (om/build chartjs-view (-> stats :filter-metrics :data)
;;              {:opts {:conf (chartjs-pie-conf :filter :count {:depth3D 10})
;;                       :type :pie
;;                       :format {:row-transformer #(:account %) :filter nil :count nil}
;;                       :subtitle (loc "Filters on our pictures pie chart")
;;                       :time-ranges [
;;                                      {:title (loc "Yearly") :key :yearly}
;;                                      {:title (loc "Monthly") :key :monthly}
;;                                      {:title (loc "Weekly") :key :weekly}
;;                                      {:title (loc "Daily") :key :daily}
;;                                      {:title (loc "Total") :key :total}
;;                                      ]
;;                       :settings {:height 530
;;                                   :title ""}}})
;;
;;            )

          (grid-half-width
            (om/build table-view (-> stats :best-performing-images :data :upcoming)
              {:opts {:settings {:title (loc "Currently Trending Pictures")}
                       :columns [:likes :comments :low_resolution :allLikes :score] ;:link 
                       :time (stat-time stats :best-performing-images)
                       :format {:height 300
                                 ;:link {:type :url}
                                 :score {:type :percent}
                                 :allLikes {:type :sparkline
                                             :custom-sort (partial reduce +)
                                             :ctx {:type "line" :barColor "#7799FF"}}
                                 :low_resolution {:type :image-url
                                                   :vkey :id
                                                   :action (fn [row field] )}}}})

            (om/build table-view (-> stats :best-performing-images :data :evergreens)
              {:opts {:settings {:title ""}
                       :columns [:likes :comments :low_resolution :allLikes :score] ;:link 
                       :subtitle (loc "Pictures continously gaining likes and comments")
                       :time (stat-time stats :best-performing-images)
                       :format {:height 300
                                 ;:link {:type :url}
                                 :score {:type :percent}
                                 :allLikes {:title (loc "Trend") :type :sparkline
                                             :custom-sort (partial reduce +)
                                             :ctx {:type "line" :barColor "#7799FF"}}
                                 :low_resolution {:type :image-url
                                                   :vkey :id
                                                   :action (fn [row field] )}}}}))

          (grid-half-width

            (om/build table-view (-> stats :most-important-images :data)
              {:opts {:settings {:title (loc "Highest Rated Content")}
                       :columns [:low_resolution :created :comments :likes] ;:link 
                       :format {:height 371
                                 :low_resolution {:type :image-url :vkey :id}
                                 ;:link {:type :url}
                                 :created {:type :isodatetime}}
                       :time (stat-time stats :most-important-images)
                       :time-ranges [{:title (loc "Total") :key :total}
                                      {:title (loc "Yearly") :key :yearly}
                                      {:title (loc "Monthly") :key :monthly}
                                      {:title (loc "Weekly") :key :weekly}
                                      {:title (loc "Daily") :key :daily}
                                      ]}})

            (om/build chartjs-view (-> stats :image-content-types :data :account)
              {:opts {:conf (chartjs-pie-conf :type :count {:depth3D 10})
                       :type :pie
                       :format {:count nil :type nil}
                       :settings {:height 400
                                   :title (loc "Image vs. Video in our content")}}})

            )

          (grid-half-width
            (om/build chartjs-view (-> stats :most-used-my-images :data :101)
              {:opts {:conf (chartjs-pie-conf :name :count {:depth3D 10})
                       :type :pie
                       :format {:name nil :count nil}
                       :time-ranges [{:title (loc "Total") :key :total}
                                      {:title (loc "Yearly") :key :yearly}
                                      {:title (loc "Monthly") :key :monthly}
                                      {:title (loc "Weekly") :key :weekly}
                                      {:title (loc "Daily") :key :daily}
                                      ]
                       :settings {:height 530
                                   :title (loc "Tags on our images")}}})

            (om/build table-view (-> stats :most-used-my-images :data :100)
              {:opts {:settings {:title ""}
                       :columns [:name :count]
                       :format {:height 500 
                                 :count {:type :score}
                                 :name {:type :tag}}
                       :time-ranges [{:title (loc "Total") :key :total}
                                      {:title (loc "Yearly") :key :yearly}
                                      {:title (loc "Monthly") :key :monthly}
                                      {:title (loc "Weekly") :key :weekly}
                                      {:title (loc "Daily") :key :daily}
                                      ]
                       :time (stat-time stats :most-used-my-images)
                       :subtitle (loc "Tags on our images pie chart")
                       }})

            )

          )))

    om/IDidMount
    (did-mount [_]
      (bundle-calls :content))))


(defmethod application-views :audience [cursor owner opts]
  (reify
    om/IRender
    (render [_]
      (let [page (pageinfo-from-cursor cursor)
             stats (:statistics cursor)]
        (odom/div 

          (comment grid-full-width
            (om/build table-view (-> stats :important-followers :data)
              {:opts {:settings {:title (loc "Important Followers")}
                       :columns [:image_count :follow_count :follower_count :username :full_name :website :profile_picture :bio :igscore]
                       :format {:height 500 :profile_picture {:type :image-url :vkey nil}
                                 :website {:type :url}
                                 :username {:type :iguser :vkey :id}
                                 ;:link {:type :url}
                                 :igscore {:type :percent}
                                 :created {:type :datetime}}
                       :time (stat-time stats :important-followers)
                       :subtitle (loc "Some of your most important followers")
                       }}))

          (grid-full-width
            (om/build table-view (-> stats :important-commenters :data)
              {:opts {:settings {:title (loc "Important Commenters") }
                       :columns [:igscore :commenter_name :commenter_followers :commenter_follows :commenter_media :commenter_bio :commenter_website :commenter_image]
                       :format {:commenter_image {:type :image-url :vkey nil}
                                 :commenter_name {:type :iguser :vkey :commenter_id}
                                 :commenter_website {:type :url}
                                 :igscore {:type :score :maxscore 1}
                                 :height 315}
                       :time (stat-time stats :important-commenters)
                       :subtitle (loc "Some of your most important commenters")
                       :time-ranges [{:title (loc "Total") :key :total}
                                      {:title (loc "Yearly") :key :yearly}
                                      {:title (loc "Monthly") :key :monthly}
                                      {:title (loc "Weekly") :key :weekly}
                                      {:title (loc "Daily") :key :daily}
                                      ]}}))

          (grid-half-width
            (om/build table-view (-> stats :mostused-followers :data :tags)
              {:opts {:settings {:title (loc "Tag Analysis of your followers")}
                       :columns [:name :count]
                       :format {:height 500 :name {:type :tag}}
                       :time (stat-time stats :mostused-followers)
                       :time-ranges [
                                      {:title (loc "In captions") :key :caption}
                                      {:title (loc "In images' comments") :key :comments}
                                      {:title (loc "In bios") :key :bio}
                                      ]}})

            (om/build table-view (-> stats :mostused-commenters :data :tags)
              {:opts {:settings {:title (loc "Tag Analysis of your commenters")}
                       :columns [:name :count]
                       :format {:height 500 :name {:type :tag}}
                       :time (stat-time stats :mostused-commenters)
                       :time-ranges [
                                      {:title (loc "In images' comments") :key :comments}
                                      {:title (loc "In bios") :key :bio}
                                      ]}})

            )

          (grid-half-width
            (om/build table-view (-> stats :mostused-followers :data :words)
              {:opts {:settings {:title (loc "Word Analysis of your followers")}
                       :columns [:name :count]
                       :format {:height 500}
                       :time (stat-time stats :mostused-followers)
                       :time-ranges [
                                      {:title (loc "In captions") :key :caption}
                                      {:title (loc "In images' comments") :key :comments}
                                      {:title (loc "In bios") :key :bio}
                                      ]}})

            (om/build table-view (-> stats :mostused-commenters :data :words)
              {:opts {:settings {:title (loc "Word Analysis of your commenters")}
                       :columns [:name :count]
                       :format {:height 500}
                       :time (stat-time stats :mostused-commenters)
                       :time-ranges [
                                      {:title (loc "In images' comments") :key :comments}
                                      {:title (loc "In bios") :key :bio}
                                      ]}})
            )

          (grid-half-width
            (om/build table-view (-> stats :mostused-followers :data :mentions)
              {:opts {:settings {:title (loc "@Mention Analysis of your followers")}
                       :columns [:name :count]
                       :format {:height 500}
                       :time (stat-time stats :mostused-followers)
                       :time-ranges [
                                      {:title (loc "In captions") :key :caption}
                                      {:title (loc "In images' comments") :key :comments}
                                      {:title (loc "In bios") :key :bio}
                                      ]}})

            (om/build table-view (-> stats :mostused-commenters :data :mentions)
              {:opts {:settings {:title (loc "@Mention Analysis of your commenters")}
                       :columns [:name :count]
                       :format {:height 500}
                       :time (stat-time stats :mostused-commenters)
                       :time-ranges [
                                      {:title (loc "In images' comments") :key :comments}
                                      {:title (loc "In bios") :key :bio}
                                      ]}})

            )

;;          (grid-half-width
;;            (om/build table-view (-> stats :filter-metrics :data)
;;              {:opts {:settings {:title (loc "Filters of your followers")}
;;                       :columns [:filter :count]
;;                       :format {:height 500
;;                                 :row-transformer #(:follower %)
;;                                 :filter nil
;;                                 :count {:type :score}}
;;                       :time-ranges [{:title (loc "Total") :key :total}
;;                                      {:title (loc "Yearly") :key :yearly}
;;                                      {:title (loc "Monthly") :key :monthly}
;;                                      {:title (loc "Weekly") :key :weekly}
;;                                      {:title (loc "Daily") :key :daily}
;;                                      ]
;;                       :time (stat-time stats :filter-metrics)
;;                       }})
;;
;;
;;            (om/build chartjs-view (-> stats :filter-metrics :data)
;;              {:opts {:conf (chartjs-pie-conf :filter :count {:depth3D 10})
;;                       :type :pie
;;                       :format {:row-transformer #(:follower %) :filter nil :count nil}
;;                       :time-ranges [
;;                                      {:title (loc "Yearly") :key :yearly}
;;                                      {:title (loc "Monthly") :key :monthly}
;;                                      {:title (loc "Weekly") :key :weekly}
;;                                      {:title (loc "Daily") :key :daily}
;;                                      {:title (loc "Total") :key :total}
;;                                      ]
;;                       :settings {:height 530
;;                                   :title ""}}})
;;
;;            )

          (grid-half-width
            (om/build chartjs-view (-> stats :image-content-types :data :follower)
              {:opts {:conf (chartjs-pie-conf :type :count {:depth3D 10})
                       :type :pie
                       :format {:count nil :type nil}
                       :settings {:height 400
                                   :title (loc "Image vs. Video in follower content")}}})

            (om/build table-view (-> stats :most-important-follower-images :data)
              {:opts {:settings {:title (loc "Highest Rated Sample")}
                       :columns [:low_resolution :created :comments :likes]
                       :format {:height 371
                                 :low_resolution {:type :image-url :vkey :id}
                                        ;:link {:type :url}
                                 :created {:type :isodatetime}}
                       :time (stat-time stats :most-important-follower-images)
                       :time-ranges [{:title (loc "Total") :key :total}
                                      {:title (loc "Yearly") :key :yearly}
                                      {:title (loc "Monthly") :key :monthly}
                                      {:title (loc "Weekly") :key :weekly}
                                      {:title (loc "Daily") :key :daily}
                                      ]}})))))

    om/IDidMount
    (did-mount [_]
      (bundle-calls :audience))))

;; ---------------------------------------------------------------
;; Main Application
;; ---------------------------------------------------------------

(defn main []
  (-> js/document .-location (set! "#/"))
  (om/root
    (fn [app owner]

      (reify
        om/IRender
        (render [_]
          (dom/div nil
            (dom/div #js {:id "wrapper" :style #js {:marginTop "0px"}}
              (dom/div #js {:id "page-content-wrapper"}
                (dom/div #js {:className "container-fluid"}
                  (om/build application-views app)
                  ))))
          )))
    app-state
    {:target (. js/document (getElementById "app"))}))

