// Compiled by ClojureScript 1.7.145 {:static-fns true, :optimize-constants true}
goog.provide('schema.core');
goog.require('cljs.core');
goog.require('clojure.string');
goog.require('schema.utils');

/**
 * @interface
 */
schema.core.Schema = function(){};

/**
 * Produce a function that takes [data], and either returns a walked version of data
 *   (by default, usually just data), or a utils/ErrorContainer containing value that looks
 *   like the 'bad' parts of data with ValidationErrors at the leaves describing the failures.
 * 
 *   If this is a composite schema, should let-bind (subschema-walker sub-schema) for each
 *   subschema outside the returned fn.  Within the returned fn, should break down data
 *   into constituents, call the let-bound subschema walkers on each component, and then
 *   reassemble the components into a walked version of the data (or an ErrorContainer
 *   describing the validaiton failures).
 * 
 *   Attempting to walk a value that already contains a utils/ErrorContainer produces undefined
 *   behavior.
 * 
 *   User code should never call `walker` directly.  Instead, it should call `start-walker`
 *   below.
 */
schema.core.walker = (function schema$core$walker(this$){
if((!((this$ == null))) && (!((this$.schema$core$Schema$walker$arity$1 == null)))){
return this$.schema$core$Schema$walker$arity$1(this$);
} else {
var x__5167__auto__ = (((this$ == null))?null:this$);
var m__5168__auto__ = (schema.core.walker[goog.typeOf(x__5167__auto__)]);
if(!((m__5168__auto__ == null))){
return (m__5168__auto__.cljs$core$IFn$_invoke$arity$1 ? m__5168__auto__.cljs$core$IFn$_invoke$arity$1(this$) : m__5168__auto__.call(null,this$));
} else {
var m__5168__auto____$1 = (schema.core.walker["_"]);
if(!((m__5168__auto____$1 == null))){
return (m__5168__auto____$1.cljs$core$IFn$_invoke$arity$1 ? m__5168__auto____$1.cljs$core$IFn$_invoke$arity$1(this$) : m__5168__auto____$1.call(null,this$));
} else {
throw cljs.core.missing_protocol("Schema.walker",this$);
}
}
}
});

/**
 * Expand this schema to a human-readable format suitable for pprinting,
 *   also expanding class schematas at the leaves.  Example:
 * 
 *   user> (s/explain {:a s/Keyword :b [s/Int]} )
 *   {:a Keyword, :b [Int]}
 */
schema.core.explain = (function schema$core$explain(this$){
if((!((this$ == null))) && (!((this$.schema$core$Schema$explain$arity$1 == null)))){
return this$.schema$core$Schema$explain$arity$1(this$);
} else {
var x__5167__auto__ = (((this$ == null))?null:this$);
var m__5168__auto__ = (schema.core.explain[goog.typeOf(x__5167__auto__)]);
if(!((m__5168__auto__ == null))){
return (m__5168__auto__.cljs$core$IFn$_invoke$arity$1 ? m__5168__auto__.cljs$core$IFn$_invoke$arity$1(this$) : m__5168__auto__.call(null,this$));
} else {
var m__5168__auto____$1 = (schema.core.explain["_"]);
if(!((m__5168__auto____$1 == null))){
return (m__5168__auto____$1.cljs$core$IFn$_invoke$arity$1 ? m__5168__auto____$1.cljs$core$IFn$_invoke$arity$1(this$) : m__5168__auto____$1.call(null,this$));
} else {
throw cljs.core.missing_protocol("Schema.explain",this$);
}
}
}
});

/**
 * The function to call within 'walker' implementations to create walkers for subschemas.
 * Can be dynamically bound (using start-walker below) to create different walking behaviors.
 * 
 * For the curious, implemented using dynamic binding rather than making walker take a
 * subschema-walker as an argument because some behaviors (e.g. recursive schema walkers)
 * seem to require mind-bending things like fixed-point combinators that way, but are
 * simple this way.
 */
schema.core.subschema_walker = (function schema$core$subschema_walker(s){
throw (new Error([cljs.core.str("Walking is unsupported outside of start-walker; "),cljs.core.str("all composite schemas must eagerly bind subschema-walkers "),cljs.core.str("outside the returned walker.")].join('')));
});
/**
 * The entry point for creating walkers.  Binds the provided walker to subschema-walker,
 * then calls it on the provided schema.  For simple validation, pass walker as sub-walker.
 * More sophisticated behavior (coercion, etc), can be achieved by passing a sub-walker
 * that wraps walker with additional behavior.
 */
schema.core.start_walker = (function schema$core$start_walker(sub_walker,schema__$1){
var subschema_walker27150 = schema.core.subschema_walker;
schema.core.subschema_walker = sub_walker;

try{return (schema.core.subschema_walker.cljs$core$IFn$_invoke$arity$1 ? schema.core.subschema_walker.cljs$core$IFn$_invoke$arity$1(schema__$1) : schema.core.subschema_walker.call(null,schema__$1));
}finally {schema.core.subschema_walker = subschema_walker27150;
}});
/**
 * Compile an efficient checker for schema, which returns nil for valid values and
 * error descriptions otherwise.
 */
schema.core.checker = (function schema$core$checker(schema__$1){
return cljs.core.comp.cljs$core$IFn$_invoke$arity$2(schema.utils.error_val,schema.core.start_walker(schema.utils.memoize_id(schema.core.walker),schema__$1));
});
/**
 * Return nil if x matches schema; otherwise, returns a value that looks like the
 * 'bad' parts of x with ValidationErrors at the leaves describing the failures.
 */
schema.core.check = (function schema$core$check(schema__$1,x){
return schema.core.checker(schema__$1).call(null,x);
});
/**
 * Throw an exception if value does not satisfy schema; otherwise, return value.
 */
schema.core.validate = (function schema$core$validate(schema__$1,value){
var temp__4425__auto___27151 = schema.core.check(schema__$1,value);
if(cljs.core.truth_(temp__4425__auto___27151)){
var error_27152 = temp__4425__auto___27151;
throw cljs.core.ex_info.cljs$core$IFn$_invoke$arity$2(schema.utils.format_STAR_.cljs$core$IFn$_invoke$arity$variadic("Value does not match schema: %s",cljs.core.array_seq([cljs.core.pr_str.cljs$core$IFn$_invoke$arity$variadic(cljs.core.array_seq([error_27152], 0))], 0)),new cljs.core.PersistentArrayMap(null, 4, [cljs.core.cst$kw$type,cljs.core.cst$kw$schema$core_SLASH_error,cljs.core.cst$kw$schema,schema__$1,cljs.core.cst$kw$value,value,cljs.core.cst$kw$error,error_27152], null));
} else {
}

return value;
});
(schema.core.Schema["function"] = true);

(schema.core.walker["function"] = (function (this$){
var class_walker = (function (){var temp__4423__auto__ = schema.utils.class_schema(this$);
if(cljs.core.truth_(temp__4423__auto__)){
var more_schema = temp__4423__auto__;
return (schema.core.subschema_walker.cljs$core$IFn$_invoke$arity$1 ? schema.core.subschema_walker.cljs$core$IFn$_invoke$arity$1(more_schema) : schema.core.subschema_walker.call(null,more_schema));
} else {
return cljs.core.identity;
}
})();
return ((function (class_walker){
return (function (x){
var or__4512__auto__ = ((((x == null)) || (cljs.core.not((function (){var or__4512__auto__ = (this$ === x.constructor);
if(or__4512__auto__){
return or__4512__auto__;
} else {
return x instanceof this$;
}
})())))?schema.utils.error(schema.utils.__GT_ValidationError(this$,x,(new cljs.core.Delay(((function (class_walker){
return (function (){
return cljs.core._conj(cljs.core._conj(cljs.core._conj(cljs.core.List.EMPTY,schema.utils.value_name(x)),this$),cljs.core.cst$sym$instance_QMARK_);
});})(class_walker))
,null)),null)):null);
if(cljs.core.truth_(or__4512__auto__)){
return or__4512__auto__;
} else {
return (class_walker.cljs$core$IFn$_invoke$arity$1 ? class_walker.cljs$core$IFn$_invoke$arity$1(x) : class_walker.call(null,x));
}
});
;})(class_walker))
}));

(schema.core.explain["function"] = (function (this$){
var temp__4423__auto__ = schema.utils.class_schema(this$);
if(cljs.core.truth_(temp__4423__auto__)){
var more_schema = temp__4423__auto__;
return schema.core.explain(more_schema);
} else {
return this$;
}
}));

/**
* @constructor
 * @implements {cljs.core.IRecord}
 * @implements {cljs.core.IEquiv}
 * @implements {cljs.core.IHash}
 * @implements {cljs.core.ICollection}
 * @implements {schema.core.Schema}
 * @implements {cljs.core.ICounted}
 * @implements {cljs.core.ISeqable}
 * @implements {cljs.core.IMeta}
 * @implements {cljs.core.ICloneable}
 * @implements {cljs.core.IPrintWithWriter}
 * @implements {cljs.core.IIterable}
 * @implements {cljs.core.IWithMeta}
 * @implements {cljs.core.IAssociative}
 * @implements {cljs.core.IMap}
 * @implements {cljs.core.ILookup}
*/
schema.core.AnythingSchema = (function (_,__meta,__extmap,__hash){
this._ = _;
this.__meta = __meta;
this.__extmap = __extmap;
this.__hash = __hash;
this.cljs$lang$protocol_mask$partition0$ = 2229667594;
this.cljs$lang$protocol_mask$partition1$ = 8192;
})
schema.core.AnythingSchema.prototype.cljs$core$ILookup$_lookup$arity$2 = (function (this__5126__auto__,k__5127__auto__){
var self__ = this;
var this__5126__auto____$1 = this;
return cljs.core._lookup.cljs$core$IFn$_invoke$arity$3(this__5126__auto____$1,k__5127__auto__,null);
});

schema.core.AnythingSchema.prototype.cljs$core$ILookup$_lookup$arity$3 = (function (this__5128__auto__,k27154,else__5129__auto__){
var self__ = this;
var this__5128__auto____$1 = this;
var G__27156 = (((k27154 instanceof cljs.core.Keyword))?k27154.fqn:null);
switch (G__27156) {
case "_":
return self__._;

break;
default:
return cljs.core.get.cljs$core$IFn$_invoke$arity$3(self__.__extmap,k27154,else__5129__auto__);

}
});

schema.core.AnythingSchema.prototype.cljs$core$IPrintWithWriter$_pr_writer$arity$3 = (function (this__5140__auto__,writer__5141__auto__,opts__5142__auto__){
var self__ = this;
var this__5140__auto____$1 = this;
var pr_pair__5143__auto__ = ((function (this__5140__auto____$1){
return (function (keyval__5144__auto__){
return cljs.core.pr_sequential_writer(writer__5141__auto__,cljs.core.pr_writer,""," ","",opts__5142__auto__,keyval__5144__auto__);
});})(this__5140__auto____$1))
;
return cljs.core.pr_sequential_writer(writer__5141__auto__,pr_pair__5143__auto__,"#schema.core.AnythingSchema{",", ","}",opts__5142__auto__,cljs.core.concat.cljs$core$IFn$_invoke$arity$2(new cljs.core.PersistentVector(null, 1, 5, cljs.core.PersistentVector.EMPTY_NODE, [(new cljs.core.PersistentVector(null,2,(5),cljs.core.PersistentVector.EMPTY_NODE,[cljs.core.cst$kw$_,self__._],null))], null),self__.__extmap));
});

schema.core.AnythingSchema.prototype.cljs$core$IIterable$ = true;

schema.core.AnythingSchema.prototype.cljs$core$IIterable$_iterator$arity$1 = (function (G__27153){
var self__ = this;
var G__27153__$1 = this;
return (new cljs.core.RecordIter((0),G__27153__$1,1,new cljs.core.PersistentVector(null, 1, 5, cljs.core.PersistentVector.EMPTY_NODE, [cljs.core.cst$kw$_], null),cljs.core._iterator(self__.__extmap)));
});

schema.core.AnythingSchema.prototype.cljs$core$IMeta$_meta$arity$1 = (function (this__5124__auto__){
var self__ = this;
var this__5124__auto____$1 = this;
return self__.__meta;
});

schema.core.AnythingSchema.prototype.cljs$core$ICloneable$_clone$arity$1 = (function (this__5120__auto__){
var self__ = this;
var this__5120__auto____$1 = this;
return (new schema.core.AnythingSchema(self__._,self__.__meta,self__.__extmap,self__.__hash));
});

schema.core.AnythingSchema.prototype.cljs$core$ICounted$_count$arity$1 = (function (this__5130__auto__){
var self__ = this;
var this__5130__auto____$1 = this;
return (1 + cljs.core.count(self__.__extmap));
});

schema.core.AnythingSchema.prototype.cljs$core$IHash$_hash$arity$1 = (function (this__5121__auto__){
var self__ = this;
var this__5121__auto____$1 = this;
var h__4947__auto__ = self__.__hash;
if(!((h__4947__auto__ == null))){
return h__4947__auto__;
} else {
var h__4947__auto____$1 = cljs.core.hash_imap(this__5121__auto____$1);
self__.__hash = h__4947__auto____$1;

return h__4947__auto____$1;
}
});

schema.core.AnythingSchema.prototype.cljs$core$IEquiv$_equiv$arity$2 = (function (this__5122__auto__,other__5123__auto__){
var self__ = this;
var this__5122__auto____$1 = this;
if(cljs.core.truth_((function (){var and__4500__auto__ = other__5123__auto__;
if(cljs.core.truth_(and__4500__auto__)){
var and__4500__auto____$1 = (this__5122__auto____$1.constructor === other__5123__auto__.constructor);
if(and__4500__auto____$1){
return cljs.core.equiv_map(this__5122__auto____$1,other__5123__auto__);
} else {
return and__4500__auto____$1;
}
} else {
return and__4500__auto__;
}
})())){
return true;
} else {
return false;
}
});

schema.core.AnythingSchema.prototype.cljs$core$IMap$_dissoc$arity$2 = (function (this__5135__auto__,k__5136__auto__){
var self__ = this;
var this__5135__auto____$1 = this;
if(cljs.core.contains_QMARK_(new cljs.core.PersistentHashSet(null, new cljs.core.PersistentArrayMap(null, 1, [cljs.core.cst$kw$_,null], null), null),k__5136__auto__)){
return cljs.core.dissoc.cljs$core$IFn$_invoke$arity$2(cljs.core.with_meta(cljs.core.into.cljs$core$IFn$_invoke$arity$2(cljs.core.PersistentArrayMap.EMPTY,this__5135__auto____$1),self__.__meta),k__5136__auto__);
} else {
return (new schema.core.AnythingSchema(self__._,self__.__meta,cljs.core.not_empty(cljs.core.dissoc.cljs$core$IFn$_invoke$arity$2(self__.__extmap,k__5136__auto__)),null));
}
});

schema.core.AnythingSchema.prototype.cljs$core$IAssociative$_assoc$arity$3 = (function (this__5133__auto__,k__5134__auto__,G__27153){
var self__ = this;
var this__5133__auto____$1 = this;
var pred__27157 = cljs.core.keyword_identical_QMARK_;
var expr__27158 = k__5134__auto__;
if(cljs.core.truth_((pred__27157.cljs$core$IFn$_invoke$arity$2 ? pred__27157.cljs$core$IFn$_invoke$arity$2(cljs.core.cst$kw$_,expr__27158) : pred__27157.call(null,cljs.core.cst$kw$_,expr__27158)))){
return (new schema.core.AnythingSchema(G__27153,self__.__meta,self__.__extmap,null));
} else {
return (new schema.core.AnythingSchema(self__._,self__.__meta,cljs.core.assoc.cljs$core$IFn$_invoke$arity$3(self__.__extmap,k__5134__auto__,G__27153),null));
}
});

schema.core.AnythingSchema.prototype.cljs$core$ISeqable$_seq$arity$1 = (function (this__5138__auto__){
var self__ = this;
var this__5138__auto____$1 = this;
return cljs.core.seq(cljs.core.concat.cljs$core$IFn$_invoke$arity$2(new cljs.core.PersistentVector(null, 1, 5, cljs.core.PersistentVector.EMPTY_NODE, [(new cljs.core.PersistentVector(null,2,(5),cljs.core.PersistentVector.EMPTY_NODE,[cljs.core.cst$kw$_,self__._],null))], null),self__.__extmap));
});

schema.core.AnythingSchema.prototype.cljs$core$IWithMeta$_with_meta$arity$2 = (function (this__5125__auto__,G__27153){
var self__ = this;
var this__5125__auto____$1 = this;
return (new schema.core.AnythingSchema(self__._,G__27153,self__.__extmap,self__.__hash));
});

schema.core.AnythingSchema.prototype.cljs$core$ICollection$_conj$arity$2 = (function (this__5131__auto__,entry__5132__auto__){
var self__ = this;
var this__5131__auto____$1 = this;
if(cljs.core.vector_QMARK_(entry__5132__auto__)){
return cljs.core._assoc(this__5131__auto____$1,cljs.core._nth.cljs$core$IFn$_invoke$arity$2(entry__5132__auto__,(0)),cljs.core._nth.cljs$core$IFn$_invoke$arity$2(entry__5132__auto__,(1)));
} else {
return cljs.core.reduce.cljs$core$IFn$_invoke$arity$3(cljs.core._conj,this__5131__auto____$1,entry__5132__auto__);
}
});

schema.core.AnythingSchema.prototype.schema$core$Schema$ = true;

schema.core.AnythingSchema.prototype.schema$core$Schema$walker$arity$1 = (function (this$){
var self__ = this;
var this$__$1 = this;
return cljs.core.identity;
});

schema.core.AnythingSchema.prototype.schema$core$Schema$explain$arity$1 = (function (this$){
var self__ = this;
var this$__$1 = this;
return cljs.core.cst$sym$Any;
});

schema.core.AnythingSchema.getBasis = (function (){
return new cljs.core.PersistentVector(null, 1, 5, cljs.core.PersistentVector.EMPTY_NODE, [cljs.core.cst$sym$_], null);
});

schema.core.AnythingSchema.cljs$lang$type = true;

schema.core.AnythingSchema.cljs$lang$ctorPrSeq = (function (this__5160__auto__){
return cljs.core._conj(cljs.core.List.EMPTY,"schema.core/AnythingSchema");
});

schema.core.AnythingSchema.cljs$lang$ctorPrWriter = (function (this__5160__auto__,writer__5161__auto__){
return cljs.core._write(writer__5161__auto__,"schema.core/AnythingSchema");
});

schema.core.__GT_AnythingSchema = (function schema$core$__GT_AnythingSchema(_){
return (new schema.core.AnythingSchema(_,null,null,null));
});

schema.core.map__GT_AnythingSchema = (function schema$core$map__GT_AnythingSchema(G__27155){
return (new schema.core.AnythingSchema(cljs.core.cst$kw$_.cljs$core$IFn$_invoke$arity$1(G__27155),null,cljs.core.dissoc.cljs$core$IFn$_invoke$arity$2(G__27155,cljs.core.cst$kw$_),null));
});

/**
 * Any value, including nil.
 */
schema.core.Any = (new schema.core.AnythingSchema(null,null,null,null));

/**
* @constructor
 * @implements {cljs.core.IRecord}
 * @implements {cljs.core.IEquiv}
 * @implements {cljs.core.IHash}
 * @implements {cljs.core.ICollection}
 * @implements {schema.core.Schema}
 * @implements {cljs.core.ICounted}
 * @implements {cljs.core.ISeqable}
 * @implements {cljs.core.IMeta}
 * @implements {cljs.core.ICloneable}
 * @implements {cljs.core.IPrintWithWriter}
 * @implements {cljs.core.IIterable}
 * @implements {cljs.core.IWithMeta}
 * @implements {cljs.core.IAssociative}
 * @implements {cljs.core.IMap}
 * @implements {cljs.core.ILookup}
*/
schema.core.EqSchema = (function (v,__meta,__extmap,__hash){
this.v = v;
this.__meta = __meta;
this.__extmap = __extmap;
this.__hash = __hash;
this.cljs$lang$protocol_mask$partition0$ = 2229667594;
this.cljs$lang$protocol_mask$partition1$ = 8192;
})
schema.core.EqSchema.prototype.cljs$core$ILookup$_lookup$arity$2 = (function (this__5126__auto__,k__5127__auto__){
var self__ = this;
var this__5126__auto____$1 = this;
return cljs.core._lookup.cljs$core$IFn$_invoke$arity$3(this__5126__auto____$1,k__5127__auto__,null);
});

schema.core.EqSchema.prototype.cljs$core$ILookup$_lookup$arity$3 = (function (this__5128__auto__,k27162,else__5129__auto__){
var self__ = this;
var this__5128__auto____$1 = this;
var G__27164 = (((k27162 instanceof cljs.core.Keyword))?k27162.fqn:null);
switch (G__27164) {
case "v":
return self__.v;

break;
default:
return cljs.core.get.cljs$core$IFn$_invoke$arity$3(self__.__extmap,k27162,else__5129__auto__);

}
});

schema.core.EqSchema.prototype.cljs$core$IPrintWithWriter$_pr_writer$arity$3 = (function (this__5140__auto__,writer__5141__auto__,opts__5142__auto__){
var self__ = this;
var this__5140__auto____$1 = this;
var pr_pair__5143__auto__ = ((function (this__5140__auto____$1){
return (function (keyval__5144__auto__){
return cljs.core.pr_sequential_writer(writer__5141__auto__,cljs.core.pr_writer,""," ","",opts__5142__auto__,keyval__5144__auto__);
});})(this__5140__auto____$1))
;
return cljs.core.pr_sequential_writer(writer__5141__auto__,pr_pair__5143__auto__,"#schema.core.EqSchema{",", ","}",opts__5142__auto__,cljs.core.concat.cljs$core$IFn$_invoke$arity$2(new cljs.core.PersistentVector(null, 1, 5, cljs.core.PersistentVector.EMPTY_NODE, [(new cljs.core.PersistentVector(null,2,(5),cljs.core.PersistentVector.EMPTY_NODE,[cljs.core.cst$kw$v,self__.v],null))], null),self__.__extmap));
});

schema.core.EqSchema.prototype.cljs$core$IIterable$ = true;

schema.core.EqSchema.prototype.cljs$core$IIterable$_iterator$arity$1 = (function (G__27161){
var self__ = this;
var G__27161__$1 = this;
return (new cljs.core.RecordIter((0),G__27161__$1,1,new cljs.core.PersistentVector(null, 1, 5, cljs.core.PersistentVector.EMPTY_NODE, [cljs.core.cst$kw$v], null),cljs.core._iterator(self__.__extmap)));
});

schema.core.EqSchema.prototype.cljs$core$IMeta$_meta$arity$1 = (function (this__5124__auto__){
var self__ = this;
var this__5124__auto____$1 = this;
return self__.__meta;
});

schema.core.EqSchema.prototype.cljs$core$ICloneable$_clone$arity$1 = (function (this__5120__auto__){
var self__ = this;
var this__5120__auto____$1 = this;
return (new schema.core.EqSchema(self__.v,self__.__meta,self__.__extmap,self__.__hash));
});

schema.core.EqSchema.prototype.cljs$core$ICounted$_count$arity$1 = (function (this__5130__auto__){
var self__ = this;
var this__5130__auto____$1 = this;
return (1 + cljs.core.count(self__.__extmap));
});

schema.core.EqSchema.prototype.cljs$core$IHash$_hash$arity$1 = (function (this__5121__auto__){
var self__ = this;
var this__5121__auto____$1 = this;
var h__4947__auto__ = self__.__hash;
if(!((h__4947__auto__ == null))){
return h__4947__auto__;
} else {
var h__4947__auto____$1 = cljs.core.hash_imap(this__5121__auto____$1);
self__.__hash = h__4947__auto____$1;

return h__4947__auto____$1;
}
});

schema.core.EqSchema.prototype.cljs$core$IEquiv$_equiv$arity$2 = (function (this__5122__auto__,other__5123__auto__){
var self__ = this;
var this__5122__auto____$1 = this;
if(cljs.core.truth_((function (){var and__4500__auto__ = other__5123__auto__;
if(cljs.core.truth_(and__4500__auto__)){
var and__4500__auto____$1 = (this__5122__auto____$1.constructor === other__5123__auto__.constructor);
if(and__4500__auto____$1){
return cljs.core.equiv_map(this__5122__auto____$1,other__5123__auto__);
} else {
return and__4500__auto____$1;
}
} else {
return and__4500__auto__;
}
})())){
return true;
} else {
return false;
}
});

schema.core.EqSchema.prototype.cljs$core$IMap$_dissoc$arity$2 = (function (this__5135__auto__,k__5136__auto__){
var self__ = this;
var this__5135__auto____$1 = this;
if(cljs.core.contains_QMARK_(new cljs.core.PersistentHashSet(null, new cljs.core.PersistentArrayMap(null, 1, [cljs.core.cst$kw$v,null], null), null),k__5136__auto__)){
return cljs.core.dissoc.cljs$core$IFn$_invoke$arity$2(cljs.core.with_meta(cljs.core.into.cljs$core$IFn$_invoke$arity$2(cljs.core.PersistentArrayMap.EMPTY,this__5135__auto____$1),self__.__meta),k__5136__auto__);
} else {
return (new schema.core.EqSchema(self__.v,self__.__meta,cljs.core.not_empty(cljs.core.dissoc.cljs$core$IFn$_invoke$arity$2(self__.__extmap,k__5136__auto__)),null));
}
});

schema.core.EqSchema.prototype.cljs$core$IAssociative$_assoc$arity$3 = (function (this__5133__auto__,k__5134__auto__,G__27161){
var self__ = this;
var this__5133__auto____$1 = this;
var pred__27165 = cljs.core.keyword_identical_QMARK_;
var expr__27166 = k__5134__auto__;
if(cljs.core.truth_((pred__27165.cljs$core$IFn$_invoke$arity$2 ? pred__27165.cljs$core$IFn$_invoke$arity$2(cljs.core.cst$kw$v,expr__27166) : pred__27165.call(null,cljs.core.cst$kw$v,expr__27166)))){
return (new schema.core.EqSchema(G__27161,self__.__meta,self__.__extmap,null));
} else {
return (new schema.core.EqSchema(self__.v,self__.__meta,cljs.core.assoc.cljs$core$IFn$_invoke$arity$3(self__.__extmap,k__5134__auto__,G__27161),null));
}
});

schema.core.EqSchema.prototype.cljs$core$ISeqable$_seq$arity$1 = (function (this__5138__auto__){
var self__ = this;
var this__5138__auto____$1 = this;
return cljs.core.seq(cljs.core.concat.cljs$core$IFn$_invoke$arity$2(new cljs.core.PersistentVector(null, 1, 5, cljs.core.PersistentVector.EMPTY_NODE, [(new cljs.core.PersistentVector(null,2,(5),cljs.core.PersistentVector.EMPTY_NODE,[cljs.core.cst$kw$v,self__.v],null))], null),self__.__extmap));
});

schema.core.EqSchema.prototype.cljs$core$IWithMeta$_with_meta$arity$2 = (function (this__5125__auto__,G__27161){
var self__ = this;
var this__5125__auto____$1 = this;
return (new schema.core.EqSchema(self__.v,G__27161,self__.__extmap,self__.__hash));
});

schema.core.EqSchema.prototype.cljs$core$ICollection$_conj$arity$2 = (function (this__5131__auto__,entry__5132__auto__){
var self__ = this;
var this__5131__auto____$1 = this;
if(cljs.core.vector_QMARK_(entry__5132__auto__)){
return cljs.core._assoc(this__5131__auto____$1,cljs.core._nth.cljs$core$IFn$_invoke$arity$2(entry__5132__auto__,(0)),cljs.core._nth.cljs$core$IFn$_invoke$arity$2(entry__5132__auto__,(1)));
} else {
return cljs.core.reduce.cljs$core$IFn$_invoke$arity$3(cljs.core._conj,this__5131__auto____$1,entry__5132__auto__);
}
});

schema.core.EqSchema.prototype.schema$core$Schema$ = true;

schema.core.EqSchema.prototype.schema$core$Schema$walker$arity$1 = (function (this$){
var self__ = this;
var this$__$1 = this;
return ((function (this$__$1){
return (function (x){
if(cljs.core._EQ_.cljs$core$IFn$_invoke$arity$2(self__.v,x)){
return x;
} else {
return schema.utils.error(schema.utils.__GT_ValidationError(this$__$1,x,(new cljs.core.Delay(((function (this$__$1){
return (function (){
return cljs.core._conj(cljs.core._conj(cljs.core._conj(cljs.core.List.EMPTY,schema.utils.value_name(x)),self__.v),cljs.core.cst$sym$_EQ_);
});})(this$__$1))
,null)),null));
}
});
;})(this$__$1))
});

schema.core.EqSchema.prototype.schema$core$Schema$explain$arity$1 = (function (this$){
var self__ = this;
var this$__$1 = this;
return cljs.core._conj(cljs.core._conj(cljs.core.List.EMPTY,self__.v),cljs.core.cst$sym$eq);
});

schema.core.EqSchema.getBasis = (function (){
return new cljs.core.PersistentVector(null, 1, 5, cljs.core.PersistentVector.EMPTY_NODE, [cljs.core.cst$sym$v], null);
});

schema.core.EqSchema.cljs$lang$type = true;

schema.core.EqSchema.cljs$lang$ctorPrSeq = (function (this__5160__auto__){
return cljs.core._conj(cljs.core.List.EMPTY,"schema.core/EqSchema");
});

schema.core.EqSchema.cljs$lang$ctorPrWriter = (function (this__5160__auto__,writer__5161__auto__){
return cljs.core._write(writer__5161__auto__,"schema.core/EqSchema");
});

schema.core.__GT_EqSchema = (function schema$core$__GT_EqSchema(v){
return (new schema.core.EqSchema(v,null,null,null));
});

schema.core.map__GT_EqSchema = (function schema$core$map__GT_EqSchema(G__27163){
return (new schema.core.EqSchema(cljs.core.cst$kw$v.cljs$core$IFn$_invoke$arity$1(G__27163),null,cljs.core.dissoc.cljs$core$IFn$_invoke$arity$2(G__27163,cljs.core.cst$kw$v),null));
});

/**
 * A value that must be (= v).
 */
schema.core.eq = (function schema$core$eq(v){
return (new schema.core.EqSchema(v,null,null,null));
});

/**
* @constructor
 * @implements {cljs.core.IRecord}
 * @implements {cljs.core.IEquiv}
 * @implements {cljs.core.IHash}
 * @implements {cljs.core.ICollection}
 * @implements {schema.core.Schema}
 * @implements {cljs.core.ICounted}
 * @implements {cljs.core.ISeqable}
 * @implements {cljs.core.IMeta}
 * @implements {cljs.core.ICloneable}
 * @implements {cljs.core.IPrintWithWriter}
 * @implements {cljs.core.IIterable}
 * @implements {cljs.core.IWithMeta}
 * @implements {cljs.core.IAssociative}
 * @implements {cljs.core.IMap}
 * @implements {cljs.core.ILookup}
*/
schema.core.Isa = (function (h,parent,__meta,__extmap,__hash){
this.h = h;
this.parent = parent;
this.__meta = __meta;
this.__extmap = __extmap;
this.__hash = __hash;
this.cljs$lang$protocol_mask$partition0$ = 2229667594;
this.cljs$lang$protocol_mask$partition1$ = 8192;
})
schema.core.Isa.prototype.cljs$core$ILookup$_lookup$arity$2 = (function (this__5126__auto__,k__5127__auto__){
var self__ = this;
var this__5126__auto____$1 = this;
return cljs.core._lookup.cljs$core$IFn$_invoke$arity$3(this__5126__auto____$1,k__5127__auto__,null);
});

schema.core.Isa.prototype.cljs$core$ILookup$_lookup$arity$3 = (function (this__5128__auto__,k27170,else__5129__auto__){
var self__ = this;
var this__5128__auto____$1 = this;
var G__27172 = (((k27170 instanceof cljs.core.Keyword))?k27170.fqn:null);
switch (G__27172) {
case "h":
return self__.h;

break;
case "parent":
return self__.parent;

break;
default:
return cljs.core.get.cljs$core$IFn$_invoke$arity$3(self__.__extmap,k27170,else__5129__auto__);

}
});

schema.core.Isa.prototype.cljs$core$IPrintWithWriter$_pr_writer$arity$3 = (function (this__5140__auto__,writer__5141__auto__,opts__5142__auto__){
var self__ = this;
var this__5140__auto____$1 = this;
var pr_pair__5143__auto__ = ((function (this__5140__auto____$1){
return (function (keyval__5144__auto__){
return cljs.core.pr_sequential_writer(writer__5141__auto__,cljs.core.pr_writer,""," ","",opts__5142__auto__,keyval__5144__auto__);
});})(this__5140__auto____$1))
;
return cljs.core.pr_sequential_writer(writer__5141__auto__,pr_pair__5143__auto__,"#schema.core.Isa{",", ","}",opts__5142__auto__,cljs.core.concat.cljs$core$IFn$_invoke$arity$2(new cljs.core.PersistentVector(null, 2, 5, cljs.core.PersistentVector.EMPTY_NODE, [(new cljs.core.PersistentVector(null,2,(5),cljs.core.PersistentVector.EMPTY_NODE,[cljs.core.cst$kw$h,self__.h],null)),(new cljs.core.PersistentVector(null,2,(5),cljs.core.PersistentVector.EMPTY_NODE,[cljs.core.cst$kw$parent,self__.parent],null))], null),self__.__extmap));
});

schema.core.Isa.prototype.cljs$core$IIterable$ = true;

schema.core.Isa.prototype.cljs$core$IIterable$_iterator$arity$1 = (function (G__27169){
var self__ = this;
var G__27169__$1 = this;
return (new cljs.core.RecordIter((0),G__27169__$1,2,new cljs.core.PersistentVector(null, 2, 5, cljs.core.PersistentVector.EMPTY_NODE, [cljs.core.cst$kw$h,cljs.core.cst$kw$parent], null),cljs.core._iterator(self__.__extmap)));
});

schema.core.Isa.prototype.cljs$core$IMeta$_meta$arity$1 = (function (this__5124__auto__){
var self__ = this;
var this__5124__auto____$1 = this;
return self__.__meta;
});

schema.core.Isa.prototype.cljs$core$ICloneable$_clone$arity$1 = (function (this__5120__auto__){
var self__ = this;
var this__5120__auto____$1 = this;
return (new schema.core.Isa(self__.h,self__.parent,self__.__meta,self__.__extmap,self__.__hash));
});

schema.core.Isa.prototype.cljs$core$ICounted$_count$arity$1 = (function (this__5130__auto__){
var self__ = this;
var this__5130__auto____$1 = this;
return (2 + cljs.core.count(self__.__extmap));
});

schema.core.Isa.prototype.cljs$core$IHash$_hash$arity$1 = (function (this__5121__auto__){
var self__ = this;
var this__5121__auto____$1 = this;
var h__4947__auto__ = self__.__hash;
if(!((h__4947__auto__ == null))){
return h__4947__auto__;
} else {
var h__4947__auto____$1 = cljs.core.hash_imap(this__5121__auto____$1);
self__.__hash = h__4947__auto____$1;

return h__4947__auto____$1;
}
});

schema.core.Isa.prototype.cljs$core$IEquiv$_equiv$arity$2 = (function (this__5122__auto__,other__5123__auto__){
var self__ = this;
var this__5122__auto____$1 = this;
if(cljs.core.truth_((function (){var and__4500__auto__ = other__5123__auto__;
if(cljs.core.truth_(and__4500__auto__)){
var and__4500__auto____$1 = (this__5122__auto____$1.constructor === other__5123__auto__.constructor);
if(and__4500__auto____$1){
return cljs.core.equiv_map(this__5122__auto____$1,other__5123__auto__);
} else {
return and__4500__auto____$1;
}
} else {
return and__4500__auto__;
}
})())){
return true;
} else {
return false;
}
});

schema.core.Isa.prototype.cljs$core$IMap$_dissoc$arity$2 = (function (this__5135__auto__,k__5136__auto__){
var self__ = this;
var this__5135__auto____$1 = this;
if(cljs.core.contains_QMARK_(new cljs.core.PersistentHashSet(null, new cljs.core.PersistentArrayMap(null, 2, [cljs.core.cst$kw$parent,null,cljs.core.cst$kw$h,null], null), null),k__5136__auto__)){
return cljs.core.dissoc.cljs$core$IFn$_invoke$arity$2(cljs.core.with_meta(cljs.core.into.cljs$core$IFn$_invoke$arity$2(cljs.core.PersistentArrayMap.EMPTY,this__5135__auto____$1),self__.__meta),k__5136__auto__);
} else {
return (new schema.core.Isa(self__.h,self__.parent,self__.__meta,cljs.core.not_empty(cljs.core.dissoc.cljs$core$IFn$_invoke$arity$2(self__.__extmap,k__5136__auto__)),null));
}
});

schema.core.Isa.prototype.cljs$core$IAssociative$_assoc$arity$3 = (function (this__5133__auto__,k__5134__auto__,G__27169){
var self__ = this;
var this__5133__auto____$1 = this;
var pred__27173 = cljs.core.keyword_identical_QMARK_;
var expr__27174 = k__5134__auto__;
if(cljs.core.truth_((pred__27173.cljs$core$IFn$_invoke$arity$2 ? pred__27173.cljs$core$IFn$_invoke$arity$2(cljs.core.cst$kw$h,expr__27174) : pred__27173.call(null,cljs.core.cst$kw$h,expr__27174)))){
return (new schema.core.Isa(G__27169,self__.parent,self__.__meta,self__.__extmap,null));
} else {
if(cljs.core.truth_((pred__27173.cljs$core$IFn$_invoke$arity$2 ? pred__27173.cljs$core$IFn$_invoke$arity$2(cljs.core.cst$kw$parent,expr__27174) : pred__27173.call(null,cljs.core.cst$kw$parent,expr__27174)))){
return (new schema.core.Isa(self__.h,G__27169,self__.__meta,self__.__extmap,null));
} else {
return (new schema.core.Isa(self__.h,self__.parent,self__.__meta,cljs.core.assoc.cljs$core$IFn$_invoke$arity$3(self__.__extmap,k__5134__auto__,G__27169),null));
}
}
});

schema.core.Isa.prototype.cljs$core$ISeqable$_seq$arity$1 = (function (this__5138__auto__){
var self__ = this;
var this__5138__auto____$1 = this;
return cljs.core.seq(cljs.core.concat.cljs$core$IFn$_invoke$arity$2(new cljs.core.PersistentVector(null, 2, 5, cljs.core.PersistentVector.EMPTY_NODE, [(new cljs.core.PersistentVector(null,2,(5),cljs.core.PersistentVector.EMPTY_NODE,[cljs.core.cst$kw$h,self__.h],null)),(new cljs.core.PersistentVector(null,2,(5),cljs.core.PersistentVector.EMPTY_NODE,[cljs.core.cst$kw$parent,self__.parent],null))], null),self__.__extmap));
});

schema.core.Isa.prototype.cljs$core$IWithMeta$_with_meta$arity$2 = (function (this__5125__auto__,G__27169){
var self__ = this;
var this__5125__auto____$1 = this;
return (new schema.core.Isa(self__.h,self__.parent,G__27169,self__.__extmap,self__.__hash));
});

schema.core.Isa.prototype.cljs$core$ICollection$_conj$arity$2 = (function (this__5131__auto__,entry__5132__auto__){
var self__ = this;
var this__5131__auto____$1 = this;
if(cljs.core.vector_QMARK_(entry__5132__auto__)){
return cljs.core._assoc(this__5131__auto____$1,cljs.core._nth.cljs$core$IFn$_invoke$arity$2(entry__5132__auto__,(0)),cljs.core._nth.cljs$core$IFn$_invoke$arity$2(entry__5132__auto__,(1)));
} else {
return cljs.core.reduce.cljs$core$IFn$_invoke$arity$3(cljs.core._conj,this__5131__auto____$1,entry__5132__auto__);
}
});

schema.core.Isa.prototype.schema$core$Schema$ = true;

schema.core.Isa.prototype.schema$core$Schema$walker$arity$1 = (function (this$){
var self__ = this;
var this$__$1 = this;
return ((function (this$__$1){
return (function (child){
if(cljs.core.truth_((function (){var or__4512__auto__ = (function (){var and__4500__auto__ = self__.h;
if(cljs.core.truth_(and__4500__auto__)){
return cljs.core.isa_QMARK_.cljs$core$IFn$_invoke$arity$3(self__.h,child,self__.parent);
} else {
return and__4500__auto__;
}
})();
if(cljs.core.truth_(or__4512__auto__)){
return or__4512__auto__;
} else {
return cljs.core.isa_QMARK_.cljs$core$IFn$_invoke$arity$2(child,self__.parent);
}
})())){
return child;
} else {
return schema.utils.error(schema.utils.__GT_ValidationError(this$__$1,child,(new cljs.core.Delay(((function (this$__$1){
return (function (){
return cljs.core._conj(cljs.core._conj(cljs.core._conj(cljs.core.List.EMPTY,self__.parent),child),cljs.core.cst$sym$isa_QMARK_);
});})(this$__$1))
,null)),null));
}
});
;})(this$__$1))
});

schema.core.Isa.prototype.schema$core$Schema$explain$arity$1 = (function (this$){
var self__ = this;
var this$__$1 = this;
return cljs.core._conj(cljs.core._conj(cljs.core.List.EMPTY,self__.parent),cljs.core.cst$sym$isa_QMARK_);
});

schema.core.Isa.getBasis = (function (){
return new cljs.core.PersistentVector(null, 2, 5, cljs.core.PersistentVector.EMPTY_NODE, [cljs.core.cst$sym$h,cljs.core.cst$sym$parent], null);
});

schema.core.Isa.cljs$lang$type = true;

schema.core.Isa.cljs$lang$ctorPrSeq = (function (this__5160__auto__){
return cljs.core._conj(cljs.core.List.EMPTY,"schema.core/Isa");
});

schema.core.Isa.cljs$lang$ctorPrWriter = (function (this__5160__auto__,writer__5161__auto__){
return cljs.core._write(writer__5161__auto__,"schema.core/Isa");
});

schema.core.__GT_Isa = (function schema$core$__GT_Isa(h,parent){
return (new schema.core.Isa(h,parent,null,null,null));
});

schema.core.map__GT_Isa = (function schema$core$map__GT_Isa(G__27171){
return (new schema.core.Isa(cljs.core.cst$kw$h.cljs$core$IFn$_invoke$arity$1(G__27171),cljs.core.cst$kw$parent.cljs$core$IFn$_invoke$arity$1(G__27171),null,cljs.core.dissoc.cljs$core$IFn$_invoke$arity$variadic(G__27171,cljs.core.cst$kw$h,cljs.core.array_seq([cljs.core.cst$kw$parent], 0)),null));
});

/**
 * A value that must be a child of parent.
 */
schema.core.isa = (function schema$core$isa(var_args){
var args27177 = [];
var len__5570__auto___27180 = arguments.length;
var i__5571__auto___27181 = (0);
while(true){
if((i__5571__auto___27181 < len__5570__auto___27180)){
args27177.push((arguments[i__5571__auto___27181]));

var G__27182 = (i__5571__auto___27181 + (1));
i__5571__auto___27181 = G__27182;
continue;
} else {
}
break;
}

var G__27179 = args27177.length;
switch (G__27179) {
case 1:
return schema.core.isa.cljs$core$IFn$_invoke$arity$1((arguments[(0)]));

break;
case 2:
return schema.core.isa.cljs$core$IFn$_invoke$arity$2((arguments[(0)]),(arguments[(1)]));

break;
default:
throw (new Error([cljs.core.str("Invalid arity: "),cljs.core.str(args27177.length)].join('')));

}
});

schema.core.isa.cljs$core$IFn$_invoke$arity$1 = (function (parent){
return (new schema.core.Isa(null,parent,null,null,null));
});

schema.core.isa.cljs$core$IFn$_invoke$arity$2 = (function (h,parent){
return (new schema.core.Isa(h,parent,null,null,null));
});

schema.core.isa.cljs$lang$maxFixedArity = 2;

/**
* @constructor
 * @implements {cljs.core.IRecord}
 * @implements {cljs.core.IEquiv}
 * @implements {cljs.core.IHash}
 * @implements {cljs.core.ICollection}
 * @implements {schema.core.Schema}
 * @implements {cljs.core.ICounted}
 * @implements {cljs.core.ISeqable}
 * @implements {cljs.core.IMeta}
 * @implements {cljs.core.ICloneable}
 * @implements {cljs.core.IPrintWithWriter}
 * @implements {cljs.core.IIterable}
 * @implements {cljs.core.IWithMeta}
 * @implements {cljs.core.IAssociative}
 * @implements {cljs.core.IMap}
 * @implements {cljs.core.ILookup}
*/
schema.core.EnumSchema = (function (vs,__meta,__extmap,__hash){
this.vs = vs;
this.__meta = __meta;
this.__extmap = __extmap;
this.__hash = __hash;
this.cljs$lang$protocol_mask$partition0$ = 2229667594;
this.cljs$lang$protocol_mask$partition1$ = 8192;
})
schema.core.EnumSchema.prototype.cljs$core$ILookup$_lookup$arity$2 = (function (this__5126__auto__,k__5127__auto__){
var self__ = this;
var this__5126__auto____$1 = this;
return cljs.core._lookup.cljs$core$IFn$_invoke$arity$3(this__5126__auto____$1,k__5127__auto__,null);
});

schema.core.EnumSchema.prototype.cljs$core$ILookup$_lookup$arity$3 = (function (this__5128__auto__,k27185,else__5129__auto__){
var self__ = this;
var this__5128__auto____$1 = this;
var G__27187 = (((k27185 instanceof cljs.core.Keyword))?k27185.fqn:null);
switch (G__27187) {
case "vs":
return self__.vs;

break;
default:
return cljs.core.get.cljs$core$IFn$_invoke$arity$3(self__.__extmap,k27185,else__5129__auto__);

}
});

schema.core.EnumSchema.prototype.cljs$core$IPrintWithWriter$_pr_writer$arity$3 = (function (this__5140__auto__,writer__5141__auto__,opts__5142__auto__){
var self__ = this;
var this__5140__auto____$1 = this;
var pr_pair__5143__auto__ = ((function (this__5140__auto____$1){
return (function (keyval__5144__auto__){
return cljs.core.pr_sequential_writer(writer__5141__auto__,cljs.core.pr_writer,""," ","",opts__5142__auto__,keyval__5144__auto__);
});})(this__5140__auto____$1))
;
return cljs.core.pr_sequential_writer(writer__5141__auto__,pr_pair__5143__auto__,"#schema.core.EnumSchema{",", ","}",opts__5142__auto__,cljs.core.concat.cljs$core$IFn$_invoke$arity$2(new cljs.core.PersistentVector(null, 1, 5, cljs.core.PersistentVector.EMPTY_NODE, [(new cljs.core.PersistentVector(null,2,(5),cljs.core.PersistentVector.EMPTY_NODE,[cljs.core.cst$kw$vs,self__.vs],null))], null),self__.__extmap));
});

schema.core.EnumSchema.prototype.cljs$core$IIterable$ = true;

schema.core.EnumSchema.prototype.cljs$core$IIterable$_iterator$arity$1 = (function (G__27184){
var self__ = this;
var G__27184__$1 = this;
return (new cljs.core.RecordIter((0),G__27184__$1,1,new cljs.core.PersistentVector(null, 1, 5, cljs.core.PersistentVector.EMPTY_NODE, [cljs.core.cst$kw$vs], null),cljs.core._iterator(self__.__extmap)));
});

schema.core.EnumSchema.prototype.cljs$core$IMeta$_meta$arity$1 = (function (this__5124__auto__){
var self__ = this;
var this__5124__auto____$1 = this;
return self__.__meta;
});

schema.core.EnumSchema.prototype.cljs$core$ICloneable$_clone$arity$1 = (function (this__5120__auto__){
var self__ = this;
var this__5120__auto____$1 = this;
return (new schema.core.EnumSchema(self__.vs,self__.__meta,self__.__extmap,self__.__hash));
});

schema.core.EnumSchema.prototype.cljs$core$ICounted$_count$arity$1 = (function (this__5130__auto__){
var self__ = this;
var this__5130__auto____$1 = this;
return (1 + cljs.core.count(self__.__extmap));
});

schema.core.EnumSchema.prototype.cljs$core$IHash$_hash$arity$1 = (function (this__5121__auto__){
var self__ = this;
var this__5121__auto____$1 = this;
var h__4947__auto__ = self__.__hash;
if(!((h__4947__auto__ == null))){
return h__4947__auto__;
} else {
var h__4947__auto____$1 = cljs.core.hash_imap(this__5121__auto____$1);
self__.__hash = h__4947__auto____$1;

return h__4947__auto____$1;
}
});

schema.core.EnumSchema.prototype.cljs$core$IEquiv$_equiv$arity$2 = (function (this__5122__auto__,other__5123__auto__){
var self__ = this;
var this__5122__auto____$1 = this;
if(cljs.core.truth_((function (){var and__4500__auto__ = other__5123__auto__;
if(cljs.core.truth_(and__4500__auto__)){
var and__4500__auto____$1 = (this__5122__auto____$1.constructor === other__5123__auto__.constructor);
if(and__4500__auto____$1){
return cljs.core.equiv_map(this__5122__auto____$1,other__5123__auto__);
} else {
return and__4500__auto____$1;
}
} else {
return and__4500__auto__;
}
})())){
return true;
} else {
return false;
}
});

schema.core.EnumSchema.prototype.cljs$core$IMap$_dissoc$arity$2 = (function (this__5135__auto__,k__5136__auto__){
var self__ = this;
var this__5135__auto____$1 = this;
if(cljs.core.contains_QMARK_(new cljs.core.PersistentHashSet(null, new cljs.core.PersistentArrayMap(null, 1, [cljs.core.cst$kw$vs,null], null), null),k__5136__auto__)){
return cljs.core.dissoc.cljs$core$IFn$_invoke$arity$2(cljs.core.with_meta(cljs.core.into.cljs$core$IFn$_invoke$arity$2(cljs.core.PersistentArrayMap.EMPTY,this__5135__auto____$1),self__.__meta),k__5136__auto__);
} else {
return (new schema.core.EnumSchema(self__.vs,self__.__meta,cljs.core.not_empty(cljs.core.dissoc.cljs$core$IFn$_invoke$arity$2(self__.__extmap,k__5136__auto__)),null));
}
});

schema.core.EnumSchema.prototype.cljs$core$IAssociative$_assoc$arity$3 = (function (this__5133__auto__,k__5134__auto__,G__27184){
var self__ = this;
var this__5133__auto____$1 = this;
var pred__27188 = cljs.core.keyword_identical_QMARK_;
var expr__27189 = k__5134__auto__;
if(cljs.core.truth_((pred__27188.cljs$core$IFn$_invoke$arity$2 ? pred__27188.cljs$core$IFn$_invoke$arity$2(cljs.core.cst$kw$vs,expr__27189) : pred__27188.call(null,cljs.core.cst$kw$vs,expr__27189)))){
return (new schema.core.EnumSchema(G__27184,self__.__meta,self__.__extmap,null));
} else {
return (new schema.core.EnumSchema(self__.vs,self__.__meta,cljs.core.assoc.cljs$core$IFn$_invoke$arity$3(self__.__extmap,k__5134__auto__,G__27184),null));
}
});

schema.core.EnumSchema.prototype.cljs$core$ISeqable$_seq$arity$1 = (function (this__5138__auto__){
var self__ = this;
var this__5138__auto____$1 = this;
return cljs.core.seq(cljs.core.concat.cljs$core$IFn$_invoke$arity$2(new cljs.core.PersistentVector(null, 1, 5, cljs.core.PersistentVector.EMPTY_NODE, [(new cljs.core.PersistentVector(null,2,(5),cljs.core.PersistentVector.EMPTY_NODE,[cljs.core.cst$kw$vs,self__.vs],null))], null),self__.__extmap));
});

schema.core.EnumSchema.prototype.cljs$core$IWithMeta$_with_meta$arity$2 = (function (this__5125__auto__,G__27184){
var self__ = this;
var this__5125__auto____$1 = this;
return (new schema.core.EnumSchema(self__.vs,G__27184,self__.__extmap,self__.__hash));
});

schema.core.EnumSchema.prototype.cljs$core$ICollection$_conj$arity$2 = (function (this__5131__auto__,entry__5132__auto__){
var self__ = this;
var this__5131__auto____$1 = this;
if(cljs.core.vector_QMARK_(entry__5132__auto__)){
return cljs.core._assoc(this__5131__auto____$1,cljs.core._nth.cljs$core$IFn$_invoke$arity$2(entry__5132__auto__,(0)),cljs.core._nth.cljs$core$IFn$_invoke$arity$2(entry__5132__auto__,(1)));
} else {
return cljs.core.reduce.cljs$core$IFn$_invoke$arity$3(cljs.core._conj,this__5131__auto____$1,entry__5132__auto__);
}
});

schema.core.EnumSchema.prototype.schema$core$Schema$ = true;

schema.core.EnumSchema.prototype.schema$core$Schema$walker$arity$1 = (function (this$){
var self__ = this;
var this$__$1 = this;
return ((function (this$__$1){
return (function (x){
if(cljs.core.contains_QMARK_(self__.vs,x)){
return x;
} else {
return schema.utils.error(schema.utils.__GT_ValidationError(this$__$1,x,(new cljs.core.Delay(((function (this$__$1){
return (function (){
return cljs.core._conj(cljs.core._conj(cljs.core.List.EMPTY,schema.utils.value_name(x)),self__.vs);
});})(this$__$1))
,null)),null));
}
});
;})(this$__$1))
});

schema.core.EnumSchema.prototype.schema$core$Schema$explain$arity$1 = (function (this$){
var self__ = this;
var this$__$1 = this;
return cljs.core.cons(cljs.core.cst$sym$enum,self__.vs);
});

schema.core.EnumSchema.getBasis = (function (){
return new cljs.core.PersistentVector(null, 1, 5, cljs.core.PersistentVector.EMPTY_NODE, [cljs.core.cst$sym$vs], null);
});

schema.core.EnumSchema.cljs$lang$type = true;

schema.core.EnumSchema.cljs$lang$ctorPrSeq = (function (this__5160__auto__){
return cljs.core._conj(cljs.core.List.EMPTY,"schema.core/EnumSchema");
});

schema.core.EnumSchema.cljs$lang$ctorPrWriter = (function (this__5160__auto__,writer__5161__auto__){
return cljs.core._write(writer__5161__auto__,"schema.core/EnumSchema");
});

schema.core.__GT_EnumSchema = (function schema$core$__GT_EnumSchema(vs){
return (new schema.core.EnumSchema(vs,null,null,null));
});

schema.core.map__GT_EnumSchema = (function schema$core$map__GT_EnumSchema(G__27186){
return (new schema.core.EnumSchema(cljs.core.cst$kw$vs.cljs$core$IFn$_invoke$arity$1(G__27186),null,cljs.core.dissoc.cljs$core$IFn$_invoke$arity$2(G__27186,cljs.core.cst$kw$vs),null));
});

/**
 * A value that must be = to some element of vs.
 */
schema.core.enum$ = (function schema$core$enum(var_args){
var args__5577__auto__ = [];
var len__5570__auto___27193 = arguments.length;
var i__5571__auto___27194 = (0);
while(true){
if((i__5571__auto___27194 < len__5570__auto___27193)){
args__5577__auto__.push((arguments[i__5571__auto___27194]));

var G__27195 = (i__5571__auto___27194 + (1));
i__5571__auto___27194 = G__27195;
continue;
} else {
}
break;
}

var argseq__5578__auto__ = ((((0) < args__5577__auto__.length))?(new cljs.core.IndexedSeq(args__5577__auto__.slice((0)),(0))):null);
return schema.core.enum$.cljs$core$IFn$_invoke$arity$variadic(argseq__5578__auto__);
});

schema.core.enum$.cljs$core$IFn$_invoke$arity$variadic = (function (vs){
return (new schema.core.EnumSchema(cljs.core.set(vs),null,null,null));
});

schema.core.enum$.cljs$lang$maxFixedArity = (0);

schema.core.enum$.cljs$lang$applyTo = (function (seq27192){
return schema.core.enum$.cljs$core$IFn$_invoke$arity$variadic(cljs.core.seq(seq27192));
});

/**
* @constructor
 * @implements {cljs.core.IRecord}
 * @implements {cljs.core.IEquiv}
 * @implements {cljs.core.IHash}
 * @implements {cljs.core.ICollection}
 * @implements {schema.core.Schema}
 * @implements {cljs.core.ICounted}
 * @implements {cljs.core.ISeqable}
 * @implements {cljs.core.IMeta}
 * @implements {cljs.core.ICloneable}
 * @implements {cljs.core.IPrintWithWriter}
 * @implements {cljs.core.IIterable}
 * @implements {cljs.core.IWithMeta}
 * @implements {cljs.core.IAssociative}
 * @implements {cljs.core.IMap}
 * @implements {cljs.core.ILookup}
*/
schema.core.Predicate = (function (p_QMARK_,pred_name,__meta,__extmap,__hash){
this.p_QMARK_ = p_QMARK_;
this.pred_name = pred_name;
this.__meta = __meta;
this.__extmap = __extmap;
this.__hash = __hash;
this.cljs$lang$protocol_mask$partition0$ = 2229667594;
this.cljs$lang$protocol_mask$partition1$ = 8192;
})
schema.core.Predicate.prototype.cljs$core$ILookup$_lookup$arity$2 = (function (this__5126__auto__,k__5127__auto__){
var self__ = this;
var this__5126__auto____$1 = this;
return cljs.core._lookup.cljs$core$IFn$_invoke$arity$3(this__5126__auto____$1,k__5127__auto__,null);
});

schema.core.Predicate.prototype.cljs$core$ILookup$_lookup$arity$3 = (function (this__5128__auto__,k27197,else__5129__auto__){
var self__ = this;
var this__5128__auto____$1 = this;
var G__27199 = (((k27197 instanceof cljs.core.Keyword))?k27197.fqn:null);
switch (G__27199) {
case "p?":
return self__.p_QMARK_;

break;
case "pred-name":
return self__.pred_name;

break;
default:
return cljs.core.get.cljs$core$IFn$_invoke$arity$3(self__.__extmap,k27197,else__5129__auto__);

}
});

schema.core.Predicate.prototype.cljs$core$IPrintWithWriter$_pr_writer$arity$3 = (function (this__5140__auto__,writer__5141__auto__,opts__5142__auto__){
var self__ = this;
var this__5140__auto____$1 = this;
var pr_pair__5143__auto__ = ((function (this__5140__auto____$1){
return (function (keyval__5144__auto__){
return cljs.core.pr_sequential_writer(writer__5141__auto__,cljs.core.pr_writer,""," ","",opts__5142__auto__,keyval__5144__auto__);
});})(this__5140__auto____$1))
;
return cljs.core.pr_sequential_writer(writer__5141__auto__,pr_pair__5143__auto__,"#schema.core.Predicate{",", ","}",opts__5142__auto__,cljs.core.concat.cljs$core$IFn$_invoke$arity$2(new cljs.core.PersistentVector(null, 2, 5, cljs.core.PersistentVector.EMPTY_NODE, [(new cljs.core.PersistentVector(null,2,(5),cljs.core.PersistentVector.EMPTY_NODE,[cljs.core.cst$kw$p_QMARK_,self__.p_QMARK_],null)),(new cljs.core.PersistentVector(null,2,(5),cljs.core.PersistentVector.EMPTY_NODE,[cljs.core.cst$kw$pred_DASH_name,self__.pred_name],null))], null),self__.__extmap));
});

schema.core.Predicate.prototype.cljs$core$IIterable$ = true;

schema.core.Predicate.prototype.cljs$core$IIterable$_iterator$arity$1 = (function (G__27196){
var self__ = this;
var G__27196__$1 = this;
return (new cljs.core.RecordIter((0),G__27196__$1,2,new cljs.core.PersistentVector(null, 2, 5, cljs.core.PersistentVector.EMPTY_NODE, [cljs.core.cst$kw$p_QMARK_,cljs.core.cst$kw$pred_DASH_name], null),cljs.core._iterator(self__.__extmap)));
});

schema.core.Predicate.prototype.cljs$core$IMeta$_meta$arity$1 = (function (this__5124__auto__){
var self__ = this;
var this__5124__auto____$1 = this;
return self__.__meta;
});

schema.core.Predicate.prototype.cljs$core$ICloneable$_clone$arity$1 = (function (this__5120__auto__){
var self__ = this;
var this__5120__auto____$1 = this;
return (new schema.core.Predicate(self__.p_QMARK_,self__.pred_name,self__.__meta,self__.__extmap,self__.__hash));
});

schema.core.Predicate.prototype.cljs$core$ICounted$_count$arity$1 = (function (this__5130__auto__){
var self__ = this;
var this__5130__auto____$1 = this;
return (2 + cljs.core.count(self__.__extmap));
});

schema.core.Predicate.prototype.cljs$core$IHash$_hash$arity$1 = (function (this__5121__auto__){
var self__ = this;
var this__5121__auto____$1 = this;
var h__4947__auto__ = self__.__hash;
if(!((h__4947__auto__ == null))){
return h__4947__auto__;
} else {
var h__4947__auto____$1 = cljs.core.hash_imap(this__5121__auto____$1);
self__.__hash = h__4947__auto____$1;

return h__4947__auto____$1;
}
});

schema.core.Predicate.prototype.cljs$core$IEquiv$_equiv$arity$2 = (function (this__5122__auto__,other__5123__auto__){
var self__ = this;
var this__5122__auto____$1 = this;
if(cljs.core.truth_((function (){var and__4500__auto__ = other__5123__auto__;
if(cljs.core.truth_(and__4500__auto__)){
var and__4500__auto____$1 = (this__5122__auto____$1.constructor === other__5123__auto__.constructor);
if(and__4500__auto____$1){
return cljs.core.equiv_map(this__5122__auto____$1,other__5123__auto__);
} else {
return and__4500__auto____$1;
}
} else {
return and__4500__auto__;
}
})())){
return true;
} else {
return false;
}
});

schema.core.Predicate.prototype.cljs$core$IMap$_dissoc$arity$2 = (function (this__5135__auto__,k__5136__auto__){
var self__ = this;
var this__5135__auto____$1 = this;
if(cljs.core.contains_QMARK_(new cljs.core.PersistentHashSet(null, new cljs.core.PersistentArrayMap(null, 2, [cljs.core.cst$kw$pred_DASH_name,null,cljs.core.cst$kw$p_QMARK_,null], null), null),k__5136__auto__)){
return cljs.core.dissoc.cljs$core$IFn$_invoke$arity$2(cljs.core.with_meta(cljs.core.into.cljs$core$IFn$_invoke$arity$2(cljs.core.PersistentArrayMap.EMPTY,this__5135__auto____$1),self__.__meta),k__5136__auto__);
} else {
return (new schema.core.Predicate(self__.p_QMARK_,self__.pred_name,self__.__meta,cljs.core.not_empty(cljs.core.dissoc.cljs$core$IFn$_invoke$arity$2(self__.__extmap,k__5136__auto__)),null));
}
});

schema.core.Predicate.prototype.cljs$core$IAssociative$_assoc$arity$3 = (function (this__5133__auto__,k__5134__auto__,G__27196){
var self__ = this;
var this__5133__auto____$1 = this;
var pred__27200 = cljs.core.keyword_identical_QMARK_;
var expr__27201 = k__5134__auto__;
if(cljs.core.truth_((pred__27200.cljs$core$IFn$_invoke$arity$2 ? pred__27200.cljs$core$IFn$_invoke$arity$2(cljs.core.cst$kw$p_QMARK_,expr__27201) : pred__27200.call(null,cljs.core.cst$kw$p_QMARK_,expr__27201)))){
return (new schema.core.Predicate(G__27196,self__.pred_name,self__.__meta,self__.__extmap,null));
} else {
if(cljs.core.truth_((pred__27200.cljs$core$IFn$_invoke$arity$2 ? pred__27200.cljs$core$IFn$_invoke$arity$2(cljs.core.cst$kw$pred_DASH_name,expr__27201) : pred__27200.call(null,cljs.core.cst$kw$pred_DASH_name,expr__27201)))){
return (new schema.core.Predicate(self__.p_QMARK_,G__27196,self__.__meta,self__.__extmap,null));
} else {
return (new schema.core.Predicate(self__.p_QMARK_,self__.pred_name,self__.__meta,cljs.core.assoc.cljs$core$IFn$_invoke$arity$3(self__.__extmap,k__5134__auto__,G__27196),null));
}
}
});

schema.core.Predicate.prototype.cljs$core$ISeqable$_seq$arity$1 = (function (this__5138__auto__){
var self__ = this;
var this__5138__auto____$1 = this;
return cljs.core.seq(cljs.core.concat.cljs$core$IFn$_invoke$arity$2(new cljs.core.PersistentVector(null, 2, 5, cljs.core.PersistentVector.EMPTY_NODE, [(new cljs.core.PersistentVector(null,2,(5),cljs.core.PersistentVector.EMPTY_NODE,[cljs.core.cst$kw$p_QMARK_,self__.p_QMARK_],null)),(new cljs.core.PersistentVector(null,2,(5),cljs.core.PersistentVector.EMPTY_NODE,[cljs.core.cst$kw$pred_DASH_name,self__.pred_name],null))], null),self__.__extmap));
});

schema.core.Predicate.prototype.cljs$core$IWithMeta$_with_meta$arity$2 = (function (this__5125__auto__,G__27196){
var self__ = this;
var this__5125__auto____$1 = this;
return (new schema.core.Predicate(self__.p_QMARK_,self__.pred_name,G__27196,self__.__extmap,self__.__hash));
});

schema.core.Predicate.prototype.cljs$core$ICollection$_conj$arity$2 = (function (this__5131__auto__,entry__5132__auto__){
var self__ = this;
var this__5131__auto____$1 = this;
if(cljs.core.vector_QMARK_(entry__5132__auto__)){
return cljs.core._assoc(this__5131__auto____$1,cljs.core._nth.cljs$core$IFn$_invoke$arity$2(entry__5132__auto__,(0)),cljs.core._nth.cljs$core$IFn$_invoke$arity$2(entry__5132__auto__,(1)));
} else {
return cljs.core.reduce.cljs$core$IFn$_invoke$arity$3(cljs.core._conj,this__5131__auto____$1,entry__5132__auto__);
}
});

schema.core.Predicate.prototype.schema$core$Schema$ = true;

schema.core.Predicate.prototype.schema$core$Schema$walker$arity$1 = (function (this$){
var self__ = this;
var this$__$1 = this;
return ((function (this$__$1){
return (function (x){
var temp__4423__auto__ = (function (){try{if(cljs.core.truth_((self__.p_QMARK_.cljs$core$IFn$_invoke$arity$1 ? self__.p_QMARK_.cljs$core$IFn$_invoke$arity$1(x) : self__.p_QMARK_.call(null,x)))){
return null;
} else {
return cljs.core.cst$sym$not;
}
}catch (e27203){if((e27203 instanceof Object)){
var e = e27203;
return cljs.core.cst$sym$throws_QMARK_;
} else {
throw e27203;

}
}})();
if(cljs.core.truth_(temp__4423__auto__)){
var reason = temp__4423__auto__;
return schema.utils.error(schema.utils.__GT_ValidationError(this$__$1,x,(new cljs.core.Delay(((function (reason,temp__4423__auto__,this$__$1){
return (function (){
return cljs.core._conj(cljs.core._conj(cljs.core.List.EMPTY,schema.utils.value_name(x)),self__.pred_name);
});})(reason,temp__4423__auto__,this$__$1))
,null)),reason));
} else {
return x;
}
});
;})(this$__$1))
});

schema.core.Predicate.prototype.schema$core$Schema$explain$arity$1 = (function (this$){
var self__ = this;
var this$__$1 = this;
if(cljs.core._EQ_.cljs$core$IFn$_invoke$arity$2(self__.p_QMARK_,cljs.core.integer_QMARK_)){
return cljs.core.cst$sym$Int;
} else {
if(cljs.core._EQ_.cljs$core$IFn$_invoke$arity$2(self__.p_QMARK_,cljs.core.keyword_QMARK_)){
return cljs.core.cst$sym$Keyword;
} else {
if(cljs.core._EQ_.cljs$core$IFn$_invoke$arity$2(self__.p_QMARK_,cljs.core.symbol_QMARK_)){
return cljs.core.cst$sym$Symbol;
} else {
if(cljs.core._EQ_.cljs$core$IFn$_invoke$arity$2(self__.p_QMARK_,cljs.core.string_QMARK_)){
return cljs.core.cst$sym$Str;
} else {
return cljs.core._conj(cljs.core._conj(cljs.core.List.EMPTY,self__.pred_name),cljs.core.cst$sym$pred);

}
}
}
}
});

schema.core.Predicate.getBasis = (function (){
return new cljs.core.PersistentVector(null, 2, 5, cljs.core.PersistentVector.EMPTY_NODE, [cljs.core.cst$sym$p_QMARK_,cljs.core.cst$sym$pred_DASH_name], null);
});

schema.core.Predicate.cljs$lang$type = true;

schema.core.Predicate.cljs$lang$ctorPrSeq = (function (this__5160__auto__){
return cljs.core._conj(cljs.core.List.EMPTY,"schema.core/Predicate");
});

schema.core.Predicate.cljs$lang$ctorPrWriter = (function (this__5160__auto__,writer__5161__auto__){
return cljs.core._write(writer__5161__auto__,"schema.core/Predicate");
});

schema.core.__GT_Predicate = (function schema$core$__GT_Predicate(p_QMARK_,pred_name){
return (new schema.core.Predicate(p_QMARK_,pred_name,null,null,null));
});

schema.core.map__GT_Predicate = (function schema$core$map__GT_Predicate(G__27198){
return (new schema.core.Predicate(cljs.core.cst$kw$p_QMARK_.cljs$core$IFn$_invoke$arity$1(G__27198),cljs.core.cst$kw$pred_DASH_name.cljs$core$IFn$_invoke$arity$1(G__27198),null,cljs.core.dissoc.cljs$core$IFn$_invoke$arity$variadic(G__27198,cljs.core.cst$kw$p_QMARK_,cljs.core.array_seq([cljs.core.cst$kw$pred_DASH_name], 0)),null));
});

/**
 * A value for which p? returns true (and does not throw).
 * Optional pred-name can be passed for nicer validation errors.
 */
schema.core.pred = (function schema$core$pred(var_args){
var args27205 = [];
var len__5570__auto___27208 = arguments.length;
var i__5571__auto___27209 = (0);
while(true){
if((i__5571__auto___27209 < len__5570__auto___27208)){
args27205.push((arguments[i__5571__auto___27209]));

var G__27210 = (i__5571__auto___27209 + (1));
i__5571__auto___27209 = G__27210;
continue;
} else {
}
break;
}

var G__27207 = args27205.length;
switch (G__27207) {
case 1:
return schema.core.pred.cljs$core$IFn$_invoke$arity$1((arguments[(0)]));

break;
case 2:
return schema.core.pred.cljs$core$IFn$_invoke$arity$2((arguments[(0)]),(arguments[(1)]));

break;
default:
throw (new Error([cljs.core.str("Invalid arity: "),cljs.core.str(args27205.length)].join('')));

}
});

schema.core.pred.cljs$core$IFn$_invoke$arity$1 = (function (p_QMARK_){
return schema.core.pred.cljs$core$IFn$_invoke$arity$2(p_QMARK_,p_QMARK_);
});

schema.core.pred.cljs$core$IFn$_invoke$arity$2 = (function (p_QMARK_,pred_name){
if(cljs.core.ifn_QMARK_(p_QMARK_)){
} else {
throw (new Error(schema.utils.format_STAR_.cljs$core$IFn$_invoke$arity$variadic("Not a function: %s",cljs.core.array_seq([p_QMARK_], 0))));
}

return (new schema.core.Predicate(p_QMARK_,pred_name,null,null,null));
});

schema.core.pred.cljs$lang$maxFixedArity = 2;
schema.core.protocol_name = (function schema$core$protocol_name(protocol){
return cljs.core.cst$kw$proto_DASH_sym.cljs$core$IFn$_invoke$arity$1(cljs.core.meta(protocol));
});

/**
* @constructor
 * @implements {cljs.core.IRecord}
 * @implements {cljs.core.IEquiv}
 * @implements {cljs.core.IHash}
 * @implements {cljs.core.ICollection}
 * @implements {schema.core.Schema}
 * @implements {cljs.core.ICounted}
 * @implements {cljs.core.ISeqable}
 * @implements {cljs.core.IMeta}
 * @implements {cljs.core.ICloneable}
 * @implements {cljs.core.IPrintWithWriter}
 * @implements {cljs.core.IIterable}
 * @implements {cljs.core.IWithMeta}
 * @implements {cljs.core.IAssociative}
 * @implements {cljs.core.IMap}
 * @implements {cljs.core.ILookup}
*/
schema.core.Protocol = (function (p,__meta,__extmap,__hash){
this.p = p;
this.__meta = __meta;
this.__extmap = __extmap;
this.__hash = __hash;
this.cljs$lang$protocol_mask$partition0$ = 2229667594;
this.cljs$lang$protocol_mask$partition1$ = 8192;
})
schema.core.Protocol.prototype.cljs$core$ILookup$_lookup$arity$2 = (function (this__5126__auto__,k__5127__auto__){
var self__ = this;
var this__5126__auto____$1 = this;
return cljs.core._lookup.cljs$core$IFn$_invoke$arity$3(this__5126__auto____$1,k__5127__auto__,null);
});

schema.core.Protocol.prototype.cljs$core$ILookup$_lookup$arity$3 = (function (this__5128__auto__,k27213,else__5129__auto__){
var self__ = this;
var this__5128__auto____$1 = this;
var G__27215 = (((k27213 instanceof cljs.core.Keyword))?k27213.fqn:null);
switch (G__27215) {
case "p":
return self__.p;

break;
default:
return cljs.core.get.cljs$core$IFn$_invoke$arity$3(self__.__extmap,k27213,else__5129__auto__);

}
});

schema.core.Protocol.prototype.cljs$core$IPrintWithWriter$_pr_writer$arity$3 = (function (this__5140__auto__,writer__5141__auto__,opts__5142__auto__){
var self__ = this;
var this__5140__auto____$1 = this;
var pr_pair__5143__auto__ = ((function (this__5140__auto____$1){
return (function (keyval__5144__auto__){
return cljs.core.pr_sequential_writer(writer__5141__auto__,cljs.core.pr_writer,""," ","",opts__5142__auto__,keyval__5144__auto__);
});})(this__5140__auto____$1))
;
return cljs.core.pr_sequential_writer(writer__5141__auto__,pr_pair__5143__auto__,"#schema.core.Protocol{",", ","}",opts__5142__auto__,cljs.core.concat.cljs$core$IFn$_invoke$arity$2(new cljs.core.PersistentVector(null, 1, 5, cljs.core.PersistentVector.EMPTY_NODE, [(new cljs.core.PersistentVector(null,2,(5),cljs.core.PersistentVector.EMPTY_NODE,[cljs.core.cst$kw$p,self__.p],null))], null),self__.__extmap));
});

schema.core.Protocol.prototype.cljs$core$IIterable$ = true;

schema.core.Protocol.prototype.cljs$core$IIterable$_iterator$arity$1 = (function (G__27212){
var self__ = this;
var G__27212__$1 = this;
return (new cljs.core.RecordIter((0),G__27212__$1,1,new cljs.core.PersistentVector(null, 1, 5, cljs.core.PersistentVector.EMPTY_NODE, [cljs.core.cst$kw$p], null),cljs.core._iterator(self__.__extmap)));
});

schema.core.Protocol.prototype.cljs$core$IMeta$_meta$arity$1 = (function (this__5124__auto__){
var self__ = this;
var this__5124__auto____$1 = this;
return self__.__meta;
});

schema.core.Protocol.prototype.cljs$core$ICloneable$_clone$arity$1 = (function (this__5120__auto__){
var self__ = this;
var this__5120__auto____$1 = this;
return (new schema.core.Protocol(self__.p,self__.__meta,self__.__extmap,self__.__hash));
});

schema.core.Protocol.prototype.cljs$core$ICounted$_count$arity$1 = (function (this__5130__auto__){
var self__ = this;
var this__5130__auto____$1 = this;
return (1 + cljs.core.count(self__.__extmap));
});

schema.core.Protocol.prototype.cljs$core$IHash$_hash$arity$1 = (function (this__5121__auto__){
var self__ = this;
var this__5121__auto____$1 = this;
var h__4947__auto__ = self__.__hash;
if(!((h__4947__auto__ == null))){
return h__4947__auto__;
} else {
var h__4947__auto____$1 = cljs.core.hash_imap(this__5121__auto____$1);
self__.__hash = h__4947__auto____$1;

return h__4947__auto____$1;
}
});

schema.core.Protocol.prototype.cljs$core$IEquiv$_equiv$arity$2 = (function (this__5122__auto__,other__5123__auto__){
var self__ = this;
var this__5122__auto____$1 = this;
if(cljs.core.truth_((function (){var and__4500__auto__ = other__5123__auto__;
if(cljs.core.truth_(and__4500__auto__)){
var and__4500__auto____$1 = (this__5122__auto____$1.constructor === other__5123__auto__.constructor);
if(and__4500__auto____$1){
return cljs.core.equiv_map(this__5122__auto____$1,other__5123__auto__);
} else {
return and__4500__auto____$1;
}
} else {
return and__4500__auto__;
}
})())){
return true;
} else {
return false;
}
});

schema.core.Protocol.prototype.cljs$core$IMap$_dissoc$arity$2 = (function (this__5135__auto__,k__5136__auto__){
var self__ = this;
var this__5135__auto____$1 = this;
if(cljs.core.contains_QMARK_(new cljs.core.PersistentHashSet(null, new cljs.core.PersistentArrayMap(null, 1, [cljs.core.cst$kw$p,null], null), null),k__5136__auto__)){
return cljs.core.dissoc.cljs$core$IFn$_invoke$arity$2(cljs.core.with_meta(cljs.core.into.cljs$core$IFn$_invoke$arity$2(cljs.core.PersistentArrayMap.EMPTY,this__5135__auto____$1),self__.__meta),k__5136__auto__);
} else {
return (new schema.core.Protocol(self__.p,self__.__meta,cljs.core.not_empty(cljs.core.dissoc.cljs$core$IFn$_invoke$arity$2(self__.__extmap,k__5136__auto__)),null));
}
});

schema.core.Protocol.prototype.cljs$core$IAssociative$_assoc$arity$3 = (function (this__5133__auto__,k__5134__auto__,G__27212){
var self__ = this;
var this__5133__auto____$1 = this;
var pred__27216 = cljs.core.keyword_identical_QMARK_;
var expr__27217 = k__5134__auto__;
if(cljs.core.truth_((pred__27216.cljs$core$IFn$_invoke$arity$2 ? pred__27216.cljs$core$IFn$_invoke$arity$2(cljs.core.cst$kw$p,expr__27217) : pred__27216.call(null,cljs.core.cst$kw$p,expr__27217)))){
return (new schema.core.Protocol(G__27212,self__.__meta,self__.__extmap,null));
} else {
return (new schema.core.Protocol(self__.p,self__.__meta,cljs.core.assoc.cljs$core$IFn$_invoke$arity$3(self__.__extmap,k__5134__auto__,G__27212),null));
}
});

schema.core.Protocol.prototype.cljs$core$ISeqable$_seq$arity$1 = (function (this__5138__auto__){
var self__ = this;
var this__5138__auto____$1 = this;
return cljs.core.seq(cljs.core.concat.cljs$core$IFn$_invoke$arity$2(new cljs.core.PersistentVector(null, 1, 5, cljs.core.PersistentVector.EMPTY_NODE, [(new cljs.core.PersistentVector(null,2,(5),cljs.core.PersistentVector.EMPTY_NODE,[cljs.core.cst$kw$p,self__.p],null))], null),self__.__extmap));
});

schema.core.Protocol.prototype.cljs$core$IWithMeta$_with_meta$arity$2 = (function (this__5125__auto__,G__27212){
var self__ = this;
var this__5125__auto____$1 = this;
return (new schema.core.Protocol(self__.p,G__27212,self__.__extmap,self__.__hash));
});

schema.core.Protocol.prototype.cljs$core$ICollection$_conj$arity$2 = (function (this__5131__auto__,entry__5132__auto__){
var self__ = this;
var this__5131__auto____$1 = this;
if(cljs.core.vector_QMARK_(entry__5132__auto__)){
return cljs.core._assoc(this__5131__auto____$1,cljs.core._nth.cljs$core$IFn$_invoke$arity$2(entry__5132__auto__,(0)),cljs.core._nth.cljs$core$IFn$_invoke$arity$2(entry__5132__auto__,(1)));
} else {
return cljs.core.reduce.cljs$core$IFn$_invoke$arity$3(cljs.core._conj,this__5131__auto____$1,entry__5132__auto__);
}
});

schema.core.Protocol.prototype.schema$core$Schema$ = true;

schema.core.Protocol.prototype.schema$core$Schema$walker$arity$1 = (function (this$){
var self__ = this;
var this$__$1 = this;
return ((function (this$__$1){
return (function (x){
if(cljs.core.truth_(cljs.core.cst$kw$proto_DASH_pred.cljs$core$IFn$_invoke$arity$1(cljs.core.meta(this$__$1)).call(null,x))){
return x;
} else {
return schema.utils.error(schema.utils.__GT_ValidationError(this$__$1,x,(new cljs.core.Delay(((function (this$__$1){
return (function (){
return cljs.core._conj(cljs.core._conj(cljs.core._conj(cljs.core.List.EMPTY,schema.utils.value_name(x)),schema.core.protocol_name(this$__$1)),cljs.core.cst$sym$satisfies_QMARK_);
});})(this$__$1))
,null)),null));
}
});
;})(this$__$1))
});

schema.core.Protocol.prototype.schema$core$Schema$explain$arity$1 = (function (this$){
var self__ = this;
var this$__$1 = this;
return cljs.core._conj(cljs.core._conj(cljs.core.List.EMPTY,schema.core.protocol_name(this$__$1)),cljs.core.cst$sym$protocol);
});

schema.core.Protocol.getBasis = (function (){
return new cljs.core.PersistentVector(null, 1, 5, cljs.core.PersistentVector.EMPTY_NODE, [cljs.core.cst$sym$p], null);
});

schema.core.Protocol.cljs$lang$type = true;

schema.core.Protocol.cljs$lang$ctorPrSeq = (function (this__5160__auto__){
return cljs.core._conj(cljs.core.List.EMPTY,"schema.core/Protocol");
});

schema.core.Protocol.cljs$lang$ctorPrWriter = (function (this__5160__auto__,writer__5161__auto__){
return cljs.core._write(writer__5161__auto__,"schema.core/Protocol");
});

schema.core.__GT_Protocol = (function schema$core$__GT_Protocol(p){
return (new schema.core.Protocol(p,null,null,null));
});

schema.core.map__GT_Protocol = (function schema$core$map__GT_Protocol(G__27214){
return (new schema.core.Protocol(cljs.core.cst$kw$p.cljs$core$IFn$_invoke$arity$1(G__27214),null,cljs.core.dissoc.cljs$core$IFn$_invoke$arity$2(G__27214,cljs.core.cst$kw$p),null));
});

RegExp.prototype.schema$core$Schema$ = true;

RegExp.prototype.schema$core$Schema$walker$arity$1 = (function (this$){
var this$__$1 = this;
return ((function (this$__$1){
return (function (x){
if(!(typeof x === 'string')){
return schema.utils.error(schema.utils.__GT_ValidationError(this$__$1,x,(new cljs.core.Delay(((function (this$__$1){
return (function (){
return cljs.core._conj(cljs.core._conj(cljs.core.List.EMPTY,schema.utils.value_name(x)),cljs.core.cst$sym$string_QMARK_);
});})(this$__$1))
,null)),null));
} else {
if(cljs.core.not(cljs.core.re_find(this$__$1,x))){
return schema.utils.error(schema.utils.__GT_ValidationError(this$__$1,x,(new cljs.core.Delay(((function (this$__$1){
return (function (){
return cljs.core._conj(cljs.core._conj(cljs.core._conj(cljs.core.List.EMPTY,schema.utils.value_name(x)),schema.core.explain(this$__$1)),cljs.core.cst$sym$re_DASH_find);
});})(this$__$1))
,null)),null));
} else {
return x;

}
}
});
;})(this$__$1))
});

RegExp.prototype.schema$core$Schema$explain$arity$1 = (function (this$){
var this$__$1 = this;
return cljs.core.symbol.cljs$core$IFn$_invoke$arity$1([cljs.core.str("#\""),cljs.core.str([cljs.core.str(this$__$1)].join('').slice((1),(-1))),cljs.core.str("\"")].join(''));
});
/**
 * Satisfied only by String.
 * Is (pred string?) and not js/String in cljs because of keywords.
 */
schema.core.Str = schema.core.pred.cljs$core$IFn$_invoke$arity$1(cljs.core.string_QMARK_);
/**
 * Boolean true or false
 */
schema.core.Bool = Boolean;
/**
 * Any number
 */
schema.core.Num = Number;
/**
 * Any integral number
 */
schema.core.Int = schema.core.pred.cljs$core$IFn$_invoke$arity$2(cljs.core.integer_QMARK_,cljs.core.cst$sym$integer_QMARK_);
/**
 * A keyword
 */
schema.core.Keyword = schema.core.pred.cljs$core$IFn$_invoke$arity$2(cljs.core.keyword_QMARK_,cljs.core.cst$sym$keyword_QMARK_);
/**
 * A symbol
 */
schema.core.Symbol = schema.core.pred.cljs$core$IFn$_invoke$arity$2(cljs.core.symbol_QMARK_,cljs.core.cst$sym$symbol_QMARK_);
/**
 * A regular expression
 */
schema.core.Regex = (function (){
if(typeof schema.core.t_schema$core27220 !== 'undefined'){
} else {

/**
* @constructor
 * @implements {schema.core.Schema}
 * @implements {cljs.core.IMeta}
 * @implements {cljs.core.IWithMeta}
*/
schema.core.t_schema$core27220 = (function (meta27221){
this.meta27221 = meta27221;
this.cljs$lang$protocol_mask$partition0$ = 393216;
this.cljs$lang$protocol_mask$partition1$ = 0;
})
schema.core.t_schema$core27220.prototype.cljs$core$IWithMeta$_with_meta$arity$2 = (function (_27222,meta27221__$1){
var self__ = this;
var _27222__$1 = this;
return (new schema.core.t_schema$core27220(meta27221__$1));
});

schema.core.t_schema$core27220.prototype.cljs$core$IMeta$_meta$arity$1 = (function (_27222){
var self__ = this;
var _27222__$1 = this;
return self__.meta27221;
});

schema.core.t_schema$core27220.prototype.schema$core$Schema$ = true;

schema.core.t_schema$core27220.prototype.schema$core$Schema$walker$arity$1 = (function (this$){
var self__ = this;
var this$__$1 = this;
return ((function (this$__$1){
return (function (x){
if((x instanceof RegExp)){
return x;
} else {
return schema.utils.error(schema.utils.__GT_ValidationError(this$__$1,x,(new cljs.core.Delay(((function (this$__$1){
return (function (){
return cljs.core._conj(cljs.core._conj(cljs.core._conj(cljs.core.List.EMPTY,schema.utils.value_name(x)),cljs.core.cst$sym$js_SLASH_RegExp),cljs.core.cst$sym$instance_QMARK_);
});})(this$__$1))
,null)),null));
}
});
;})(this$__$1))
});

schema.core.t_schema$core27220.prototype.schema$core$Schema$explain$arity$1 = (function (this$){
var self__ = this;
var this$__$1 = this;
return cljs.core.cst$sym$Regex;
});

schema.core.t_schema$core27220.getBasis = (function (){
return new cljs.core.PersistentVector(null, 1, 5, cljs.core.PersistentVector.EMPTY_NODE, [cljs.core.cst$sym$meta27221], null);
});

schema.core.t_schema$core27220.cljs$lang$type = true;

schema.core.t_schema$core27220.cljs$lang$ctorStr = "schema.core/t_schema$core27220";

schema.core.t_schema$core27220.cljs$lang$ctorPrWriter = (function (this__5110__auto__,writer__5111__auto__,opt__5112__auto__){
return cljs.core._write(writer__5111__auto__,"schema.core/t_schema$core27220");
});

schema.core.__GT_t_schema$core27220 = (function schema$core$__GT_t_schema$core27220(meta27221){
return (new schema.core.t_schema$core27220(meta27221));
});

}

return (new schema.core.t_schema$core27220(cljs.core.PersistentArrayMap.EMPTY));
})()
;
/**
 * The local representation of #inst ...
 */
schema.core.Inst = Date;
/**
 * The local representation of #uuid ...
 */
schema.core.Uuid = cljs.core.UUID;

/**
* @constructor
 * @implements {cljs.core.IRecord}
 * @implements {cljs.core.IEquiv}
 * @implements {cljs.core.IHash}
 * @implements {cljs.core.ICollection}
 * @implements {schema.core.Schema}
 * @implements {cljs.core.ICounted}
 * @implements {cljs.core.ISeqable}
 * @implements {cljs.core.IMeta}
 * @implements {cljs.core.ICloneable}
 * @implements {cljs.core.IPrintWithWriter}
 * @implements {cljs.core.IIterable}
 * @implements {cljs.core.IWithMeta}
 * @implements {cljs.core.IAssociative}
 * @implements {cljs.core.IMap}
 * @implements {cljs.core.ILookup}
*/
schema.core.Maybe = (function (schema,__meta,__extmap,__hash){
this.schema = schema;
this.__meta = __meta;
this.__extmap = __extmap;
this.__hash = __hash;
this.cljs$lang$protocol_mask$partition0$ = 2229667594;
this.cljs$lang$protocol_mask$partition1$ = 8192;
})
schema.core.Maybe.prototype.cljs$core$ILookup$_lookup$arity$2 = (function (this__5126__auto__,k__5127__auto__){
var self__ = this;
var this__5126__auto____$1 = this;
return cljs.core._lookup.cljs$core$IFn$_invoke$arity$3(this__5126__auto____$1,k__5127__auto__,null);
});

schema.core.Maybe.prototype.cljs$core$ILookup$_lookup$arity$3 = (function (this__5128__auto__,k27224,else__5129__auto__){
var self__ = this;
var this__5128__auto____$1 = this;
var G__27226 = (((k27224 instanceof cljs.core.Keyword))?k27224.fqn:null);
switch (G__27226) {
case "schema":
return self__.schema;

break;
default:
return cljs.core.get.cljs$core$IFn$_invoke$arity$3(self__.__extmap,k27224,else__5129__auto__);

}
});

schema.core.Maybe.prototype.cljs$core$IPrintWithWriter$_pr_writer$arity$3 = (function (this__5140__auto__,writer__5141__auto__,opts__5142__auto__){
var self__ = this;
var this__5140__auto____$1 = this;
var pr_pair__5143__auto__ = ((function (this__5140__auto____$1){
return (function (keyval__5144__auto__){
return cljs.core.pr_sequential_writer(writer__5141__auto__,cljs.core.pr_writer,""," ","",opts__5142__auto__,keyval__5144__auto__);
});})(this__5140__auto____$1))
;
return cljs.core.pr_sequential_writer(writer__5141__auto__,pr_pair__5143__auto__,"#schema.core.Maybe{",", ","}",opts__5142__auto__,cljs.core.concat.cljs$core$IFn$_invoke$arity$2(new cljs.core.PersistentVector(null, 1, 5, cljs.core.PersistentVector.EMPTY_NODE, [(new cljs.core.PersistentVector(null,2,(5),cljs.core.PersistentVector.EMPTY_NODE,[cljs.core.cst$kw$schema,self__.schema],null))], null),self__.__extmap));
});

schema.core.Maybe.prototype.cljs$core$IIterable$ = true;

schema.core.Maybe.prototype.cljs$core$IIterable$_iterator$arity$1 = (function (G__27223){
var self__ = this;
var G__27223__$1 = this;
return (new cljs.core.RecordIter((0),G__27223__$1,1,new cljs.core.PersistentVector(null, 1, 5, cljs.core.PersistentVector.EMPTY_NODE, [cljs.core.cst$kw$schema], null),cljs.core._iterator(self__.__extmap)));
});

schema.core.Maybe.prototype.cljs$core$IMeta$_meta$arity$1 = (function (this__5124__auto__){
var self__ = this;
var this__5124__auto____$1 = this;
return self__.__meta;
});

schema.core.Maybe.prototype.cljs$core$ICloneable$_clone$arity$1 = (function (this__5120__auto__){
var self__ = this;
var this__5120__auto____$1 = this;
return (new schema.core.Maybe(self__.schema,self__.__meta,self__.__extmap,self__.__hash));
});

schema.core.Maybe.prototype.cljs$core$ICounted$_count$arity$1 = (function (this__5130__auto__){
var self__ = this;
var this__5130__auto____$1 = this;
return (1 + cljs.core.count(self__.__extmap));
});

schema.core.Maybe.prototype.cljs$core$IHash$_hash$arity$1 = (function (this__5121__auto__){
var self__ = this;
var this__5121__auto____$1 = this;
var h__4947__auto__ = self__.__hash;
if(!((h__4947__auto__ == null))){
return h__4947__auto__;
} else {
var h__4947__auto____$1 = cljs.core.hash_imap(this__5121__auto____$1);
self__.__hash = h__4947__auto____$1;

return h__4947__auto____$1;
}
});

schema.core.Maybe.prototype.cljs$core$IEquiv$_equiv$arity$2 = (function (this__5122__auto__,other__5123__auto__){
var self__ = this;
var this__5122__auto____$1 = this;
if(cljs.core.truth_((function (){var and__4500__auto__ = other__5123__auto__;
if(cljs.core.truth_(and__4500__auto__)){
var and__4500__auto____$1 = (this__5122__auto____$1.constructor === other__5123__auto__.constructor);
if(and__4500__auto____$1){
return cljs.core.equiv_map(this__5122__auto____$1,other__5123__auto__);
} else {
return and__4500__auto____$1;
}
} else {
return and__4500__auto__;
}
})())){
return true;
} else {
return false;
}
});

schema.core.Maybe.prototype.cljs$core$IMap$_dissoc$arity$2 = (function (this__5135__auto__,k__5136__auto__){
var self__ = this;
var this__5135__auto____$1 = this;
if(cljs.core.contains_QMARK_(new cljs.core.PersistentHashSet(null, new cljs.core.PersistentArrayMap(null, 1, [cljs.core.cst$kw$schema,null], null), null),k__5136__auto__)){
return cljs.core.dissoc.cljs$core$IFn$_invoke$arity$2(cljs.core.with_meta(cljs.core.into.cljs$core$IFn$_invoke$arity$2(cljs.core.PersistentArrayMap.EMPTY,this__5135__auto____$1),self__.__meta),k__5136__auto__);
} else {
return (new schema.core.Maybe(self__.schema,self__.__meta,cljs.core.not_empty(cljs.core.dissoc.cljs$core$IFn$_invoke$arity$2(self__.__extmap,k__5136__auto__)),null));
}
});

schema.core.Maybe.prototype.cljs$core$IAssociative$_assoc$arity$3 = (function (this__5133__auto__,k__5134__auto__,G__27223){
var self__ = this;
var this__5133__auto____$1 = this;
var pred__27227 = cljs.core.keyword_identical_QMARK_;
var expr__27228 = k__5134__auto__;
if(cljs.core.truth_((pred__27227.cljs$core$IFn$_invoke$arity$2 ? pred__27227.cljs$core$IFn$_invoke$arity$2(cljs.core.cst$kw$schema,expr__27228) : pred__27227.call(null,cljs.core.cst$kw$schema,expr__27228)))){
return (new schema.core.Maybe(G__27223,self__.__meta,self__.__extmap,null));
} else {
return (new schema.core.Maybe(self__.schema,self__.__meta,cljs.core.assoc.cljs$core$IFn$_invoke$arity$3(self__.__extmap,k__5134__auto__,G__27223),null));
}
});

schema.core.Maybe.prototype.cljs$core$ISeqable$_seq$arity$1 = (function (this__5138__auto__){
var self__ = this;
var this__5138__auto____$1 = this;
return cljs.core.seq(cljs.core.concat.cljs$core$IFn$_invoke$arity$2(new cljs.core.PersistentVector(null, 1, 5, cljs.core.PersistentVector.EMPTY_NODE, [(new cljs.core.PersistentVector(null,2,(5),cljs.core.PersistentVector.EMPTY_NODE,[cljs.core.cst$kw$schema,self__.schema],null))], null),self__.__extmap));
});

schema.core.Maybe.prototype.cljs$core$IWithMeta$_with_meta$arity$2 = (function (this__5125__auto__,G__27223){
var self__ = this;
var this__5125__auto____$1 = this;
return (new schema.core.Maybe(self__.schema,G__27223,self__.__extmap,self__.__hash));
});

schema.core.Maybe.prototype.cljs$core$ICollection$_conj$arity$2 = (function (this__5131__auto__,entry__5132__auto__){
var self__ = this;
var this__5131__auto____$1 = this;
if(cljs.core.vector_QMARK_(entry__5132__auto__)){
return cljs.core._assoc(this__5131__auto____$1,cljs.core._nth.cljs$core$IFn$_invoke$arity$2(entry__5132__auto__,(0)),cljs.core._nth.cljs$core$IFn$_invoke$arity$2(entry__5132__auto__,(1)));
} else {
return cljs.core.reduce.cljs$core$IFn$_invoke$arity$3(cljs.core._conj,this__5131__auto____$1,entry__5132__auto__);
}
});

schema.core.Maybe.prototype.schema$core$Schema$ = true;

schema.core.Maybe.prototype.schema$core$Schema$walker$arity$1 = (function (this$){
var self__ = this;
var this$__$1 = this;
var sub_walker = (schema.core.subschema_walker.cljs$core$IFn$_invoke$arity$1 ? schema.core.subschema_walker.cljs$core$IFn$_invoke$arity$1(self__.schema) : schema.core.subschema_walker.call(null,self__.schema));
return ((function (sub_walker,this$__$1){
return (function (x){
if((x == null)){
return null;
} else {
return (sub_walker.cljs$core$IFn$_invoke$arity$1 ? sub_walker.cljs$core$IFn$_invoke$arity$1(x) : sub_walker.call(null,x));
}
});
;})(sub_walker,this$__$1))
});

schema.core.Maybe.prototype.schema$core$Schema$explain$arity$1 = (function (this$){
var self__ = this;
var this$__$1 = this;
return cljs.core._conj(cljs.core._conj(cljs.core.List.EMPTY,schema.core.explain(self__.schema)),cljs.core.cst$sym$maybe);
});

schema.core.Maybe.getBasis = (function (){
return new cljs.core.PersistentVector(null, 1, 5, cljs.core.PersistentVector.EMPTY_NODE, [cljs.core.cst$sym$schema], null);
});

schema.core.Maybe.cljs$lang$type = true;

schema.core.Maybe.cljs$lang$ctorPrSeq = (function (this__5160__auto__){
return cljs.core._conj(cljs.core.List.EMPTY,"schema.core/Maybe");
});

schema.core.Maybe.cljs$lang$ctorPrWriter = (function (this__5160__auto__,writer__5161__auto__){
return cljs.core._write(writer__5161__auto__,"schema.core/Maybe");
});

schema.core.__GT_Maybe = (function schema$core$__GT_Maybe(schema__$1){
return (new schema.core.Maybe(schema__$1,null,null,null));
});

schema.core.map__GT_Maybe = (function schema$core$map__GT_Maybe(G__27225){
return (new schema.core.Maybe(cljs.core.cst$kw$schema.cljs$core$IFn$_invoke$arity$1(G__27225),null,cljs.core.dissoc.cljs$core$IFn$_invoke$arity$2(G__27225,cljs.core.cst$kw$schema),null));
});

/**
 * A value that must either be nil or satisfy schema
 */
schema.core.maybe = (function schema$core$maybe(schema__$1){
return (new schema.core.Maybe(schema__$1,null,null,null));
});

/**
* @constructor
 * @implements {cljs.core.IRecord}
 * @implements {cljs.core.IEquiv}
 * @implements {cljs.core.IHash}
 * @implements {cljs.core.ICollection}
 * @implements {schema.core.Schema}
 * @implements {cljs.core.ICounted}
 * @implements {cljs.core.ISeqable}
 * @implements {cljs.core.IMeta}
 * @implements {cljs.core.ICloneable}
 * @implements {cljs.core.IPrintWithWriter}
 * @implements {cljs.core.IIterable}
 * @implements {cljs.core.IWithMeta}
 * @implements {cljs.core.IAssociative}
 * @implements {cljs.core.IMap}
 * @implements {cljs.core.ILookup}
*/
schema.core.NamedSchema = (function (schema,name,__meta,__extmap,__hash){
this.schema = schema;
this.name = name;
this.__meta = __meta;
this.__extmap = __extmap;
this.__hash = __hash;
this.cljs$lang$protocol_mask$partition0$ = 2229667594;
this.cljs$lang$protocol_mask$partition1$ = 8192;
})
schema.core.NamedSchema.prototype.cljs$core$ILookup$_lookup$arity$2 = (function (this__5126__auto__,k__5127__auto__){
var self__ = this;
var this__5126__auto____$1 = this;
return cljs.core._lookup.cljs$core$IFn$_invoke$arity$3(this__5126__auto____$1,k__5127__auto__,null);
});

schema.core.NamedSchema.prototype.cljs$core$ILookup$_lookup$arity$3 = (function (this__5128__auto__,k27232,else__5129__auto__){
var self__ = this;
var this__5128__auto____$1 = this;
var G__27234 = (((k27232 instanceof cljs.core.Keyword))?k27232.fqn:null);
switch (G__27234) {
case "schema":
return self__.schema;

break;
case "name":
return self__.name;

break;
default:
return cljs.core.get.cljs$core$IFn$_invoke$arity$3(self__.__extmap,k27232,else__5129__auto__);

}
});

schema.core.NamedSchema.prototype.cljs$core$IPrintWithWriter$_pr_writer$arity$3 = (function (this__5140__auto__,writer__5141__auto__,opts__5142__auto__){
var self__ = this;
var this__5140__auto____$1 = this;
var pr_pair__5143__auto__ = ((function (this__5140__auto____$1){
return (function (keyval__5144__auto__){
return cljs.core.pr_sequential_writer(writer__5141__auto__,cljs.core.pr_writer,""," ","",opts__5142__auto__,keyval__5144__auto__);
});})(this__5140__auto____$1))
;
return cljs.core.pr_sequential_writer(writer__5141__auto__,pr_pair__5143__auto__,"#schema.core.NamedSchema{",", ","}",opts__5142__auto__,cljs.core.concat.cljs$core$IFn$_invoke$arity$2(new cljs.core.PersistentVector(null, 2, 5, cljs.core.PersistentVector.EMPTY_NODE, [(new cljs.core.PersistentVector(null,2,(5),cljs.core.PersistentVector.EMPTY_NODE,[cljs.core.cst$kw$schema,self__.schema],null)),(new cljs.core.PersistentVector(null,2,(5),cljs.core.PersistentVector.EMPTY_NODE,[cljs.core.cst$kw$name,self__.name],null))], null),self__.__extmap));
});

schema.core.NamedSchema.prototype.cljs$core$IIterable$ = true;

schema.core.NamedSchema.prototype.cljs$core$IIterable$_iterator$arity$1 = (function (G__27231){
var self__ = this;
var G__27231__$1 = this;
return (new cljs.core.RecordIter((0),G__27231__$1,2,new cljs.core.PersistentVector(null, 2, 5, cljs.core.PersistentVector.EMPTY_NODE, [cljs.core.cst$kw$schema,cljs.core.cst$kw$name], null),cljs.core._iterator(self__.__extmap)));
});

schema.core.NamedSchema.prototype.cljs$core$IMeta$_meta$arity$1 = (function (this__5124__auto__){
var self__ = this;
var this__5124__auto____$1 = this;
return self__.__meta;
});

schema.core.NamedSchema.prototype.cljs$core$ICloneable$_clone$arity$1 = (function (this__5120__auto__){
var self__ = this;
var this__5120__auto____$1 = this;
return (new schema.core.NamedSchema(self__.schema,self__.name,self__.__meta,self__.__extmap,self__.__hash));
});

schema.core.NamedSchema.prototype.cljs$core$ICounted$_count$arity$1 = (function (this__5130__auto__){
var self__ = this;
var this__5130__auto____$1 = this;
return (2 + cljs.core.count(self__.__extmap));
});

schema.core.NamedSchema.prototype.cljs$core$IHash$_hash$arity$1 = (function (this__5121__auto__){
var self__ = this;
var this__5121__auto____$1 = this;
var h__4947__auto__ = self__.__hash;
if(!((h__4947__auto__ == null))){
return h__4947__auto__;
} else {
var h__4947__auto____$1 = cljs.core.hash_imap(this__5121__auto____$1);
self__.__hash = h__4947__auto____$1;

return h__4947__auto____$1;
}
});

schema.core.NamedSchema.prototype.cljs$core$IEquiv$_equiv$arity$2 = (function (this__5122__auto__,other__5123__auto__){
var self__ = this;
var this__5122__auto____$1 = this;
if(cljs.core.truth_((function (){var and__4500__auto__ = other__5123__auto__;
if(cljs.core.truth_(and__4500__auto__)){
var and__4500__auto____$1 = (this__5122__auto____$1.constructor === other__5123__auto__.constructor);
if(and__4500__auto____$1){
return cljs.core.equiv_map(this__5122__auto____$1,other__5123__auto__);
} else {
return and__4500__auto____$1;
}
} else {
return and__4500__auto__;
}
})())){
return true;
} else {
return false;
}
});

schema.core.NamedSchema.prototype.cljs$core$IMap$_dissoc$arity$2 = (function (this__5135__auto__,k__5136__auto__){
var self__ = this;
var this__5135__auto____$1 = this;
if(cljs.core.contains_QMARK_(new cljs.core.PersistentHashSet(null, new cljs.core.PersistentArrayMap(null, 2, [cljs.core.cst$kw$schema,null,cljs.core.cst$kw$name,null], null), null),k__5136__auto__)){
return cljs.core.dissoc.cljs$core$IFn$_invoke$arity$2(cljs.core.with_meta(cljs.core.into.cljs$core$IFn$_invoke$arity$2(cljs.core.PersistentArrayMap.EMPTY,this__5135__auto____$1),self__.__meta),k__5136__auto__);
} else {
return (new schema.core.NamedSchema(self__.schema,self__.name,self__.__meta,cljs.core.not_empty(cljs.core.dissoc.cljs$core$IFn$_invoke$arity$2(self__.__extmap,k__5136__auto__)),null));
}
});

schema.core.NamedSchema.prototype.cljs$core$IAssociative$_assoc$arity$3 = (function (this__5133__auto__,k__5134__auto__,G__27231){
var self__ = this;
var this__5133__auto____$1 = this;
var pred__27235 = cljs.core.keyword_identical_QMARK_;
var expr__27236 = k__5134__auto__;
if(cljs.core.truth_((pred__27235.cljs$core$IFn$_invoke$arity$2 ? pred__27235.cljs$core$IFn$_invoke$arity$2(cljs.core.cst$kw$schema,expr__27236) : pred__27235.call(null,cljs.core.cst$kw$schema,expr__27236)))){
return (new schema.core.NamedSchema(G__27231,self__.name,self__.__meta,self__.__extmap,null));
} else {
if(cljs.core.truth_((pred__27235.cljs$core$IFn$_invoke$arity$2 ? pred__27235.cljs$core$IFn$_invoke$arity$2(cljs.core.cst$kw$name,expr__27236) : pred__27235.call(null,cljs.core.cst$kw$name,expr__27236)))){
return (new schema.core.NamedSchema(self__.schema,G__27231,self__.__meta,self__.__extmap,null));
} else {
return (new schema.core.NamedSchema(self__.schema,self__.name,self__.__meta,cljs.core.assoc.cljs$core$IFn$_invoke$arity$3(self__.__extmap,k__5134__auto__,G__27231),null));
}
}
});

schema.core.NamedSchema.prototype.cljs$core$ISeqable$_seq$arity$1 = (function (this__5138__auto__){
var self__ = this;
var this__5138__auto____$1 = this;
return cljs.core.seq(cljs.core.concat.cljs$core$IFn$_invoke$arity$2(new cljs.core.PersistentVector(null, 2, 5, cljs.core.PersistentVector.EMPTY_NODE, [(new cljs.core.PersistentVector(null,2,(5),cljs.core.PersistentVector.EMPTY_NODE,[cljs.core.cst$kw$schema,self__.schema],null)),(new cljs.core.PersistentVector(null,2,(5),cljs.core.PersistentVector.EMPTY_NODE,[cljs.core.cst$kw$name,self__.name],null))], null),self__.__extmap));
});

schema.core.NamedSchema.prototype.cljs$core$IWithMeta$_with_meta$arity$2 = (function (this__5125__auto__,G__27231){
var self__ = this;
var this__5125__auto____$1 = this;
return (new schema.core.NamedSchema(self__.schema,self__.name,G__27231,self__.__extmap,self__.__hash));
});

schema.core.NamedSchema.prototype.cljs$core$ICollection$_conj$arity$2 = (function (this__5131__auto__,entry__5132__auto__){
var self__ = this;
var this__5131__auto____$1 = this;
if(cljs.core.vector_QMARK_(entry__5132__auto__)){
return cljs.core._assoc(this__5131__auto____$1,cljs.core._nth.cljs$core$IFn$_invoke$arity$2(entry__5132__auto__,(0)),cljs.core._nth.cljs$core$IFn$_invoke$arity$2(entry__5132__auto__,(1)));
} else {
return cljs.core.reduce.cljs$core$IFn$_invoke$arity$3(cljs.core._conj,this__5131__auto____$1,entry__5132__auto__);
}
});

schema.core.NamedSchema.prototype.schema$core$Schema$ = true;

schema.core.NamedSchema.prototype.schema$core$Schema$walker$arity$1 = (function (this$){
var self__ = this;
var this$__$1 = this;
var sub_walker = (schema.core.subschema_walker.cljs$core$IFn$_invoke$arity$1 ? schema.core.subschema_walker.cljs$core$IFn$_invoke$arity$1(self__.schema) : schema.core.subschema_walker.call(null,self__.schema));
return ((function (sub_walker,this$__$1){
return (function (x){
return schema.utils.wrap_error_name(self__.name,(sub_walker.cljs$core$IFn$_invoke$arity$1 ? sub_walker.cljs$core$IFn$_invoke$arity$1(x) : sub_walker.call(null,x)));
});
;})(sub_walker,this$__$1))
});

schema.core.NamedSchema.prototype.schema$core$Schema$explain$arity$1 = (function (this$){
var self__ = this;
var this$__$1 = this;
return cljs.core._conj(cljs.core._conj(cljs.core._conj(cljs.core.List.EMPTY,self__.name),schema.core.explain(self__.schema)),cljs.core.cst$sym$named);
});

schema.core.NamedSchema.getBasis = (function (){
return new cljs.core.PersistentVector(null, 2, 5, cljs.core.PersistentVector.EMPTY_NODE, [cljs.core.cst$sym$schema,cljs.core.cst$sym$name], null);
});

schema.core.NamedSchema.cljs$lang$type = true;

schema.core.NamedSchema.cljs$lang$ctorPrSeq = (function (this__5160__auto__){
return cljs.core._conj(cljs.core.List.EMPTY,"schema.core/NamedSchema");
});

schema.core.NamedSchema.cljs$lang$ctorPrWriter = (function (this__5160__auto__,writer__5161__auto__){
return cljs.core._write(writer__5161__auto__,"schema.core/NamedSchema");
});

schema.core.__GT_NamedSchema = (function schema$core$__GT_NamedSchema(schema__$1,name){
return (new schema.core.NamedSchema(schema__$1,name,null,null,null));
});

schema.core.map__GT_NamedSchema = (function schema$core$map__GT_NamedSchema(G__27233){
return (new schema.core.NamedSchema(cljs.core.cst$kw$schema.cljs$core$IFn$_invoke$arity$1(G__27233),cljs.core.cst$kw$name.cljs$core$IFn$_invoke$arity$1(G__27233),null,cljs.core.dissoc.cljs$core$IFn$_invoke$arity$variadic(G__27233,cljs.core.cst$kw$schema,cljs.core.array_seq([cljs.core.cst$kw$name], 0)),null));
});

/**
 * A value that must satisfy schema, and has a name for documentation purposes.
 */
schema.core.named = (function schema$core$named(schema__$1,name){
return (new schema.core.NamedSchema(schema__$1,name,null,null,null));
});

/**
* @constructor
 * @implements {cljs.core.IRecord}
 * @implements {cljs.core.IEquiv}
 * @implements {cljs.core.IHash}
 * @implements {cljs.core.ICollection}
 * @implements {schema.core.Schema}
 * @implements {cljs.core.ICounted}
 * @implements {cljs.core.ISeqable}
 * @implements {cljs.core.IMeta}
 * @implements {cljs.core.ICloneable}
 * @implements {cljs.core.IPrintWithWriter}
 * @implements {cljs.core.IIterable}
 * @implements {cljs.core.IWithMeta}
 * @implements {cljs.core.IAssociative}
 * @implements {cljs.core.IMap}
 * @implements {cljs.core.ILookup}
*/
schema.core.Either = (function (schemas,__meta,__extmap,__hash){
this.schemas = schemas;
this.__meta = __meta;
this.__extmap = __extmap;
this.__hash = __hash;
this.cljs$lang$protocol_mask$partition0$ = 2229667594;
this.cljs$lang$protocol_mask$partition1$ = 8192;
})
schema.core.Either.prototype.cljs$core$ILookup$_lookup$arity$2 = (function (this__5126__auto__,k__5127__auto__){
var self__ = this;
var this__5126__auto____$1 = this;
return cljs.core._lookup.cljs$core$IFn$_invoke$arity$3(this__5126__auto____$1,k__5127__auto__,null);
});

schema.core.Either.prototype.cljs$core$ILookup$_lookup$arity$3 = (function (this__5128__auto__,k27240,else__5129__auto__){
var self__ = this;
var this__5128__auto____$1 = this;
var G__27242 = (((k27240 instanceof cljs.core.Keyword))?k27240.fqn:null);
switch (G__27242) {
case "schemas":
return self__.schemas;

break;
default:
return cljs.core.get.cljs$core$IFn$_invoke$arity$3(self__.__extmap,k27240,else__5129__auto__);

}
});

schema.core.Either.prototype.cljs$core$IPrintWithWriter$_pr_writer$arity$3 = (function (this__5140__auto__,writer__5141__auto__,opts__5142__auto__){
var self__ = this;
var this__5140__auto____$1 = this;
var pr_pair__5143__auto__ = ((function (this__5140__auto____$1){
return (function (keyval__5144__auto__){
return cljs.core.pr_sequential_writer(writer__5141__auto__,cljs.core.pr_writer,""," ","",opts__5142__auto__,keyval__5144__auto__);
});})(this__5140__auto____$1))
;
return cljs.core.pr_sequential_writer(writer__5141__auto__,pr_pair__5143__auto__,"#schema.core.Either{",", ","}",opts__5142__auto__,cljs.core.concat.cljs$core$IFn$_invoke$arity$2(new cljs.core.PersistentVector(null, 1, 5, cljs.core.PersistentVector.EMPTY_NODE, [(new cljs.core.PersistentVector(null,2,(5),cljs.core.PersistentVector.EMPTY_NODE,[cljs.core.cst$kw$schemas,self__.schemas],null))], null),self__.__extmap));
});

schema.core.Either.prototype.cljs$core$IIterable$ = true;

schema.core.Either.prototype.cljs$core$IIterable$_iterator$arity$1 = (function (G__27239){
var self__ = this;
var G__27239__$1 = this;
return (new cljs.core.RecordIter((0),G__27239__$1,1,new cljs.core.PersistentVector(null, 1, 5, cljs.core.PersistentVector.EMPTY_NODE, [cljs.core.cst$kw$schemas], null),cljs.core._iterator(self__.__extmap)));
});

schema.core.Either.prototype.cljs$core$IMeta$_meta$arity$1 = (function (this__5124__auto__){
var self__ = this;
var this__5124__auto____$1 = this;
return self__.__meta;
});

schema.core.Either.prototype.cljs$core$ICloneable$_clone$arity$1 = (function (this__5120__auto__){
var self__ = this;
var this__5120__auto____$1 = this;
return (new schema.core.Either(self__.schemas,self__.__meta,self__.__extmap,self__.__hash));
});

schema.core.Either.prototype.cljs$core$ICounted$_count$arity$1 = (function (this__5130__auto__){
var self__ = this;
var this__5130__auto____$1 = this;
return (1 + cljs.core.count(self__.__extmap));
});

schema.core.Either.prototype.cljs$core$IHash$_hash$arity$1 = (function (this__5121__auto__){
var self__ = this;
var this__5121__auto____$1 = this;
var h__4947__auto__ = self__.__hash;
if(!((h__4947__auto__ == null))){
return h__4947__auto__;
} else {
var h__4947__auto____$1 = cljs.core.hash_imap(this__5121__auto____$1);
self__.__hash = h__4947__auto____$1;

return h__4947__auto____$1;
}
});

schema.core.Either.prototype.cljs$core$IEquiv$_equiv$arity$2 = (function (this__5122__auto__,other__5123__auto__){
var self__ = this;
var this__5122__auto____$1 = this;
if(cljs.core.truth_((function (){var and__4500__auto__ = other__5123__auto__;
if(cljs.core.truth_(and__4500__auto__)){
var and__4500__auto____$1 = (this__5122__auto____$1.constructor === other__5123__auto__.constructor);
if(and__4500__auto____$1){
return cljs.core.equiv_map(this__5122__auto____$1,other__5123__auto__);
} else {
return and__4500__auto____$1;
}
} else {
return and__4500__auto__;
}
})())){
return true;
} else {
return false;
}
});

schema.core.Either.prototype.cljs$core$IMap$_dissoc$arity$2 = (function (this__5135__auto__,k__5136__auto__){
var self__ = this;
var this__5135__auto____$1 = this;
if(cljs.core.contains_QMARK_(new cljs.core.PersistentHashSet(null, new cljs.core.PersistentArrayMap(null, 1, [cljs.core.cst$kw$schemas,null], null), null),k__5136__auto__)){
return cljs.core.dissoc.cljs$core$IFn$_invoke$arity$2(cljs.core.with_meta(cljs.core.into.cljs$core$IFn$_invoke$arity$2(cljs.core.PersistentArrayMap.EMPTY,this__5135__auto____$1),self__.__meta),k__5136__auto__);
} else {
return (new schema.core.Either(self__.schemas,self__.__meta,cljs.core.not_empty(cljs.core.dissoc.cljs$core$IFn$_invoke$arity$2(self__.__extmap,k__5136__auto__)),null));
}
});

schema.core.Either.prototype.cljs$core$IAssociative$_assoc$arity$3 = (function (this__5133__auto__,k__5134__auto__,G__27239){
var self__ = this;
var this__5133__auto____$1 = this;
var pred__27243 = cljs.core.keyword_identical_QMARK_;
var expr__27244 = k__5134__auto__;
if(cljs.core.truth_((pred__27243.cljs$core$IFn$_invoke$arity$2 ? pred__27243.cljs$core$IFn$_invoke$arity$2(cljs.core.cst$kw$schemas,expr__27244) : pred__27243.call(null,cljs.core.cst$kw$schemas,expr__27244)))){
return (new schema.core.Either(G__27239,self__.__meta,self__.__extmap,null));
} else {
return (new schema.core.Either(self__.schemas,self__.__meta,cljs.core.assoc.cljs$core$IFn$_invoke$arity$3(self__.__extmap,k__5134__auto__,G__27239),null));
}
});

schema.core.Either.prototype.cljs$core$ISeqable$_seq$arity$1 = (function (this__5138__auto__){
var self__ = this;
var this__5138__auto____$1 = this;
return cljs.core.seq(cljs.core.concat.cljs$core$IFn$_invoke$arity$2(new cljs.core.PersistentVector(null, 1, 5, cljs.core.PersistentVector.EMPTY_NODE, [(new cljs.core.PersistentVector(null,2,(5),cljs.core.PersistentVector.EMPTY_NODE,[cljs.core.cst$kw$schemas,self__.schemas],null))], null),self__.__extmap));
});

schema.core.Either.prototype.cljs$core$IWithMeta$_with_meta$arity$2 = (function (this__5125__auto__,G__27239){
var self__ = this;
var this__5125__auto____$1 = this;
return (new schema.core.Either(self__.schemas,G__27239,self__.__extmap,self__.__hash));
});

schema.core.Either.prototype.cljs$core$ICollection$_conj$arity$2 = (function (this__5131__auto__,entry__5132__auto__){
var self__ = this;
var this__5131__auto____$1 = this;
if(cljs.core.vector_QMARK_(entry__5132__auto__)){
return cljs.core._assoc(this__5131__auto____$1,cljs.core._nth.cljs$core$IFn$_invoke$arity$2(entry__5132__auto__,(0)),cljs.core._nth.cljs$core$IFn$_invoke$arity$2(entry__5132__auto__,(1)));
} else {
return cljs.core.reduce.cljs$core$IFn$_invoke$arity$3(cljs.core._conj,this__5131__auto____$1,entry__5132__auto__);
}
});

schema.core.Either.prototype.schema$core$Schema$ = true;

schema.core.Either.prototype.schema$core$Schema$walker$arity$1 = (function (this$){
var self__ = this;
var this$__$1 = this;
var sub_walkers = cljs.core.mapv.cljs$core$IFn$_invoke$arity$2(schema.core.subschema_walker,self__.schemas);
return ((function (sub_walkers,this$__$1){
return (function (x){
var sub_walkers__$1 = cljs.core.seq(sub_walkers);
while(true){
if(cljs.core.not(sub_walkers__$1)){
return schema.utils.error(schema.utils.__GT_ValidationError(this$__$1,x,(new cljs.core.Delay(((function (sub_walkers__$1,sub_walkers,this$__$1){
return (function (){
return cljs.core._conj(cljs.core._conj(cljs.core._conj(cljs.core.List.EMPTY,cljs.core.cst$sym$schemas),cljs.core._conj(cljs.core._conj(cljs.core._conj(cljs.core.List.EMPTY,schema.utils.value_name(x)),cljs.core.cst$sym$_PERCENT_),cljs.core.cst$sym$check)),cljs.core.cst$sym$some);
});})(sub_walkers__$1,sub_walkers,this$__$1))
,null)),null));
} else {
var res = cljs.core.first(sub_walkers__$1).call(null,x);
if(cljs.core.not(schema.utils.error_QMARK_(res))){
return res;
} else {
var G__27247 = cljs.core.next(sub_walkers__$1);
sub_walkers__$1 = G__27247;
continue;
}
}
break;
}
});
;})(sub_walkers,this$__$1))
});

schema.core.Either.prototype.schema$core$Schema$explain$arity$1 = (function (this$){
var self__ = this;
var this$__$1 = this;
return cljs.core.cons(cljs.core.cst$sym$either,cljs.core.map.cljs$core$IFn$_invoke$arity$2(schema.core.explain,self__.schemas));
});

schema.core.Either.getBasis = (function (){
return new cljs.core.PersistentVector(null, 1, 5, cljs.core.PersistentVector.EMPTY_NODE, [cljs.core.cst$sym$schemas], null);
});

schema.core.Either.cljs$lang$type = true;

schema.core.Either.cljs$lang$ctorPrSeq = (function (this__5160__auto__){
return cljs.core._conj(cljs.core.List.EMPTY,"schema.core/Either");
});

schema.core.Either.cljs$lang$ctorPrWriter = (function (this__5160__auto__,writer__5161__auto__){
return cljs.core._write(writer__5161__auto__,"schema.core/Either");
});

schema.core.__GT_Either = (function schema$core$__GT_Either(schemas){
return (new schema.core.Either(schemas,null,null,null));
});

schema.core.map__GT_Either = (function schema$core$map__GT_Either(G__27241){
return (new schema.core.Either(cljs.core.cst$kw$schemas.cljs$core$IFn$_invoke$arity$1(G__27241),null,cljs.core.dissoc.cljs$core$IFn$_invoke$arity$2(G__27241,cljs.core.cst$kw$schemas),null));
});

/**
 * A value that must satisfy at least one schema in schemas.
 */
schema.core.either = (function schema$core$either(var_args){
var args__5577__auto__ = [];
var len__5570__auto___27249 = arguments.length;
var i__5571__auto___27250 = (0);
while(true){
if((i__5571__auto___27250 < len__5570__auto___27249)){
args__5577__auto__.push((arguments[i__5571__auto___27250]));

var G__27251 = (i__5571__auto___27250 + (1));
i__5571__auto___27250 = G__27251;
continue;
} else {
}
break;
}

var argseq__5578__auto__ = ((((0) < args__5577__auto__.length))?(new cljs.core.IndexedSeq(args__5577__auto__.slice((0)),(0))):null);
return schema.core.either.cljs$core$IFn$_invoke$arity$variadic(argseq__5578__auto__);
});

schema.core.either.cljs$core$IFn$_invoke$arity$variadic = (function (schemas){
return (new schema.core.Either(schemas,null,null,null));
});

schema.core.either.cljs$lang$maxFixedArity = (0);

schema.core.either.cljs$lang$applyTo = (function (seq27248){
return schema.core.either.cljs$core$IFn$_invoke$arity$variadic(cljs.core.seq(seq27248));
});

/**
* @constructor
 * @implements {cljs.core.IRecord}
 * @implements {cljs.core.IEquiv}
 * @implements {cljs.core.IHash}
 * @implements {cljs.core.ICollection}
 * @implements {schema.core.Schema}
 * @implements {cljs.core.ICounted}
 * @implements {cljs.core.ISeqable}
 * @implements {cljs.core.IMeta}
 * @implements {cljs.core.ICloneable}
 * @implements {cljs.core.IPrintWithWriter}
 * @implements {cljs.core.IIterable}
 * @implements {cljs.core.IWithMeta}
 * @implements {cljs.core.IAssociative}
 * @implements {cljs.core.IMap}
 * @implements {cljs.core.ILookup}
*/
schema.core.Both = (function (schemas,__meta,__extmap,__hash){
this.schemas = schemas;
this.__meta = __meta;
this.__extmap = __extmap;
this.__hash = __hash;
this.cljs$lang$protocol_mask$partition0$ = 2229667594;
this.cljs$lang$protocol_mask$partition1$ = 8192;
})
schema.core.Both.prototype.cljs$core$ILookup$_lookup$arity$2 = (function (this__5126__auto__,k__5127__auto__){
var self__ = this;
var this__5126__auto____$1 = this;
return cljs.core._lookup.cljs$core$IFn$_invoke$arity$3(this__5126__auto____$1,k__5127__auto__,null);
});

schema.core.Both.prototype.cljs$core$ILookup$_lookup$arity$3 = (function (this__5128__auto__,k27253,else__5129__auto__){
var self__ = this;
var this__5128__auto____$1 = this;
var G__27255 = (((k27253 instanceof cljs.core.Keyword))?k27253.fqn:null);
switch (G__27255) {
case "schemas":
return self__.schemas;

break;
default:
return cljs.core.get.cljs$core$IFn$_invoke$arity$3(self__.__extmap,k27253,else__5129__auto__);

}
});

schema.core.Both.prototype.cljs$core$IPrintWithWriter$_pr_writer$arity$3 = (function (this__5140__auto__,writer__5141__auto__,opts__5142__auto__){
var self__ = this;
var this__5140__auto____$1 = this;
var pr_pair__5143__auto__ = ((function (this__5140__auto____$1){
return (function (keyval__5144__auto__){
return cljs.core.pr_sequential_writer(writer__5141__auto__,cljs.core.pr_writer,""," ","",opts__5142__auto__,keyval__5144__auto__);
});})(this__5140__auto____$1))
;
return cljs.core.pr_sequential_writer(writer__5141__auto__,pr_pair__5143__auto__,"#schema.core.Both{",", ","}",opts__5142__auto__,cljs.core.concat.cljs$core$IFn$_invoke$arity$2(new cljs.core.PersistentVector(null, 1, 5, cljs.core.PersistentVector.EMPTY_NODE, [(new cljs.core.PersistentVector(null,2,(5),cljs.core.PersistentVector.EMPTY_NODE,[cljs.core.cst$kw$schemas,self__.schemas],null))], null),self__.__extmap));
});

schema.core.Both.prototype.cljs$core$IIterable$ = true;

schema.core.Both.prototype.cljs$core$IIterable$_iterator$arity$1 = (function (G__27252){
var self__ = this;
var G__27252__$1 = this;
return (new cljs.core.RecordIter((0),G__27252__$1,1,new cljs.core.PersistentVector(null, 1, 5, cljs.core.PersistentVector.EMPTY_NODE, [cljs.core.cst$kw$schemas], null),cljs.core._iterator(self__.__extmap)));
});

schema.core.Both.prototype.cljs$core$IMeta$_meta$arity$1 = (function (this__5124__auto__){
var self__ = this;
var this__5124__auto____$1 = this;
return self__.__meta;
});

schema.core.Both.prototype.cljs$core$ICloneable$_clone$arity$1 = (function (this__5120__auto__){
var self__ = this;
var this__5120__auto____$1 = this;
return (new schema.core.Both(self__.schemas,self__.__meta,self__.__extmap,self__.__hash));
});

schema.core.Both.prototype.cljs$core$ICounted$_count$arity$1 = (function (this__5130__auto__){
var self__ = this;
var this__5130__auto____$1 = this;
return (1 + cljs.core.count(self__.__extmap));
});

schema.core.Both.prototype.cljs$core$IHash$_hash$arity$1 = (function (this__5121__auto__){
var self__ = this;
var this__5121__auto____$1 = this;
var h__4947__auto__ = self__.__hash;
if(!((h__4947__auto__ == null))){
return h__4947__auto__;
} else {
var h__4947__auto____$1 = cljs.core.hash_imap(this__5121__auto____$1);
self__.__hash = h__4947__auto____$1;

return h__4947__auto____$1;
}
});

schema.core.Both.prototype.cljs$core$IEquiv$_equiv$arity$2 = (function (this__5122__auto__,other__5123__auto__){
var self__ = this;
var this__5122__auto____$1 = this;
if(cljs.core.truth_((function (){var and__4500__auto__ = other__5123__auto__;
if(cljs.core.truth_(and__4500__auto__)){
var and__4500__auto____$1 = (this__5122__auto____$1.constructor === other__5123__auto__.constructor);
if(and__4500__auto____$1){
return cljs.core.equiv_map(this__5122__auto____$1,other__5123__auto__);
} else {
return and__4500__auto____$1;
}
} else {
return and__4500__auto__;
}
})())){
return true;
} else {
return false;
}
});

schema.core.Both.prototype.cljs$core$IMap$_dissoc$arity$2 = (function (this__5135__auto__,k__5136__auto__){
var self__ = this;
var this__5135__auto____$1 = this;
if(cljs.core.contains_QMARK_(new cljs.core.PersistentHashSet(null, new cljs.core.PersistentArrayMap(null, 1, [cljs.core.cst$kw$schemas,null], null), null),k__5136__auto__)){
return cljs.core.dissoc.cljs$core$IFn$_invoke$arity$2(cljs.core.with_meta(cljs.core.into.cljs$core$IFn$_invoke$arity$2(cljs.core.PersistentArrayMap.EMPTY,this__5135__auto____$1),self__.__meta),k__5136__auto__);
} else {
return (new schema.core.Both(self__.schemas,self__.__meta,cljs.core.not_empty(cljs.core.dissoc.cljs$core$IFn$_invoke$arity$2(self__.__extmap,k__5136__auto__)),null));
}
});

schema.core.Both.prototype.cljs$core$IAssociative$_assoc$arity$3 = (function (this__5133__auto__,k__5134__auto__,G__27252){
var self__ = this;
var this__5133__auto____$1 = this;
var pred__27256 = cljs.core.keyword_identical_QMARK_;
var expr__27257 = k__5134__auto__;
if(cljs.core.truth_((pred__27256.cljs$core$IFn$_invoke$arity$2 ? pred__27256.cljs$core$IFn$_invoke$arity$2(cljs.core.cst$kw$schemas,expr__27257) : pred__27256.call(null,cljs.core.cst$kw$schemas,expr__27257)))){
return (new schema.core.Both(G__27252,self__.__meta,self__.__extmap,null));
} else {
return (new schema.core.Both(self__.schemas,self__.__meta,cljs.core.assoc.cljs$core$IFn$_invoke$arity$3(self__.__extmap,k__5134__auto__,G__27252),null));
}
});

schema.core.Both.prototype.cljs$core$ISeqable$_seq$arity$1 = (function (this__5138__auto__){
var self__ = this;
var this__5138__auto____$1 = this;
return cljs.core.seq(cljs.core.concat.cljs$core$IFn$_invoke$arity$2(new cljs.core.PersistentVector(null, 1, 5, cljs.core.PersistentVector.EMPTY_NODE, [(new cljs.core.PersistentVector(null,2,(5),cljs.core.PersistentVector.EMPTY_NODE,[cljs.core.cst$kw$schemas,self__.schemas],null))], null),self__.__extmap));
});

schema.core.Both.prototype.cljs$core$IWithMeta$_with_meta$arity$2 = (function (this__5125__auto__,G__27252){
var self__ = this;
var this__5125__auto____$1 = this;
return (new schema.core.Both(self__.schemas,G__27252,self__.__extmap,self__.__hash));
});

schema.core.Both.prototype.cljs$core$ICollection$_conj$arity$2 = (function (this__5131__auto__,entry__5132__auto__){
var self__ = this;
var this__5131__auto____$1 = this;
if(cljs.core.vector_QMARK_(entry__5132__auto__)){
return cljs.core._assoc(this__5131__auto____$1,cljs.core._nth.cljs$core$IFn$_invoke$arity$2(entry__5132__auto__,(0)),cljs.core._nth.cljs$core$IFn$_invoke$arity$2(entry__5132__auto__,(1)));
} else {
return cljs.core.reduce.cljs$core$IFn$_invoke$arity$3(cljs.core._conj,this__5131__auto____$1,entry__5132__auto__);
}
});

schema.core.Both.prototype.schema$core$Schema$ = true;

schema.core.Both.prototype.schema$core$Schema$walker$arity$1 = (function (this$){
var self__ = this;
var this$__$1 = this;
var sub_walkers = cljs.core.mapv.cljs$core$IFn$_invoke$arity$2(schema.core.subschema_walker,self__.schemas);
return ((function (sub_walkers,this$__$1){
return (function (x){
return cljs.core.reduce.cljs$core$IFn$_invoke$arity$3(((function (sub_walkers,this$__$1){
return (function (x__$1,sub_walker){
if(cljs.core.truth_(schema.utils.error_QMARK_(x__$1))){
return x__$1;
} else {
return (sub_walker.cljs$core$IFn$_invoke$arity$1 ? sub_walker.cljs$core$IFn$_invoke$arity$1(x__$1) : sub_walker.call(null,x__$1));
}
});})(sub_walkers,this$__$1))
,x,sub_walkers);
});
;})(sub_walkers,this$__$1))
});

schema.core.Both.prototype.schema$core$Schema$explain$arity$1 = (function (this$){
var self__ = this;
var this$__$1 = this;
return cljs.core.cons(cljs.core.cst$sym$both,cljs.core.map.cljs$core$IFn$_invoke$arity$2(schema.core.explain,self__.schemas));
});

schema.core.Both.getBasis = (function (){
return new cljs.core.PersistentVector(null, 1, 5, cljs.core.PersistentVector.EMPTY_NODE, [cljs.core.cst$sym$schemas], null);
});

schema.core.Both.cljs$lang$type = true;

schema.core.Both.cljs$lang$ctorPrSeq = (function (this__5160__auto__){
return cljs.core._conj(cljs.core.List.EMPTY,"schema.core/Both");
});

schema.core.Both.cljs$lang$ctorPrWriter = (function (this__5160__auto__,writer__5161__auto__){
return cljs.core._write(writer__5161__auto__,"schema.core/Both");
});

schema.core.__GT_Both = (function schema$core$__GT_Both(schemas){
return (new schema.core.Both(schemas,null,null,null));
});

schema.core.map__GT_Both = (function schema$core$map__GT_Both(G__27254){
return (new schema.core.Both(cljs.core.cst$kw$schemas.cljs$core$IFn$_invoke$arity$1(G__27254),null,cljs.core.dissoc.cljs$core$IFn$_invoke$arity$2(G__27254,cljs.core.cst$kw$schemas),null));
});

/**
 * A value that must satisfy every schema in schemas.
 */
schema.core.both = (function schema$core$both(var_args){
var args__5577__auto__ = [];
var len__5570__auto___27261 = arguments.length;
var i__5571__auto___27262 = (0);
while(true){
if((i__5571__auto___27262 < len__5570__auto___27261)){
args__5577__auto__.push((arguments[i__5571__auto___27262]));

var G__27263 = (i__5571__auto___27262 + (1));
i__5571__auto___27262 = G__27263;
continue;
} else {
}
break;
}

var argseq__5578__auto__ = ((((0) < args__5577__auto__.length))?(new cljs.core.IndexedSeq(args__5577__auto__.slice((0)),(0))):null);
return schema.core.both.cljs$core$IFn$_invoke$arity$variadic(argseq__5578__auto__);
});

schema.core.both.cljs$core$IFn$_invoke$arity$variadic = (function (schemas){
return (new schema.core.Both(schemas,null,null,null));
});

schema.core.both.cljs$lang$maxFixedArity = (0);

schema.core.both.cljs$lang$applyTo = (function (seq27260){
return schema.core.both.cljs$core$IFn$_invoke$arity$variadic(cljs.core.seq(seq27260));
});

/**
* @constructor
 * @implements {cljs.core.IRecord}
 * @implements {cljs.core.IEquiv}
 * @implements {cljs.core.IHash}
 * @implements {cljs.core.ICollection}
 * @implements {schema.core.Schema}
 * @implements {cljs.core.ICounted}
 * @implements {cljs.core.ISeqable}
 * @implements {cljs.core.IMeta}
 * @implements {cljs.core.ICloneable}
 * @implements {cljs.core.IPrintWithWriter}
 * @implements {cljs.core.IIterable}
 * @implements {cljs.core.IWithMeta}
 * @implements {cljs.core.IAssociative}
 * @implements {cljs.core.IMap}
 * @implements {cljs.core.ILookup}
*/
schema.core.ConditionalSchema = (function (preds_and_schemas,__meta,__extmap,__hash){
this.preds_and_schemas = preds_and_schemas;
this.__meta = __meta;
this.__extmap = __extmap;
this.__hash = __hash;
this.cljs$lang$protocol_mask$partition0$ = 2229667594;
this.cljs$lang$protocol_mask$partition1$ = 8192;
})
schema.core.ConditionalSchema.prototype.cljs$core$ILookup$_lookup$arity$2 = (function (this__5126__auto__,k__5127__auto__){
var self__ = this;
var this__5126__auto____$1 = this;
return cljs.core._lookup.cljs$core$IFn$_invoke$arity$3(this__5126__auto____$1,k__5127__auto__,null);
});

schema.core.ConditionalSchema.prototype.cljs$core$ILookup$_lookup$arity$3 = (function (this__5128__auto__,k27265,else__5129__auto__){
var self__ = this;
var this__5128__auto____$1 = this;
var G__27267 = (((k27265 instanceof cljs.core.Keyword))?k27265.fqn:null);
switch (G__27267) {
case "preds-and-schemas":
return self__.preds_and_schemas;

break;
default:
return cljs.core.get.cljs$core$IFn$_invoke$arity$3(self__.__extmap,k27265,else__5129__auto__);

}
});

schema.core.ConditionalSchema.prototype.cljs$core$IPrintWithWriter$_pr_writer$arity$3 = (function (this__5140__auto__,writer__5141__auto__,opts__5142__auto__){
var self__ = this;
var this__5140__auto____$1 = this;
var pr_pair__5143__auto__ = ((function (this__5140__auto____$1){
return (function (keyval__5144__auto__){
return cljs.core.pr_sequential_writer(writer__5141__auto__,cljs.core.pr_writer,""," ","",opts__5142__auto__,keyval__5144__auto__);
});})(this__5140__auto____$1))
;
return cljs.core.pr_sequential_writer(writer__5141__auto__,pr_pair__5143__auto__,"#schema.core.ConditionalSchema{",", ","}",opts__5142__auto__,cljs.core.concat.cljs$core$IFn$_invoke$arity$2(new cljs.core.PersistentVector(null, 1, 5, cljs.core.PersistentVector.EMPTY_NODE, [(new cljs.core.PersistentVector(null,2,(5),cljs.core.PersistentVector.EMPTY_NODE,[cljs.core.cst$kw$preds_DASH_and_DASH_schemas,self__.preds_and_schemas],null))], null),self__.__extmap));
});

schema.core.ConditionalSchema.prototype.cljs$core$IIterable$ = true;

schema.core.ConditionalSchema.prototype.cljs$core$IIterable$_iterator$arity$1 = (function (G__27264){
var self__ = this;
var G__27264__$1 = this;
return (new cljs.core.RecordIter((0),G__27264__$1,1,new cljs.core.PersistentVector(null, 1, 5, cljs.core.PersistentVector.EMPTY_NODE, [cljs.core.cst$kw$preds_DASH_and_DASH_schemas], null),cljs.core._iterator(self__.__extmap)));
});

schema.core.ConditionalSchema.prototype.cljs$core$IMeta$_meta$arity$1 = (function (this__5124__auto__){
var self__ = this;
var this__5124__auto____$1 = this;
return self__.__meta;
});

schema.core.ConditionalSchema.prototype.cljs$core$ICloneable$_clone$arity$1 = (function (this__5120__auto__){
var self__ = this;
var this__5120__auto____$1 = this;
return (new schema.core.ConditionalSchema(self__.preds_and_schemas,self__.__meta,self__.__extmap,self__.__hash));
});

schema.core.ConditionalSchema.prototype.cljs$core$ICounted$_count$arity$1 = (function (this__5130__auto__){
var self__ = this;
var this__5130__auto____$1 = this;
return (1 + cljs.core.count(self__.__extmap));
});

schema.core.ConditionalSchema.prototype.cljs$core$IHash$_hash$arity$1 = (function (this__5121__auto__){
var self__ = this;
var this__5121__auto____$1 = this;
var h__4947__auto__ = self__.__hash;
if(!((h__4947__auto__ == null))){
return h__4947__auto__;
} else {
var h__4947__auto____$1 = cljs.core.hash_imap(this__5121__auto____$1);
self__.__hash = h__4947__auto____$1;

return h__4947__auto____$1;
}
});

schema.core.ConditionalSchema.prototype.cljs$core$IEquiv$_equiv$arity$2 = (function (this__5122__auto__,other__5123__auto__){
var self__ = this;
var this__5122__auto____$1 = this;
if(cljs.core.truth_((function (){var and__4500__auto__ = other__5123__auto__;
if(cljs.core.truth_(and__4500__auto__)){
var and__4500__auto____$1 = (this__5122__auto____$1.constructor === other__5123__auto__.constructor);
if(and__4500__auto____$1){
return cljs.core.equiv_map(this__5122__auto____$1,other__5123__auto__);
} else {
return and__4500__auto____$1;
}
} else {
return and__4500__auto__;
}
})())){
return true;
} else {
return false;
}
});

schema.core.ConditionalSchema.prototype.cljs$core$IMap$_dissoc$arity$2 = (function (this__5135__auto__,k__5136__auto__){
var self__ = this;
var this__5135__auto____$1 = this;
if(cljs.core.contains_QMARK_(new cljs.core.PersistentHashSet(null, new cljs.core.PersistentArrayMap(null, 1, [cljs.core.cst$kw$preds_DASH_and_DASH_schemas,null], null), null),k__5136__auto__)){
return cljs.core.dissoc.cljs$core$IFn$_invoke$arity$2(cljs.core.with_meta(cljs.core.into.cljs$core$IFn$_invoke$arity$2(cljs.core.PersistentArrayMap.EMPTY,this__5135__auto____$1),self__.__meta),k__5136__auto__);
} else {
return (new schema.core.ConditionalSchema(self__.preds_and_schemas,self__.__meta,cljs.core.not_empty(cljs.core.dissoc.cljs$core$IFn$_invoke$arity$2(self__.__extmap,k__5136__auto__)),null));
}
});

schema.core.ConditionalSchema.prototype.cljs$core$IAssociative$_assoc$arity$3 = (function (this__5133__auto__,k__5134__auto__,G__27264){
var self__ = this;
var this__5133__auto____$1 = this;
var pred__27268 = cljs.core.keyword_identical_QMARK_;
var expr__27269 = k__5134__auto__;
if(cljs.core.truth_((pred__27268.cljs$core$IFn$_invoke$arity$2 ? pred__27268.cljs$core$IFn$_invoke$arity$2(cljs.core.cst$kw$preds_DASH_and_DASH_schemas,expr__27269) : pred__27268.call(null,cljs.core.cst$kw$preds_DASH_and_DASH_schemas,expr__27269)))){
return (new schema.core.ConditionalSchema(G__27264,self__.__meta,self__.__extmap,null));
} else {
return (new schema.core.ConditionalSchema(self__.preds_and_schemas,self__.__meta,cljs.core.assoc.cljs$core$IFn$_invoke$arity$3(self__.__extmap,k__5134__auto__,G__27264),null));
}
});

schema.core.ConditionalSchema.prototype.cljs$core$ISeqable$_seq$arity$1 = (function (this__5138__auto__){
var self__ = this;
var this__5138__auto____$1 = this;
return cljs.core.seq(cljs.core.concat.cljs$core$IFn$_invoke$arity$2(new cljs.core.PersistentVector(null, 1, 5, cljs.core.PersistentVector.EMPTY_NODE, [(new cljs.core.PersistentVector(null,2,(5),cljs.core.PersistentVector.EMPTY_NODE,[cljs.core.cst$kw$preds_DASH_and_DASH_schemas,self__.preds_and_schemas],null))], null),self__.__extmap));
});

schema.core.ConditionalSchema.prototype.cljs$core$IWithMeta$_with_meta$arity$2 = (function (this__5125__auto__,G__27264){
var self__ = this;
var this__5125__auto____$1 = this;
return (new schema.core.ConditionalSchema(self__.preds_and_schemas,G__27264,self__.__extmap,self__.__hash));
});

schema.core.ConditionalSchema.prototype.cljs$core$ICollection$_conj$arity$2 = (function (this__5131__auto__,entry__5132__auto__){
var self__ = this;
var this__5131__auto____$1 = this;
if(cljs.core.vector_QMARK_(entry__5132__auto__)){
return cljs.core._assoc(this__5131__auto____$1,cljs.core._nth.cljs$core$IFn$_invoke$arity$2(entry__5132__auto__,(0)),cljs.core._nth.cljs$core$IFn$_invoke$arity$2(entry__5132__auto__,(1)));
} else {
return cljs.core.reduce.cljs$core$IFn$_invoke$arity$3(cljs.core._conj,this__5131__auto____$1,entry__5132__auto__);
}
});

schema.core.ConditionalSchema.prototype.schema$core$Schema$ = true;

schema.core.ConditionalSchema.prototype.schema$core$Schema$walker$arity$1 = (function (this$){
var self__ = this;
var this$__$1 = this;
var preds_and_walkers = cljs.core.mapv.cljs$core$IFn$_invoke$arity$2(((function (this$__$1){
return (function (p__27271){
var vec__27272 = p__27271;
var pred = cljs.core.nth.cljs$core$IFn$_invoke$arity$3(vec__27272,(0),null);
var schema__$1 = cljs.core.nth.cljs$core$IFn$_invoke$arity$3(vec__27272,(1),null);
return new cljs.core.PersistentVector(null, 2, 5, cljs.core.PersistentVector.EMPTY_NODE, [pred,(schema.core.subschema_walker.cljs$core$IFn$_invoke$arity$1 ? schema.core.subschema_walker.cljs$core$IFn$_invoke$arity$1(schema__$1) : schema.core.subschema_walker.call(null,schema__$1))], null);
});})(this$__$1))
,self__.preds_and_schemas);
return ((function (preds_and_walkers,this$__$1){
return (function (x){
var temp__4423__auto__ = cljs.core.first(cljs.core.filter.cljs$core$IFn$_invoke$arity$2(((function (preds_and_walkers,this$__$1){
return (function (p__27273){
var vec__27274 = p__27273;
var pred = cljs.core.nth.cljs$core$IFn$_invoke$arity$3(vec__27274,(0),null);
return (pred.cljs$core$IFn$_invoke$arity$1 ? pred.cljs$core$IFn$_invoke$arity$1(x) : pred.call(null,x));
});})(preds_and_walkers,this$__$1))
,preds_and_walkers));
if(cljs.core.truth_(temp__4423__auto__)){
var vec__27275 = temp__4423__auto__;
var _ = cljs.core.nth.cljs$core$IFn$_invoke$arity$3(vec__27275,(0),null);
var match = cljs.core.nth.cljs$core$IFn$_invoke$arity$3(vec__27275,(1),null);
return (match.cljs$core$IFn$_invoke$arity$1 ? match.cljs$core$IFn$_invoke$arity$1(x) : match.call(null,x));
} else {
return schema.utils.error(schema.utils.__GT_ValidationError(this$__$1,x,(new cljs.core.Delay(((function (temp__4423__auto__,preds_and_walkers,this$__$1){
return (function (){
return cljs.core._conj(cljs.core._conj(cljs.core.List.EMPTY,schema.utils.value_name(x)),cljs.core.cst$sym$matches_DASH_some_DASH_condition_QMARK_);
});})(temp__4423__auto__,preds_and_walkers,this$__$1))
,null)),null));
}
});
;})(preds_and_walkers,this$__$1))
});

schema.core.ConditionalSchema.prototype.schema$core$Schema$explain$arity$1 = (function (this$){
var self__ = this;
var this$__$1 = this;
return cljs.core.cons(cljs.core.cst$sym$conditional,cljs.core.mapcat.cljs$core$IFn$_invoke$arity$variadic(((function (this$__$1){
return (function (p__27276){
var vec__27277 = p__27276;
var pred = cljs.core.nth.cljs$core$IFn$_invoke$arity$3(vec__27277,(0),null);
var schema__$1 = cljs.core.nth.cljs$core$IFn$_invoke$arity$3(vec__27277,(1),null);
return new cljs.core.PersistentVector(null, 2, 5, cljs.core.PersistentVector.EMPTY_NODE, [pred,schema.core.explain(schema__$1)], null);
});})(this$__$1))
,cljs.core.array_seq([self__.preds_and_schemas], 0)));
});

schema.core.ConditionalSchema.getBasis = (function (){
return new cljs.core.PersistentVector(null, 1, 5, cljs.core.PersistentVector.EMPTY_NODE, [cljs.core.cst$sym$preds_DASH_and_DASH_schemas], null);
});

schema.core.ConditionalSchema.cljs$lang$type = true;

schema.core.ConditionalSchema.cljs$lang$ctorPrSeq = (function (this__5160__auto__){
return cljs.core._conj(cljs.core.List.EMPTY,"schema.core/ConditionalSchema");
});

schema.core.ConditionalSchema.cljs$lang$ctorPrWriter = (function (this__5160__auto__,writer__5161__auto__){
return cljs.core._write(writer__5161__auto__,"schema.core/ConditionalSchema");
});

schema.core.__GT_ConditionalSchema = (function schema$core$__GT_ConditionalSchema(preds_and_schemas){
return (new schema.core.ConditionalSchema(preds_and_schemas,null,null,null));
});

schema.core.map__GT_ConditionalSchema = (function schema$core$map__GT_ConditionalSchema(G__27266){
return (new schema.core.ConditionalSchema(cljs.core.cst$kw$preds_DASH_and_DASH_schemas.cljs$core$IFn$_invoke$arity$1(G__27266),null,cljs.core.dissoc.cljs$core$IFn$_invoke$arity$2(G__27266,cljs.core.cst$kw$preds_DASH_and_DASH_schemas),null));
});

/**
 * Define a conditional schema.  Takes args like cond,
 * (conditional pred1 schema1 pred2 schema2 ...),
 * and checks the first schema where pred is true on the value.
 * Unlike cond, throws if the value does not match any condition.
 * :else may be used as a final condition in the place of (constantly true).
 * More efficient than either, since only one schema must be checked.
 */
schema.core.conditional = (function schema$core$conditional(var_args){
var args__5577__auto__ = [];
var len__5570__auto___27290 = arguments.length;
var i__5571__auto___27291 = (0);
while(true){
if((i__5571__auto___27291 < len__5570__auto___27290)){
args__5577__auto__.push((arguments[i__5571__auto___27291]));

var G__27292 = (i__5571__auto___27291 + (1));
i__5571__auto___27291 = G__27292;
continue;
} else {
}
break;
}

var argseq__5578__auto__ = ((((0) < args__5577__auto__.length))?(new cljs.core.IndexedSeq(args__5577__auto__.slice((0)),(0))):null);
return schema.core.conditional.cljs$core$IFn$_invoke$arity$variadic(argseq__5578__auto__);
});

schema.core.conditional.cljs$core$IFn$_invoke$arity$variadic = (function (preds_and_schemas){
if((cljs.core.seq(preds_and_schemas)) && (cljs.core.even_QMARK_(cljs.core.count(preds_and_schemas)))){
} else {
throw (new Error(schema.utils.format_STAR_.cljs$core$IFn$_invoke$arity$variadic("Expected even, nonzero number of args; got %s",cljs.core.array_seq([cljs.core.count(preds_and_schemas)], 0))));
}

return (new schema.core.ConditionalSchema((function (){var iter__5284__auto__ = (function schema$core$iter__27280(s__27281){
return (new cljs.core.LazySeq(null,(function (){
var s__27281__$1 = s__27281;
while(true){
var temp__4425__auto__ = cljs.core.seq(s__27281__$1);
if(temp__4425__auto__){
var s__27281__$2 = temp__4425__auto__;
if(cljs.core.chunked_seq_QMARK_(s__27281__$2)){
var c__5282__auto__ = cljs.core.chunk_first(s__27281__$2);
var size__5283__auto__ = cljs.core.count(c__5282__auto__);
var b__27283 = cljs.core.chunk_buffer(size__5283__auto__);
if((function (){var i__27282 = (0);
while(true){
if((i__27282 < size__5283__auto__)){
var vec__27288 = cljs.core._nth.cljs$core$IFn$_invoke$arity$2(c__5282__auto__,i__27282);
var pred = cljs.core.nth.cljs$core$IFn$_invoke$arity$3(vec__27288,(0),null);
var schema__$1 = cljs.core.nth.cljs$core$IFn$_invoke$arity$3(vec__27288,(1),null);
cljs.core.chunk_append(b__27283,new cljs.core.PersistentVector(null, 2, 5, cljs.core.PersistentVector.EMPTY_NODE, [((cljs.core._EQ_.cljs$core$IFn$_invoke$arity$2(pred,cljs.core.cst$kw$else))?cljs.core.constantly(true):pred),schema__$1], null));

var G__27293 = (i__27282 + (1));
i__27282 = G__27293;
continue;
} else {
return true;
}
break;
}
})()){
return cljs.core.chunk_cons(cljs.core.chunk(b__27283),schema$core$iter__27280(cljs.core.chunk_rest(s__27281__$2)));
} else {
return cljs.core.chunk_cons(cljs.core.chunk(b__27283),null);
}
} else {
var vec__27289 = cljs.core.first(s__27281__$2);
var pred = cljs.core.nth.cljs$core$IFn$_invoke$arity$3(vec__27289,(0),null);
var schema__$1 = cljs.core.nth.cljs$core$IFn$_invoke$arity$3(vec__27289,(1),null);
return cljs.core.cons(new cljs.core.PersistentVector(null, 2, 5, cljs.core.PersistentVector.EMPTY_NODE, [((cljs.core._EQ_.cljs$core$IFn$_invoke$arity$2(pred,cljs.core.cst$kw$else))?cljs.core.constantly(true):pred),schema__$1], null),schema$core$iter__27280(cljs.core.rest(s__27281__$2)));
}
} else {
return null;
}
break;
}
}),null,null));
});
return iter__5284__auto__(cljs.core.partition.cljs$core$IFn$_invoke$arity$2((2),preds_and_schemas));
})(),null,null,null));
});

schema.core.conditional.cljs$lang$maxFixedArity = (0);

schema.core.conditional.cljs$lang$applyTo = (function (seq27279){
return schema.core.conditional.cljs$core$IFn$_invoke$arity$variadic(cljs.core.seq(seq27279));
});
/**
 * if the predicate returns truthy, use the if-schema, otherwise use the else-schema
 */
schema.core.if$ = (function schema$core$if(pred,if_schema,else_schema){
return schema.core.conditional.cljs$core$IFn$_invoke$arity$variadic(cljs.core.array_seq([pred,if_schema,cljs.core.constantly(true),else_schema], 0));
});
/**
 * A sentinel value representing missing portions of the input data.
 */
schema.core._PLUS_missing_PLUS_ = cljs.core.cst$kw$schema$core_SLASH_missing;

/**
* @constructor
 * @implements {cljs.core.IRecord}
 * @implements {cljs.core.IEquiv}
 * @implements {cljs.core.IHash}
 * @implements {cljs.core.ICollection}
 * @implements {cljs.core.ICounted}
 * @implements {cljs.core.ISeqable}
 * @implements {cljs.core.IMeta}
 * @implements {cljs.core.ICloneable}
 * @implements {cljs.core.IPrintWithWriter}
 * @implements {cljs.core.IIterable}
 * @implements {cljs.core.IWithMeta}
 * @implements {cljs.core.IAssociative}
 * @implements {cljs.core.IMap}
 * @implements {cljs.core.ILookup}
*/
schema.core.RequiredKey = (function (k,__meta,__extmap,__hash){
this.k = k;
this.__meta = __meta;
this.__extmap = __extmap;
this.__hash = __hash;
this.cljs$lang$protocol_mask$partition0$ = 2229667594;
this.cljs$lang$protocol_mask$partition1$ = 8192;
})
schema.core.RequiredKey.prototype.cljs$core$ILookup$_lookup$arity$2 = (function (this__5126__auto__,k__5127__auto__){
var self__ = this;
var this__5126__auto____$1 = this;
return cljs.core._lookup.cljs$core$IFn$_invoke$arity$3(this__5126__auto____$1,k__5127__auto__,null);
});

schema.core.RequiredKey.prototype.cljs$core$ILookup$_lookup$arity$3 = (function (this__5128__auto__,k27295,else__5129__auto__){
var self__ = this;
var this__5128__auto____$1 = this;
var G__27297 = (((k27295 instanceof cljs.core.Keyword))?k27295.fqn:null);
switch (G__27297) {
case "k":
return self__.k;

break;
default:
return cljs.core.get.cljs$core$IFn$_invoke$arity$3(self__.__extmap,k27295,else__5129__auto__);

}
});

schema.core.RequiredKey.prototype.cljs$core$IPrintWithWriter$_pr_writer$arity$3 = (function (this__5140__auto__,writer__5141__auto__,opts__5142__auto__){
var self__ = this;
var this__5140__auto____$1 = this;
var pr_pair__5143__auto__ = ((function (this__5140__auto____$1){
return (function (keyval__5144__auto__){
return cljs.core.pr_sequential_writer(writer__5141__auto__,cljs.core.pr_writer,""," ","",opts__5142__auto__,keyval__5144__auto__);
});})(this__5140__auto____$1))
;
return cljs.core.pr_sequential_writer(writer__5141__auto__,pr_pair__5143__auto__,"#schema.core.RequiredKey{",", ","}",opts__5142__auto__,cljs.core.concat.cljs$core$IFn$_invoke$arity$2(new cljs.core.PersistentVector(null, 1, 5, cljs.core.PersistentVector.EMPTY_NODE, [(new cljs.core.PersistentVector(null,2,(5),cljs.core.PersistentVector.EMPTY_NODE,[cljs.core.cst$kw$k,self__.k],null))], null),self__.__extmap));
});

schema.core.RequiredKey.prototype.cljs$core$IIterable$ = true;

schema.core.RequiredKey.prototype.cljs$core$IIterable$_iterator$arity$1 = (function (G__27294){
var self__ = this;
var G__27294__$1 = this;
return (new cljs.core.RecordIter((0),G__27294__$1,1,new cljs.core.PersistentVector(null, 1, 5, cljs.core.PersistentVector.EMPTY_NODE, [cljs.core.cst$kw$k], null),cljs.core._iterator(self__.__extmap)));
});

schema.core.RequiredKey.prototype.cljs$core$IMeta$_meta$arity$1 = (function (this__5124__auto__){
var self__ = this;
var this__5124__auto____$1 = this;
return self__.__meta;
});

schema.core.RequiredKey.prototype.cljs$core$ICloneable$_clone$arity$1 = (function (this__5120__auto__){
var self__ = this;
var this__5120__auto____$1 = this;
return (new schema.core.RequiredKey(self__.k,self__.__meta,self__.__extmap,self__.__hash));
});

schema.core.RequiredKey.prototype.cljs$core$ICounted$_count$arity$1 = (function (this__5130__auto__){
var self__ = this;
var this__5130__auto____$1 = this;
return (1 + cljs.core.count(self__.__extmap));
});

schema.core.RequiredKey.prototype.cljs$core$IHash$_hash$arity$1 = (function (this__5121__auto__){
var self__ = this;
var this__5121__auto____$1 = this;
var h__4947__auto__ = self__.__hash;
if(!((h__4947__auto__ == null))){
return h__4947__auto__;
} else {
var h__4947__auto____$1 = cljs.core.hash_imap(this__5121__auto____$1);
self__.__hash = h__4947__auto____$1;

return h__4947__auto____$1;
}
});

schema.core.RequiredKey.prototype.cljs$core$IEquiv$_equiv$arity$2 = (function (this__5122__auto__,other__5123__auto__){
var self__ = this;
var this__5122__auto____$1 = this;
if(cljs.core.truth_((function (){var and__4500__auto__ = other__5123__auto__;
if(cljs.core.truth_(and__4500__auto__)){
var and__4500__auto____$1 = (this__5122__auto____$1.constructor === other__5123__auto__.constructor);
if(and__4500__auto____$1){
return cljs.core.equiv_map(this__5122__auto____$1,other__5123__auto__);
} else {
return and__4500__auto____$1;
}
} else {
return and__4500__auto__;
}
})())){
return true;
} else {
return false;
}
});

schema.core.RequiredKey.prototype.cljs$core$IMap$_dissoc$arity$2 = (function (this__5135__auto__,k__5136__auto__){
var self__ = this;
var this__5135__auto____$1 = this;
if(cljs.core.contains_QMARK_(new cljs.core.PersistentHashSet(null, new cljs.core.PersistentArrayMap(null, 1, [cljs.core.cst$kw$k,null], null), null),k__5136__auto__)){
return cljs.core.dissoc.cljs$core$IFn$_invoke$arity$2(cljs.core.with_meta(cljs.core.into.cljs$core$IFn$_invoke$arity$2(cljs.core.PersistentArrayMap.EMPTY,this__5135__auto____$1),self__.__meta),k__5136__auto__);
} else {
return (new schema.core.RequiredKey(self__.k,self__.__meta,cljs.core.not_empty(cljs.core.dissoc.cljs$core$IFn$_invoke$arity$2(self__.__extmap,k__5136__auto__)),null));
}
});

schema.core.RequiredKey.prototype.cljs$core$IAssociative$_assoc$arity$3 = (function (this__5133__auto__,k__5134__auto__,G__27294){
var self__ = this;
var this__5133__auto____$1 = this;
var pred__27298 = cljs.core.keyword_identical_QMARK_;
var expr__27299 = k__5134__auto__;
if(cljs.core.truth_((pred__27298.cljs$core$IFn$_invoke$arity$2 ? pred__27298.cljs$core$IFn$_invoke$arity$2(cljs.core.cst$kw$k,expr__27299) : pred__27298.call(null,cljs.core.cst$kw$k,expr__27299)))){
return (new schema.core.RequiredKey(G__27294,self__.__meta,self__.__extmap,null));
} else {
return (new schema.core.RequiredKey(self__.k,self__.__meta,cljs.core.assoc.cljs$core$IFn$_invoke$arity$3(self__.__extmap,k__5134__auto__,G__27294),null));
}
});

schema.core.RequiredKey.prototype.cljs$core$ISeqable$_seq$arity$1 = (function (this__5138__auto__){
var self__ = this;
var this__5138__auto____$1 = this;
return cljs.core.seq(cljs.core.concat.cljs$core$IFn$_invoke$arity$2(new cljs.core.PersistentVector(null, 1, 5, cljs.core.PersistentVector.EMPTY_NODE, [(new cljs.core.PersistentVector(null,2,(5),cljs.core.PersistentVector.EMPTY_NODE,[cljs.core.cst$kw$k,self__.k],null))], null),self__.__extmap));
});

schema.core.RequiredKey.prototype.cljs$core$IWithMeta$_with_meta$arity$2 = (function (this__5125__auto__,G__27294){
var self__ = this;
var this__5125__auto____$1 = this;
return (new schema.core.RequiredKey(self__.k,G__27294,self__.__extmap,self__.__hash));
});

schema.core.RequiredKey.prototype.cljs$core$ICollection$_conj$arity$2 = (function (this__5131__auto__,entry__5132__auto__){
var self__ = this;
var this__5131__auto____$1 = this;
if(cljs.core.vector_QMARK_(entry__5132__auto__)){
return cljs.core._assoc(this__5131__auto____$1,cljs.core._nth.cljs$core$IFn$_invoke$arity$2(entry__5132__auto__,(0)),cljs.core._nth.cljs$core$IFn$_invoke$arity$2(entry__5132__auto__,(1)));
} else {
return cljs.core.reduce.cljs$core$IFn$_invoke$arity$3(cljs.core._conj,this__5131__auto____$1,entry__5132__auto__);
}
});

schema.core.RequiredKey.getBasis = (function (){
return new cljs.core.PersistentVector(null, 1, 5, cljs.core.PersistentVector.EMPTY_NODE, [cljs.core.cst$sym$k], null);
});

schema.core.RequiredKey.cljs$lang$type = true;

schema.core.RequiredKey.cljs$lang$ctorPrSeq = (function (this__5160__auto__){
return cljs.core._conj(cljs.core.List.EMPTY,"schema.core/RequiredKey");
});

schema.core.RequiredKey.cljs$lang$ctorPrWriter = (function (this__5160__auto__,writer__5161__auto__){
return cljs.core._write(writer__5161__auto__,"schema.core/RequiredKey");
});

schema.core.__GT_RequiredKey = (function schema$core$__GT_RequiredKey(k){
return (new schema.core.RequiredKey(k,null,null,null));
});

schema.core.map__GT_RequiredKey = (function schema$core$map__GT_RequiredKey(G__27296){
return (new schema.core.RequiredKey(cljs.core.cst$kw$k.cljs$core$IFn$_invoke$arity$1(G__27296),null,cljs.core.dissoc.cljs$core$IFn$_invoke$arity$2(G__27296,cljs.core.cst$kw$k),null));
});

/**
 * A required key in a map
 */
schema.core.required_key = (function schema$core$required_key(k){
if((k instanceof cljs.core.Keyword)){
return k;
} else {
return (new schema.core.RequiredKey(k,null,null,null));
}
});
schema.core.required_key_QMARK_ = (function schema$core$required_key_QMARK_(ks){
return ((ks instanceof cljs.core.Keyword)) || ((ks instanceof schema.core.RequiredKey));
});

/**
* @constructor
 * @implements {cljs.core.IRecord}
 * @implements {cljs.core.IEquiv}
 * @implements {cljs.core.IHash}
 * @implements {cljs.core.ICollection}
 * @implements {cljs.core.ICounted}
 * @implements {cljs.core.ISeqable}
 * @implements {cljs.core.IMeta}
 * @implements {cljs.core.ICloneable}
 * @implements {cljs.core.IPrintWithWriter}
 * @implements {cljs.core.IIterable}
 * @implements {cljs.core.IWithMeta}
 * @implements {cljs.core.IAssociative}
 * @implements {cljs.core.IMap}
 * @implements {cljs.core.ILookup}
*/
schema.core.OptionalKey = (function (k,__meta,__extmap,__hash){
this.k = k;
this.__meta = __meta;
this.__extmap = __extmap;
this.__hash = __hash;
this.cljs$lang$protocol_mask$partition0$ = 2229667594;
this.cljs$lang$protocol_mask$partition1$ = 8192;
})
schema.core.OptionalKey.prototype.cljs$core$ILookup$_lookup$arity$2 = (function (this__5126__auto__,k__5127__auto__){
var self__ = this;
var this__5126__auto____$1 = this;
return cljs.core._lookup.cljs$core$IFn$_invoke$arity$3(this__5126__auto____$1,k__5127__auto__,null);
});

schema.core.OptionalKey.prototype.cljs$core$ILookup$_lookup$arity$3 = (function (this__5128__auto__,k27303,else__5129__auto__){
var self__ = this;
var this__5128__auto____$1 = this;
var G__27305 = (((k27303 instanceof cljs.core.Keyword))?k27303.fqn:null);
switch (G__27305) {
case "k":
return self__.k;

break;
default:
return cljs.core.get.cljs$core$IFn$_invoke$arity$3(self__.__extmap,k27303,else__5129__auto__);

}
});

schema.core.OptionalKey.prototype.cljs$core$IPrintWithWriter$_pr_writer$arity$3 = (function (this__5140__auto__,writer__5141__auto__,opts__5142__auto__){
var self__ = this;
var this__5140__auto____$1 = this;
var pr_pair__5143__auto__ = ((function (this__5140__auto____$1){
return (function (keyval__5144__auto__){
return cljs.core.pr_sequential_writer(writer__5141__auto__,cljs.core.pr_writer,""," ","",opts__5142__auto__,keyval__5144__auto__);
});})(this__5140__auto____$1))
;
return cljs.core.pr_sequential_writer(writer__5141__auto__,pr_pair__5143__auto__,"#schema.core.OptionalKey{",", ","}",opts__5142__auto__,cljs.core.concat.cljs$core$IFn$_invoke$arity$2(new cljs.core.PersistentVector(null, 1, 5, cljs.core.PersistentVector.EMPTY_NODE, [(new cljs.core.PersistentVector(null,2,(5),cljs.core.PersistentVector.EMPTY_NODE,[cljs.core.cst$kw$k,self__.k],null))], null),self__.__extmap));
});

schema.core.OptionalKey.prototype.cljs$core$IIterable$ = true;

schema.core.OptionalKey.prototype.cljs$core$IIterable$_iterator$arity$1 = (function (G__27302){
var self__ = this;
var G__27302__$1 = this;
return (new cljs.core.RecordIter((0),G__27302__$1,1,new cljs.core.PersistentVector(null, 1, 5, cljs.core.PersistentVector.EMPTY_NODE, [cljs.core.cst$kw$k], null),cljs.core._iterator(self__.__extmap)));
});

schema.core.OptionalKey.prototype.cljs$core$IMeta$_meta$arity$1 = (function (this__5124__auto__){
var self__ = this;
var this__5124__auto____$1 = this;
return self__.__meta;
});

schema.core.OptionalKey.prototype.cljs$core$ICloneable$_clone$arity$1 = (function (this__5120__auto__){
var self__ = this;
var this__5120__auto____$1 = this;
return (new schema.core.OptionalKey(self__.k,self__.__meta,self__.__extmap,self__.__hash));
});

schema.core.OptionalKey.prototype.cljs$core$ICounted$_count$arity$1 = (function (this__5130__auto__){
var self__ = this;
var this__5130__auto____$1 = this;
return (1 + cljs.core.count(self__.__extmap));
});

schema.core.OptionalKey.prototype.cljs$core$IHash$_hash$arity$1 = (function (this__5121__auto__){
var self__ = this;
var this__5121__auto____$1 = this;
var h__4947__auto__ = self__.__hash;
if(!((h__4947__auto__ == null))){
return h__4947__auto__;
} else {
var h__4947__auto____$1 = cljs.core.hash_imap(this__5121__auto____$1);
self__.__hash = h__4947__auto____$1;

return h__4947__auto____$1;
}
});

schema.core.OptionalKey.prototype.cljs$core$IEquiv$_equiv$arity$2 = (function (this__5122__auto__,other__5123__auto__){
var self__ = this;
var this__5122__auto____$1 = this;
if(cljs.core.truth_((function (){var and__4500__auto__ = other__5123__auto__;
if(cljs.core.truth_(and__4500__auto__)){
var and__4500__auto____$1 = (this__5122__auto____$1.constructor === other__5123__auto__.constructor);
if(and__4500__auto____$1){
return cljs.core.equiv_map(this__5122__auto____$1,other__5123__auto__);
} else {
return and__4500__auto____$1;
}
} else {
return and__4500__auto__;
}
})())){
return true;
} else {
return false;
}
});

schema.core.OptionalKey.prototype.cljs$core$IMap$_dissoc$arity$2 = (function (this__5135__auto__,k__5136__auto__){
var self__ = this;
var this__5135__auto____$1 = this;
if(cljs.core.contains_QMARK_(new cljs.core.PersistentHashSet(null, new cljs.core.PersistentArrayMap(null, 1, [cljs.core.cst$kw$k,null], null), null),k__5136__auto__)){
return cljs.core.dissoc.cljs$core$IFn$_invoke$arity$2(cljs.core.with_meta(cljs.core.into.cljs$core$IFn$_invoke$arity$2(cljs.core.PersistentArrayMap.EMPTY,this__5135__auto____$1),self__.__meta),k__5136__auto__);
} else {
return (new schema.core.OptionalKey(self__.k,self__.__meta,cljs.core.not_empty(cljs.core.dissoc.cljs$core$IFn$_invoke$arity$2(self__.__extmap,k__5136__auto__)),null));
}
});

schema.core.OptionalKey.prototype.cljs$core$IAssociative$_assoc$arity$3 = (function (this__5133__auto__,k__5134__auto__,G__27302){
var self__ = this;
var this__5133__auto____$1 = this;
var pred__27306 = cljs.core.keyword_identical_QMARK_;
var expr__27307 = k__5134__auto__;
if(cljs.core.truth_((pred__27306.cljs$core$IFn$_invoke$arity$2 ? pred__27306.cljs$core$IFn$_invoke$arity$2(cljs.core.cst$kw$k,expr__27307) : pred__27306.call(null,cljs.core.cst$kw$k,expr__27307)))){
return (new schema.core.OptionalKey(G__27302,self__.__meta,self__.__extmap,null));
} else {
return (new schema.core.OptionalKey(self__.k,self__.__meta,cljs.core.assoc.cljs$core$IFn$_invoke$arity$3(self__.__extmap,k__5134__auto__,G__27302),null));
}
});

schema.core.OptionalKey.prototype.cljs$core$ISeqable$_seq$arity$1 = (function (this__5138__auto__){
var self__ = this;
var this__5138__auto____$1 = this;
return cljs.core.seq(cljs.core.concat.cljs$core$IFn$_invoke$arity$2(new cljs.core.PersistentVector(null, 1, 5, cljs.core.PersistentVector.EMPTY_NODE, [(new cljs.core.PersistentVector(null,2,(5),cljs.core.PersistentVector.EMPTY_NODE,[cljs.core.cst$kw$k,self__.k],null))], null),self__.__extmap));
});

schema.core.OptionalKey.prototype.cljs$core$IWithMeta$_with_meta$arity$2 = (function (this__5125__auto__,G__27302){
var self__ = this;
var this__5125__auto____$1 = this;
return (new schema.core.OptionalKey(self__.k,G__27302,self__.__extmap,self__.__hash));
});

schema.core.OptionalKey.prototype.cljs$core$ICollection$_conj$arity$2 = (function (this__5131__auto__,entry__5132__auto__){
var self__ = this;
var this__5131__auto____$1 = this;
if(cljs.core.vector_QMARK_(entry__5132__auto__)){
return cljs.core._assoc(this__5131__auto____$1,cljs.core._nth.cljs$core$IFn$_invoke$arity$2(entry__5132__auto__,(0)),cljs.core._nth.cljs$core$IFn$_invoke$arity$2(entry__5132__auto__,(1)));
} else {
return cljs.core.reduce.cljs$core$IFn$_invoke$arity$3(cljs.core._conj,this__5131__auto____$1,entry__5132__auto__);
}
});

schema.core.OptionalKey.getBasis = (function (){
return new cljs.core.PersistentVector(null, 1, 5, cljs.core.PersistentVector.EMPTY_NODE, [cljs.core.cst$sym$k], null);
});

schema.core.OptionalKey.cljs$lang$type = true;

schema.core.OptionalKey.cljs$lang$ctorPrSeq = (function (this__5160__auto__){
return cljs.core._conj(cljs.core.List.EMPTY,"schema.core/OptionalKey");
});

schema.core.OptionalKey.cljs$lang$ctorPrWriter = (function (this__5160__auto__,writer__5161__auto__){
return cljs.core._write(writer__5161__auto__,"schema.core/OptionalKey");
});

schema.core.__GT_OptionalKey = (function schema$core$__GT_OptionalKey(k){
return (new schema.core.OptionalKey(k,null,null,null));
});

schema.core.map__GT_OptionalKey = (function schema$core$map__GT_OptionalKey(G__27304){
return (new schema.core.OptionalKey(cljs.core.cst$kw$k.cljs$core$IFn$_invoke$arity$1(G__27304),null,cljs.core.dissoc.cljs$core$IFn$_invoke$arity$2(G__27304,cljs.core.cst$kw$k),null));
});

/**
 * An optional key in a map
 */
schema.core.optional_key = (function schema$core$optional_key(k){
return (new schema.core.OptionalKey(k,null,null,null));
});
schema.core.optional_key_QMARK_ = (function schema$core$optional_key_QMARK_(ks){
return (ks instanceof schema.core.OptionalKey);
});
schema.core.explicit_schema_key = (function schema$core$explicit_schema_key(ks){
if((ks instanceof cljs.core.Keyword)){
return ks;
} else {
if((ks instanceof schema.core.RequiredKey)){
return ks.k;
} else {
if(cljs.core.truth_(schema.core.optional_key_QMARK_(ks))){
return ks.k;
} else {
throw (new Error(schema.utils.format_STAR_.cljs$core$IFn$_invoke$arity$variadic("Bad explicit key: %s",cljs.core.array_seq([ks], 0))));

}
}
}
});
schema.core.specific_key_QMARK_ = (function schema$core$specific_key_QMARK_(ks){
var or__4512__auto__ = schema.core.required_key_QMARK_(ks);
if(cljs.core.truth_(or__4512__auto__)){
return or__4512__auto__;
} else {
return schema.core.optional_key_QMARK_(ks);
}
});
schema.core.explain_kspec = (function schema$core$explain_kspec(kspec){
if(cljs.core.truth_(schema.core.specific_key_QMARK_(kspec))){
if((kspec instanceof cljs.core.Keyword)){
return kspec;
} else {
return cljs.core._conj(cljs.core._conj(cljs.core.List.EMPTY,schema.core.explicit_schema_key(kspec)),(cljs.core.truth_(schema.core.required_key_QMARK_(kspec))?cljs.core.cst$sym$required_DASH_key:(cljs.core.truth_(schema.core.optional_key_QMARK_(kspec))?cljs.core.cst$sym$optional_DASH_key:null)));
}
} else {
return schema.core.explain(kspec);
}
});

/**
* @constructor
 * @implements {cljs.core.IRecord}
 * @implements {cljs.core.IEquiv}
 * @implements {cljs.core.IHash}
 * @implements {cljs.core.ICollection}
 * @implements {schema.core.Schema}
 * @implements {cljs.core.ICounted}
 * @implements {cljs.core.ISeqable}
 * @implements {cljs.core.IMeta}
 * @implements {cljs.core.ICloneable}
 * @implements {cljs.core.IPrintWithWriter}
 * @implements {cljs.core.IIterable}
 * @implements {cljs.core.IWithMeta}
 * @implements {cljs.core.IAssociative}
 * @implements {cljs.core.IMap}
 * @implements {cljs.core.ILookup}
*/
schema.core.MapEntry = (function (kspec,val_schema,__meta,__extmap,__hash){
this.kspec = kspec;
this.val_schema = val_schema;
this.__meta = __meta;
this.__extmap = __extmap;
this.__hash = __hash;
this.cljs$lang$protocol_mask$partition0$ = 2229667594;
this.cljs$lang$protocol_mask$partition1$ = 8192;
})
schema.core.MapEntry.prototype.cljs$core$ILookup$_lookup$arity$2 = (function (this__5126__auto__,k__5127__auto__){
var self__ = this;
var this__5126__auto____$1 = this;
return cljs.core._lookup.cljs$core$IFn$_invoke$arity$3(this__5126__auto____$1,k__5127__auto__,null);
});

schema.core.MapEntry.prototype.cljs$core$ILookup$_lookup$arity$3 = (function (this__5128__auto__,k27311,else__5129__auto__){
var self__ = this;
var this__5128__auto____$1 = this;
var G__27313 = (((k27311 instanceof cljs.core.Keyword))?k27311.fqn:null);
switch (G__27313) {
case "kspec":
return self__.kspec;

break;
case "val-schema":
return self__.val_schema;

break;
default:
return cljs.core.get.cljs$core$IFn$_invoke$arity$3(self__.__extmap,k27311,else__5129__auto__);

}
});

schema.core.MapEntry.prototype.cljs$core$IPrintWithWriter$_pr_writer$arity$3 = (function (this__5140__auto__,writer__5141__auto__,opts__5142__auto__){
var self__ = this;
var this__5140__auto____$1 = this;
var pr_pair__5143__auto__ = ((function (this__5140__auto____$1){
return (function (keyval__5144__auto__){
return cljs.core.pr_sequential_writer(writer__5141__auto__,cljs.core.pr_writer,""," ","",opts__5142__auto__,keyval__5144__auto__);
});})(this__5140__auto____$1))
;
return cljs.core.pr_sequential_writer(writer__5141__auto__,pr_pair__5143__auto__,"#schema.core.MapEntry{",", ","}",opts__5142__auto__,cljs.core.concat.cljs$core$IFn$_invoke$arity$2(new cljs.core.PersistentVector(null, 2, 5, cljs.core.PersistentVector.EMPTY_NODE, [(new cljs.core.PersistentVector(null,2,(5),cljs.core.PersistentVector.EMPTY_NODE,[cljs.core.cst$kw$kspec,self__.kspec],null)),(new cljs.core.PersistentVector(null,2,(5),cljs.core.PersistentVector.EMPTY_NODE,[cljs.core.cst$kw$val_DASH_schema,self__.val_schema],null))], null),self__.__extmap));
});

schema.core.MapEntry.prototype.cljs$core$IIterable$ = true;

schema.core.MapEntry.prototype.cljs$core$IIterable$_iterator$arity$1 = (function (G__27310){
var self__ = this;
var G__27310__$1 = this;
return (new cljs.core.RecordIter((0),G__27310__$1,2,new cljs.core.PersistentVector(null, 2, 5, cljs.core.PersistentVector.EMPTY_NODE, [cljs.core.cst$kw$kspec,cljs.core.cst$kw$val_DASH_schema], null),cljs.core._iterator(self__.__extmap)));
});

schema.core.MapEntry.prototype.cljs$core$IMeta$_meta$arity$1 = (function (this__5124__auto__){
var self__ = this;
var this__5124__auto____$1 = this;
return self__.__meta;
});

schema.core.MapEntry.prototype.cljs$core$ICloneable$_clone$arity$1 = (function (this__5120__auto__){
var self__ = this;
var this__5120__auto____$1 = this;
return (new schema.core.MapEntry(self__.kspec,self__.val_schema,self__.__meta,self__.__extmap,self__.__hash));
});

schema.core.MapEntry.prototype.cljs$core$ICounted$_count$arity$1 = (function (this__5130__auto__){
var self__ = this;
var this__5130__auto____$1 = this;
return (2 + cljs.core.count(self__.__extmap));
});

schema.core.MapEntry.prototype.cljs$core$IHash$_hash$arity$1 = (function (this__5121__auto__){
var self__ = this;
var this__5121__auto____$1 = this;
var h__4947__auto__ = self__.__hash;
if(!((h__4947__auto__ == null))){
return h__4947__auto__;
} else {
var h__4947__auto____$1 = cljs.core.hash_imap(this__5121__auto____$1);
self__.__hash = h__4947__auto____$1;

return h__4947__auto____$1;
}
});

schema.core.MapEntry.prototype.cljs$core$IEquiv$_equiv$arity$2 = (function (this__5122__auto__,other__5123__auto__){
var self__ = this;
var this__5122__auto____$1 = this;
if(cljs.core.truth_((function (){var and__4500__auto__ = other__5123__auto__;
if(cljs.core.truth_(and__4500__auto__)){
var and__4500__auto____$1 = (this__5122__auto____$1.constructor === other__5123__auto__.constructor);
if(and__4500__auto____$1){
return cljs.core.equiv_map(this__5122__auto____$1,other__5123__auto__);
} else {
return and__4500__auto____$1;
}
} else {
return and__4500__auto__;
}
})())){
return true;
} else {
return false;
}
});

schema.core.MapEntry.prototype.cljs$core$IMap$_dissoc$arity$2 = (function (this__5135__auto__,k__5136__auto__){
var self__ = this;
var this__5135__auto____$1 = this;
if(cljs.core.contains_QMARK_(new cljs.core.PersistentHashSet(null, new cljs.core.PersistentArrayMap(null, 2, [cljs.core.cst$kw$kspec,null,cljs.core.cst$kw$val_DASH_schema,null], null), null),k__5136__auto__)){
return cljs.core.dissoc.cljs$core$IFn$_invoke$arity$2(cljs.core.with_meta(cljs.core.into.cljs$core$IFn$_invoke$arity$2(cljs.core.PersistentArrayMap.EMPTY,this__5135__auto____$1),self__.__meta),k__5136__auto__);
} else {
return (new schema.core.MapEntry(self__.kspec,self__.val_schema,self__.__meta,cljs.core.not_empty(cljs.core.dissoc.cljs$core$IFn$_invoke$arity$2(self__.__extmap,k__5136__auto__)),null));
}
});

schema.core.MapEntry.prototype.cljs$core$IAssociative$_assoc$arity$3 = (function (this__5133__auto__,k__5134__auto__,G__27310){
var self__ = this;
var this__5133__auto____$1 = this;
var pred__27314 = cljs.core.keyword_identical_QMARK_;
var expr__27315 = k__5134__auto__;
if(cljs.core.truth_((pred__27314.cljs$core$IFn$_invoke$arity$2 ? pred__27314.cljs$core$IFn$_invoke$arity$2(cljs.core.cst$kw$kspec,expr__27315) : pred__27314.call(null,cljs.core.cst$kw$kspec,expr__27315)))){
return (new schema.core.MapEntry(G__27310,self__.val_schema,self__.__meta,self__.__extmap,null));
} else {
if(cljs.core.truth_((pred__27314.cljs$core$IFn$_invoke$arity$2 ? pred__27314.cljs$core$IFn$_invoke$arity$2(cljs.core.cst$kw$val_DASH_schema,expr__27315) : pred__27314.call(null,cljs.core.cst$kw$val_DASH_schema,expr__27315)))){
return (new schema.core.MapEntry(self__.kspec,G__27310,self__.__meta,self__.__extmap,null));
} else {
return (new schema.core.MapEntry(self__.kspec,self__.val_schema,self__.__meta,cljs.core.assoc.cljs$core$IFn$_invoke$arity$3(self__.__extmap,k__5134__auto__,G__27310),null));
}
}
});

schema.core.MapEntry.prototype.cljs$core$ISeqable$_seq$arity$1 = (function (this__5138__auto__){
var self__ = this;
var this__5138__auto____$1 = this;
return cljs.core.seq(cljs.core.concat.cljs$core$IFn$_invoke$arity$2(new cljs.core.PersistentVector(null, 2, 5, cljs.core.PersistentVector.EMPTY_NODE, [(new cljs.core.PersistentVector(null,2,(5),cljs.core.PersistentVector.EMPTY_NODE,[cljs.core.cst$kw$kspec,self__.kspec],null)),(new cljs.core.PersistentVector(null,2,(5),cljs.core.PersistentVector.EMPTY_NODE,[cljs.core.cst$kw$val_DASH_schema,self__.val_schema],null))], null),self__.__extmap));
});

schema.core.MapEntry.prototype.cljs$core$IWithMeta$_with_meta$arity$2 = (function (this__5125__auto__,G__27310){
var self__ = this;
var this__5125__auto____$1 = this;
return (new schema.core.MapEntry(self__.kspec,self__.val_schema,G__27310,self__.__extmap,self__.__hash));
});

schema.core.MapEntry.prototype.cljs$core$ICollection$_conj$arity$2 = (function (this__5131__auto__,entry__5132__auto__){
var self__ = this;
var this__5131__auto____$1 = this;
if(cljs.core.vector_QMARK_(entry__5132__auto__)){
return cljs.core._assoc(this__5131__auto____$1,cljs.core._nth.cljs$core$IFn$_invoke$arity$2(entry__5132__auto__,(0)),cljs.core._nth.cljs$core$IFn$_invoke$arity$2(entry__5132__auto__,(1)));
} else {
return cljs.core.reduce.cljs$core$IFn$_invoke$arity$3(cljs.core._conj,this__5131__auto____$1,entry__5132__auto__);
}
});

schema.core.MapEntry.prototype.schema$core$Schema$ = true;

schema.core.MapEntry.prototype.schema$core$Schema$walker$arity$1 = (function (this$){
var self__ = this;
var this$__$1 = this;
var val_walker = (schema.core.subschema_walker.cljs$core$IFn$_invoke$arity$1 ? schema.core.subschema_walker.cljs$core$IFn$_invoke$arity$1(self__.val_schema) : schema.core.subschema_walker.call(null,self__.val_schema));
if(cljs.core.truth_(schema.core.specific_key_QMARK_(self__.kspec))){
var optional_QMARK_ = schema.core.optional_key_QMARK_(self__.kspec);
var k = schema.core.explicit_schema_key(self__.kspec);
return ((function (optional_QMARK_,k,val_walker,this$__$1){
return (function (x){
if((schema.core._PLUS_missing_PLUS_ === x)){
if(cljs.core.truth_(optional_QMARK_)){
return null;
} else {
return schema.utils.error(new cljs.core.PersistentVector(null, 2, 5, cljs.core.PersistentVector.EMPTY_NODE, [k,cljs.core.cst$sym$missing_DASH_required_DASH_key], null));
}
} else {
if(!(cljs.core._EQ_.cljs$core$IFn$_invoke$arity$2((2),cljs.core.count(x)))){
return schema.utils.error(schema.utils.__GT_ValidationError(this$__$1,x,(new cljs.core.Delay(((function (optional_QMARK_,k,val_walker,this$__$1){
return (function (){
return cljs.core._conj(cljs.core._conj(cljs.core._conj(cljs.core.List.EMPTY,cljs.core._conj(cljs.core._conj(cljs.core.List.EMPTY,schema.utils.value_name(x)),cljs.core.cst$sym$count)),(2)),cljs.core._EQ_);
});})(optional_QMARK_,k,val_walker,this$__$1))
,null)),null));
} else {
var vec__27317 = x;
var xk = cljs.core.nth.cljs$core$IFn$_invoke$arity$3(vec__27317,(0),null);
var xv = cljs.core.nth.cljs$core$IFn$_invoke$arity$3(vec__27317,(1),null);
if(cljs.core._EQ_.cljs$core$IFn$_invoke$arity$2(xk,k)){
} else {
throw (new Error([cljs.core.str("Assert failed: "),cljs.core.str(cljs.core.pr_str.cljs$core$IFn$_invoke$arity$variadic(cljs.core.array_seq([cljs.core.list(cljs.core.cst$sym$_EQ_,cljs.core.cst$sym$xk,cljs.core.cst$sym$k)], 0)))].join('')));
}

var vres = (val_walker.cljs$core$IFn$_invoke$arity$1 ? val_walker.cljs$core$IFn$_invoke$arity$1(xv) : val_walker.call(null,xv));
var temp__4423__auto__ = schema.utils.error_val(vres);
if(cljs.core.truth_(temp__4423__auto__)){
var ve = temp__4423__auto__;
return schema.utils.error(new cljs.core.PersistentVector(null, 2, 5, cljs.core.PersistentVector.EMPTY_NODE, [xk,ve], null));
} else {
return new cljs.core.PersistentVector(null, 2, 5, cljs.core.PersistentVector.EMPTY_NODE, [xk,vres], null);
}

}
}
});
;})(optional_QMARK_,k,val_walker,this$__$1))
} else {
var key_walker = (schema.core.subschema_walker.cljs$core$IFn$_invoke$arity$1 ? schema.core.subschema_walker.cljs$core$IFn$_invoke$arity$1(self__.kspec) : schema.core.subschema_walker.call(null,self__.kspec));
return ((function (key_walker,val_walker,this$__$1){
return (function (x){
if(!(cljs.core._EQ_.cljs$core$IFn$_invoke$arity$2((2),cljs.core.count(x)))){
return schema.utils.error(schema.utils.__GT_ValidationError(this$__$1,x,(new cljs.core.Delay(((function (key_walker,val_walker,this$__$1){
return (function (){
return cljs.core._conj(cljs.core._conj(cljs.core._conj(cljs.core.List.EMPTY,cljs.core._conj(cljs.core._conj(cljs.core.List.EMPTY,schema.utils.value_name(x)),cljs.core.cst$sym$count)),(2)),cljs.core._EQ_);
});})(key_walker,val_walker,this$__$1))
,null)),null));
} else {
var out_k = (function (){var G__27318 = cljs.core.key(x);
return (key_walker.cljs$core$IFn$_invoke$arity$1 ? key_walker.cljs$core$IFn$_invoke$arity$1(G__27318) : key_walker.call(null,G__27318));
})();
var out_ke = schema.utils.error_val(out_k);
var out_v = (function (){var G__27319 = cljs.core.val(x);
return (val_walker.cljs$core$IFn$_invoke$arity$1 ? val_walker.cljs$core$IFn$_invoke$arity$1(G__27319) : val_walker.call(null,G__27319));
})();
var out_ve = schema.utils.error_val(out_v);
if(cljs.core.truth_((function (){var or__4512__auto__ = out_ke;
if(cljs.core.truth_(or__4512__auto__)){
return or__4512__auto__;
} else {
return out_ve;
}
})())){
return schema.utils.error(new cljs.core.PersistentVector(null, 2, 5, cljs.core.PersistentVector.EMPTY_NODE, [(function (){var or__4512__auto__ = out_ke;
if(cljs.core.truth_(or__4512__auto__)){
return or__4512__auto__;
} else {
return cljs.core.key(x);
}
})(),(function (){var or__4512__auto__ = out_ve;
if(cljs.core.truth_(or__4512__auto__)){
return or__4512__auto__;
} else {
return cljs.core.cst$sym$invalid_DASH_key;
}
})()], null));
} else {
return new cljs.core.PersistentVector(null, 2, 5, cljs.core.PersistentVector.EMPTY_NODE, [out_k,out_v], null);
}
}
});
;})(key_walker,val_walker,this$__$1))
}
});

schema.core.MapEntry.prototype.schema$core$Schema$explain$arity$1 = (function (this$){
var self__ = this;
var this$__$1 = this;
return cljs.core._conj(cljs.core._conj(cljs.core._conj(cljs.core.List.EMPTY,schema.core.explain(self__.val_schema)),schema.core.explain_kspec(self__.kspec)),cljs.core.cst$sym$map_DASH_entry);
});

schema.core.MapEntry.getBasis = (function (){
return new cljs.core.PersistentVector(null, 2, 5, cljs.core.PersistentVector.EMPTY_NODE, [cljs.core.cst$sym$kspec,cljs.core.cst$sym$val_DASH_schema], null);
});

schema.core.MapEntry.cljs$lang$type = true;

schema.core.MapEntry.cljs$lang$ctorPrSeq = (function (this__5160__auto__){
return cljs.core._conj(cljs.core.List.EMPTY,"schema.core/MapEntry");
});

schema.core.MapEntry.cljs$lang$ctorPrWriter = (function (this__5160__auto__,writer__5161__auto__){
return cljs.core._write(writer__5161__auto__,"schema.core/MapEntry");
});

schema.core.__GT_MapEntry = (function schema$core$__GT_MapEntry(kspec,val_schema){
return (new schema.core.MapEntry(kspec,val_schema,null,null,null));
});

schema.core.map__GT_MapEntry = (function schema$core$map__GT_MapEntry(G__27312){
return (new schema.core.MapEntry(cljs.core.cst$kw$kspec.cljs$core$IFn$_invoke$arity$1(G__27312),cljs.core.cst$kw$val_DASH_schema.cljs$core$IFn$_invoke$arity$1(G__27312),null,cljs.core.dissoc.cljs$core$IFn$_invoke$arity$variadic(G__27312,cljs.core.cst$kw$kspec,cljs.core.array_seq([cljs.core.cst$kw$val_DASH_schema], 0)),null));
});

schema.core.map_entry = (function schema$core$map_entry(kspec,val_schema){
return (new schema.core.MapEntry(kspec,val_schema,null,null,null));
});
schema.core.find_extra_keys_schema = (function schema$core$find_extra_keys_schema(map_schema){
var key_schemata = cljs.core.remove.cljs$core$IFn$_invoke$arity$2(schema.core.specific_key_QMARK_,cljs.core.keys(map_schema));
if((cljs.core.count(key_schemata) < (2))){
} else {
throw (new Error(schema.utils.format_STAR_.cljs$core$IFn$_invoke$arity$variadic("More than one non-optional/required key schemata: %s",cljs.core.array_seq([cljs.core.vec(key_schemata)], 0))));
}

return cljs.core.first(key_schemata);
});
schema.core.preserve_map_type = (function schema$core$preserve_map_type(original,walker_result){
if(cljs.core.truth_((function (){var and__4500__auto__ = schema.utils.record_QMARK_(original);
if(cljs.core.truth_(and__4500__auto__)){
return cljs.core.not(schema.utils.error_QMARK_(walker_result));
} else {
return and__4500__auto__;
}
})())){
return cljs.core.into.cljs$core$IFn$_invoke$arity$2(original,walker_result);
} else {
return walker_result;
}
});
schema.core.map_walker = (function schema$core$map_walker(map_schema){
var extra_keys_schema = schema.core.find_extra_keys_schema(map_schema);
var extra_walker = (cljs.core.truth_(extra_keys_schema)?(schema.core.subschema_walker.cljs$core$IFn$_invoke$arity$1 ? schema.core.subschema_walker.cljs$core$IFn$_invoke$arity$1(cljs.core.apply.cljs$core$IFn$_invoke$arity$2(schema.core.map_entry,cljs.core.find(map_schema,extra_keys_schema))) : schema.core.subschema_walker.call(null,cljs.core.apply.cljs$core$IFn$_invoke$arity$2(schema.core.map_entry,cljs.core.find(map_schema,extra_keys_schema)))):null);
var explicit_schema = cljs.core.dissoc.cljs$core$IFn$_invoke$arity$2(map_schema,extra_keys_schema);
var explicit_walkers = cljs.core.into.cljs$core$IFn$_invoke$arity$2(cljs.core.PersistentArrayMap.EMPTY,(function (){var iter__5284__auto__ = ((function (extra_keys_schema,extra_walker,explicit_schema){
return (function schema$core$map_walker_$_iter__27345(s__27346){
return (new cljs.core.LazySeq(null,((function (extra_keys_schema,extra_walker,explicit_schema){
return (function (){
var s__27346__$1 = s__27346;
while(true){
var temp__4425__auto__ = cljs.core.seq(s__27346__$1);
if(temp__4425__auto__){
var s__27346__$2 = temp__4425__auto__;
if(cljs.core.chunked_seq_QMARK_(s__27346__$2)){
var c__5282__auto__ = cljs.core.chunk_first(s__27346__$2);
var size__5283__auto__ = cljs.core.count(c__5282__auto__);
var b__27348 = cljs.core.chunk_buffer(size__5283__auto__);
if((function (){var i__27347 = (0);
while(true){
if((i__27347 < size__5283__auto__)){
var vec__27353 = cljs.core._nth.cljs$core$IFn$_invoke$arity$2(c__5282__auto__,i__27347);
var k = cljs.core.nth.cljs$core$IFn$_invoke$arity$3(vec__27353,(0),null);
var v = cljs.core.nth.cljs$core$IFn$_invoke$arity$3(vec__27353,(1),null);
cljs.core.chunk_append(b__27348,new cljs.core.PersistentVector(null, 2, 5, cljs.core.PersistentVector.EMPTY_NODE, [schema.core.explicit_schema_key(k),(schema.core.subschema_walker.cljs$core$IFn$_invoke$arity$1 ? schema.core.subschema_walker.cljs$core$IFn$_invoke$arity$1(schema.core.map_entry(k,v)) : schema.core.subschema_walker.call(null,schema.core.map_entry(k,v)))], null));

var G__27368 = (i__27347 + (1));
i__27347 = G__27368;
continue;
} else {
return true;
}
break;
}
})()){
return cljs.core.chunk_cons(cljs.core.chunk(b__27348),schema$core$map_walker_$_iter__27345(cljs.core.chunk_rest(s__27346__$2)));
} else {
return cljs.core.chunk_cons(cljs.core.chunk(b__27348),null);
}
} else {
var vec__27354 = cljs.core.first(s__27346__$2);
var k = cljs.core.nth.cljs$core$IFn$_invoke$arity$3(vec__27354,(0),null);
var v = cljs.core.nth.cljs$core$IFn$_invoke$arity$3(vec__27354,(1),null);
return cljs.core.cons(new cljs.core.PersistentVector(null, 2, 5, cljs.core.PersistentVector.EMPTY_NODE, [schema.core.explicit_schema_key(k),(schema.core.subschema_walker.cljs$core$IFn$_invoke$arity$1 ? schema.core.subschema_walker.cljs$core$IFn$_invoke$arity$1(schema.core.map_entry(k,v)) : schema.core.subschema_walker.call(null,schema.core.map_entry(k,v)))], null),schema$core$map_walker_$_iter__27345(cljs.core.rest(s__27346__$2)));
}
} else {
return null;
}
break;
}
});})(extra_keys_schema,extra_walker,explicit_schema))
,null,null));
});})(extra_keys_schema,extra_walker,explicit_schema))
;
return iter__5284__auto__(explicit_schema);
})());
var err_conj = schema.utils.result_builder(cljs.core.constantly(cljs.core.PersistentArrayMap.EMPTY));
if(cljs.core._EQ_.cljs$core$IFn$_invoke$arity$2(cljs.core.count(explicit_schema),cljs.core.count(explicit_walkers))){
} else {
throw (new Error(schema.utils.format_STAR_.cljs$core$IFn$_invoke$arity$variadic("Schema has multiple variants of the same explicit key: %s",cljs.core.array_seq([cljs.core.mapv.cljs$core$IFn$_invoke$arity$2(schema.core.explain_kspec,cljs.core.apply.cljs$core$IFn$_invoke$arity$2(cljs.core.concat,cljs.core.filter.cljs$core$IFn$_invoke$arity$2(((function (extra_keys_schema,extra_walker,explicit_schema,explicit_walkers,err_conj){
return (function (p1__27321_SHARP_){
return (cljs.core.count(p1__27321_SHARP_) > (1));
});})(extra_keys_schema,extra_walker,explicit_schema,explicit_walkers,err_conj))
,cljs.core.vals(cljs.core.group_by(schema.core.explicit_schema_key,cljs.core.keys(explicit_schema))))))], 0))));
}

return ((function (extra_keys_schema,extra_walker,explicit_schema,explicit_walkers,err_conj){
return (function (x){
if(!(cljs.core.map_QMARK_(x))){
return schema.utils.error(schema.utils.__GT_ValidationError(map_schema,x,(new cljs.core.Delay(((function (extra_keys_schema,extra_walker,explicit_schema,explicit_walkers,err_conj){
return (function (){
return cljs.core._conj(cljs.core._conj(cljs.core.List.EMPTY,schema.utils.value_name(x)),cljs.core.cst$sym$map_QMARK_);
});})(extra_keys_schema,extra_walker,explicit_schema,explicit_walkers,err_conj))
,null)),null));
} else {
return schema.core.preserve_map_type(x,(function (){var ok_key = cljs.core.PersistentHashSet.EMPTY;
var explicit_walkers__$1 = cljs.core.seq(explicit_walkers);
var out = cljs.core.PersistentArrayMap.EMPTY;
while(true){
if(cljs.core.not(explicit_walkers__$1)){
return cljs.core.reduce.cljs$core$IFn$_invoke$arity$3((cljs.core.truth_(extra_walker)?((function (ok_key,explicit_walkers__$1,out,extra_keys_schema,extra_walker,explicit_schema,explicit_walkers,err_conj){
return (function (out__$1,e){
var G__27355 = out__$1;
var G__27356 = (extra_walker.cljs$core$IFn$_invoke$arity$1 ? extra_walker.cljs$core$IFn$_invoke$arity$1(e) : extra_walker.call(null,e));
return (err_conj.cljs$core$IFn$_invoke$arity$2 ? err_conj.cljs$core$IFn$_invoke$arity$2(G__27355,G__27356) : err_conj.call(null,G__27355,G__27356));
});})(ok_key,explicit_walkers__$1,out,extra_keys_schema,extra_walker,explicit_schema,explicit_walkers,err_conj))
:((function (ok_key,explicit_walkers__$1,out,extra_keys_schema,extra_walker,explicit_schema,explicit_walkers,err_conj){
return (function (out__$1,p__27357){
var vec__27358 = p__27357;
var k = cljs.core.nth.cljs$core$IFn$_invoke$arity$3(vec__27358,(0),null);
var _ = cljs.core.nth.cljs$core$IFn$_invoke$arity$3(vec__27358,(1),null);
var G__27359 = out__$1;
var G__27360 = schema.utils.error(new cljs.core.PersistentVector(null, 2, 5, cljs.core.PersistentVector.EMPTY_NODE, [k,cljs.core.cst$sym$disallowed_DASH_key], null));
return (err_conj.cljs$core$IFn$_invoke$arity$2 ? err_conj.cljs$core$IFn$_invoke$arity$2(G__27359,G__27360) : err_conj.call(null,G__27359,G__27360));
});})(ok_key,explicit_walkers__$1,out,extra_keys_schema,extra_walker,explicit_schema,explicit_walkers,err_conj))
),out,cljs.core.remove.cljs$core$IFn$_invoke$arity$2(((function (ok_key,explicit_walkers__$1,out,extra_keys_schema,extra_walker,explicit_schema,explicit_walkers,err_conj){
return (function (p__27361){
var vec__27362 = p__27361;
var k = cljs.core.nth.cljs$core$IFn$_invoke$arity$3(vec__27362,(0),null);
var v = cljs.core.nth.cljs$core$IFn$_invoke$arity$3(vec__27362,(1),null);
return (ok_key.cljs$core$IFn$_invoke$arity$1 ? ok_key.cljs$core$IFn$_invoke$arity$1(k) : ok_key.call(null,k));
});})(ok_key,explicit_walkers__$1,out,extra_keys_schema,extra_walker,explicit_schema,explicit_walkers,err_conj))
,x));
} else {
var vec__27363 = cljs.core.first(explicit_walkers__$1);
var wk = cljs.core.nth.cljs$core$IFn$_invoke$arity$3(vec__27363,(0),null);
var wv = cljs.core.nth.cljs$core$IFn$_invoke$arity$3(vec__27363,(1),null);
var G__27369 = cljs.core.conj.cljs$core$IFn$_invoke$arity$2(ok_key,wk);
var G__27370 = cljs.core.next(explicit_walkers__$1);
var G__27371 = (function (){var G__27365 = out;
var G__27366 = (function (){var G__27367 = (function (){var or__4512__auto__ = cljs.core.find(x,wk);
if(cljs.core.truth_(or__4512__auto__)){
return or__4512__auto__;
} else {
return schema.core._PLUS_missing_PLUS_;
}
})();
return (wv.cljs$core$IFn$_invoke$arity$1 ? wv.cljs$core$IFn$_invoke$arity$1(G__27367) : wv.call(null,G__27367));
})();
return (err_conj.cljs$core$IFn$_invoke$arity$2 ? err_conj.cljs$core$IFn$_invoke$arity$2(G__27365,G__27366) : err_conj.call(null,G__27365,G__27366));
})();
ok_key = G__27369;
explicit_walkers__$1 = G__27370;
out = G__27371;
continue;
}
break;
}
})());
}
});
;})(extra_keys_schema,extra_walker,explicit_schema,explicit_walkers,err_conj))
});
schema.core.map_explain = (function schema$core$map_explain(this$){
return cljs.core.into.cljs$core$IFn$_invoke$arity$2(cljs.core.PersistentArrayMap.EMPTY,(function (){var iter__5284__auto__ = (function schema$core$map_explain_$_iter__27382(s__27383){
return (new cljs.core.LazySeq(null,(function (){
var s__27383__$1 = s__27383;
while(true){
var temp__4425__auto__ = cljs.core.seq(s__27383__$1);
if(temp__4425__auto__){
var s__27383__$2 = temp__4425__auto__;
if(cljs.core.chunked_seq_QMARK_(s__27383__$2)){
var c__5282__auto__ = cljs.core.chunk_first(s__27383__$2);
var size__5283__auto__ = cljs.core.count(c__5282__auto__);
var b__27385 = cljs.core.chunk_buffer(size__5283__auto__);
if((function (){var i__27384 = (0);
while(true){
if((i__27384 < size__5283__auto__)){
var vec__27390 = cljs.core._nth.cljs$core$IFn$_invoke$arity$2(c__5282__auto__,i__27384);
var k = cljs.core.nth.cljs$core$IFn$_invoke$arity$3(vec__27390,(0),null);
var v = cljs.core.nth.cljs$core$IFn$_invoke$arity$3(vec__27390,(1),null);
cljs.core.chunk_append(b__27385,cljs.core.vec(cljs.core.next(schema.core.explain(schema.core.map_entry(k,v)))));

var G__27392 = (i__27384 + (1));
i__27384 = G__27392;
continue;
} else {
return true;
}
break;
}
})()){
return cljs.core.chunk_cons(cljs.core.chunk(b__27385),schema$core$map_explain_$_iter__27382(cljs.core.chunk_rest(s__27383__$2)));
} else {
return cljs.core.chunk_cons(cljs.core.chunk(b__27385),null);
}
} else {
var vec__27391 = cljs.core.first(s__27383__$2);
var k = cljs.core.nth.cljs$core$IFn$_invoke$arity$3(vec__27391,(0),null);
var v = cljs.core.nth.cljs$core$IFn$_invoke$arity$3(vec__27391,(1),null);
return cljs.core.cons(cljs.core.vec(cljs.core.next(schema.core.explain(schema.core.map_entry(k,v)))),schema$core$map_explain_$_iter__27382(cljs.core.rest(s__27383__$2)));
}
} else {
return null;
}
break;
}
}),null,null));
});
return iter__5284__auto__(this$);
})());
});
cljs.core.PersistentArrayMap.prototype.schema$core$Schema$ = true;

cljs.core.PersistentArrayMap.prototype.schema$core$Schema$walker$arity$1 = (function (this$){
var this$__$1 = this;
return schema.core.map_walker(this$__$1);
});

cljs.core.PersistentArrayMap.prototype.schema$core$Schema$explain$arity$1 = (function (this$){
var this$__$1 = this;
return schema.core.map_explain(this$__$1);
});

cljs.core.PersistentHashMap.prototype.schema$core$Schema$ = true;

cljs.core.PersistentHashMap.prototype.schema$core$Schema$walker$arity$1 = (function (this$){
var this$__$1 = this;
return schema.core.map_walker(this$__$1);
});

cljs.core.PersistentHashMap.prototype.schema$core$Schema$explain$arity$1 = (function (this$){
var this$__$1 = this;
return schema.core.map_explain(this$__$1);
});
cljs.core.PersistentHashSet.prototype.schema$core$Schema$ = true;

cljs.core.PersistentHashSet.prototype.schema$core$Schema$walker$arity$1 = (function (this$){
var this$__$1 = this;
if(cljs.core._EQ_.cljs$core$IFn$_invoke$arity$2(cljs.core.count(this$__$1),(1))){
} else {
throw (new Error(schema.utils.format_STAR_("Set schema must have exactly one element")));
}

var sub_walker = (schema.core.subschema_walker.cljs$core$IFn$_invoke$arity$1 ? schema.core.subschema_walker.cljs$core$IFn$_invoke$arity$1(cljs.core.first(this$__$1)) : schema.core.subschema_walker.call(null,cljs.core.first(this$__$1)));
return ((function (sub_walker,this$__$1){
return (function (x){
var or__4512__auto__ = ((cljs.core.set_QMARK_(x))?null:schema.utils.error(schema.utils.__GT_ValidationError(this$__$1,x,(new cljs.core.Delay(((function (sub_walker,this$__$1){
return (function (){
return cljs.core._conj(cljs.core._conj(cljs.core.List.EMPTY,schema.utils.value_name(x)),cljs.core.cst$sym$set_QMARK_);
});})(sub_walker,this$__$1))
,null)),null)));
if(cljs.core.truth_(or__4512__auto__)){
return or__4512__auto__;
} else {
var vec__27393 = cljs.core.juxt.cljs$core$IFn$_invoke$arity$2(cljs.core.remove,cljs.core.keep).call(null,schema.utils.error_val,cljs.core.map.cljs$core$IFn$_invoke$arity$2(sub_walker,x));
var good = cljs.core.nth.cljs$core$IFn$_invoke$arity$3(vec__27393,(0),null);
var bad = cljs.core.nth.cljs$core$IFn$_invoke$arity$3(vec__27393,(1),null);
if(cljs.core.seq(bad)){
return schema.utils.error(cljs.core.set(bad));
} else {
return cljs.core.set(good);
}
}
});
;})(sub_walker,this$__$1))
});

cljs.core.PersistentHashSet.prototype.schema$core$Schema$explain$arity$1 = (function (this$){
var this$__$1 = this;
return cljs.core.set(new cljs.core.PersistentVector(null, 1, 5, cljs.core.PersistentVector.EMPTY_NODE, [schema.core.explain(cljs.core.first(this$__$1))], null));
});

/**
* @constructor
 * @implements {cljs.core.IRecord}
 * @implements {cljs.core.IEquiv}
 * @implements {cljs.core.IHash}
 * @implements {cljs.core.ICollection}
 * @implements {cljs.core.ICounted}
 * @implements {cljs.core.ISeqable}
 * @implements {cljs.core.IMeta}
 * @implements {cljs.core.ICloneable}
 * @implements {cljs.core.IPrintWithWriter}
 * @implements {cljs.core.IIterable}
 * @implements {cljs.core.IWithMeta}
 * @implements {cljs.core.IAssociative}
 * @implements {cljs.core.IMap}
 * @implements {cljs.core.ILookup}
*/
schema.core.One = (function (schema,optional_QMARK_,name,__meta,__extmap,__hash){
this.schema = schema;
this.optional_QMARK_ = optional_QMARK_;
this.name = name;
this.__meta = __meta;
this.__extmap = __extmap;
this.__hash = __hash;
this.cljs$lang$protocol_mask$partition0$ = 2229667594;
this.cljs$lang$protocol_mask$partition1$ = 8192;
})
schema.core.One.prototype.cljs$core$ILookup$_lookup$arity$2 = (function (this__5126__auto__,k__5127__auto__){
var self__ = this;
var this__5126__auto____$1 = this;
return cljs.core._lookup.cljs$core$IFn$_invoke$arity$3(this__5126__auto____$1,k__5127__auto__,null);
});

schema.core.One.prototype.cljs$core$ILookup$_lookup$arity$3 = (function (this__5128__auto__,k27395,else__5129__auto__){
var self__ = this;
var this__5128__auto____$1 = this;
var G__27397 = (((k27395 instanceof cljs.core.Keyword))?k27395.fqn:null);
switch (G__27397) {
case "schema":
return self__.schema;

break;
case "optional?":
return self__.optional_QMARK_;

break;
case "name":
return self__.name;

break;
default:
return cljs.core.get.cljs$core$IFn$_invoke$arity$3(self__.__extmap,k27395,else__5129__auto__);

}
});

schema.core.One.prototype.cljs$core$IPrintWithWriter$_pr_writer$arity$3 = (function (this__5140__auto__,writer__5141__auto__,opts__5142__auto__){
var self__ = this;
var this__5140__auto____$1 = this;
var pr_pair__5143__auto__ = ((function (this__5140__auto____$1){
return (function (keyval__5144__auto__){
return cljs.core.pr_sequential_writer(writer__5141__auto__,cljs.core.pr_writer,""," ","",opts__5142__auto__,keyval__5144__auto__);
});})(this__5140__auto____$1))
;
return cljs.core.pr_sequential_writer(writer__5141__auto__,pr_pair__5143__auto__,"#schema.core.One{",", ","}",opts__5142__auto__,cljs.core.concat.cljs$core$IFn$_invoke$arity$2(new cljs.core.PersistentVector(null, 3, 5, cljs.core.PersistentVector.EMPTY_NODE, [(new cljs.core.PersistentVector(null,2,(5),cljs.core.PersistentVector.EMPTY_NODE,[cljs.core.cst$kw$schema,self__.schema],null)),(new cljs.core.PersistentVector(null,2,(5),cljs.core.PersistentVector.EMPTY_NODE,[cljs.core.cst$kw$optional_QMARK_,self__.optional_QMARK_],null)),(new cljs.core.PersistentVector(null,2,(5),cljs.core.PersistentVector.EMPTY_NODE,[cljs.core.cst$kw$name,self__.name],null))], null),self__.__extmap));
});

schema.core.One.prototype.cljs$core$IIterable$ = true;

schema.core.One.prototype.cljs$core$IIterable$_iterator$arity$1 = (function (G__27394){
var self__ = this;
var G__27394__$1 = this;
return (new cljs.core.RecordIter((0),G__27394__$1,3,new cljs.core.PersistentVector(null, 3, 5, cljs.core.PersistentVector.EMPTY_NODE, [cljs.core.cst$kw$schema,cljs.core.cst$kw$optional_QMARK_,cljs.core.cst$kw$name], null),cljs.core._iterator(self__.__extmap)));
});

schema.core.One.prototype.cljs$core$IMeta$_meta$arity$1 = (function (this__5124__auto__){
var self__ = this;
var this__5124__auto____$1 = this;
return self__.__meta;
});

schema.core.One.prototype.cljs$core$ICloneable$_clone$arity$1 = (function (this__5120__auto__){
var self__ = this;
var this__5120__auto____$1 = this;
return (new schema.core.One(self__.schema,self__.optional_QMARK_,self__.name,self__.__meta,self__.__extmap,self__.__hash));
});

schema.core.One.prototype.cljs$core$ICounted$_count$arity$1 = (function (this__5130__auto__){
var self__ = this;
var this__5130__auto____$1 = this;
return (3 + cljs.core.count(self__.__extmap));
});

schema.core.One.prototype.cljs$core$IHash$_hash$arity$1 = (function (this__5121__auto__){
var self__ = this;
var this__5121__auto____$1 = this;
var h__4947__auto__ = self__.__hash;
if(!((h__4947__auto__ == null))){
return h__4947__auto__;
} else {
var h__4947__auto____$1 = cljs.core.hash_imap(this__5121__auto____$1);
self__.__hash = h__4947__auto____$1;

return h__4947__auto____$1;
}
});

schema.core.One.prototype.cljs$core$IEquiv$_equiv$arity$2 = (function (this__5122__auto__,other__5123__auto__){
var self__ = this;
var this__5122__auto____$1 = this;
if(cljs.core.truth_((function (){var and__4500__auto__ = other__5123__auto__;
if(cljs.core.truth_(and__4500__auto__)){
var and__4500__auto____$1 = (this__5122__auto____$1.constructor === other__5123__auto__.constructor);
if(and__4500__auto____$1){
return cljs.core.equiv_map(this__5122__auto____$1,other__5123__auto__);
} else {
return and__4500__auto____$1;
}
} else {
return and__4500__auto__;
}
})())){
return true;
} else {
return false;
}
});

schema.core.One.prototype.cljs$core$IMap$_dissoc$arity$2 = (function (this__5135__auto__,k__5136__auto__){
var self__ = this;
var this__5135__auto____$1 = this;
if(cljs.core.contains_QMARK_(new cljs.core.PersistentHashSet(null, new cljs.core.PersistentArrayMap(null, 3, [cljs.core.cst$kw$schema,null,cljs.core.cst$kw$name,null,cljs.core.cst$kw$optional_QMARK_,null], null), null),k__5136__auto__)){
return cljs.core.dissoc.cljs$core$IFn$_invoke$arity$2(cljs.core.with_meta(cljs.core.into.cljs$core$IFn$_invoke$arity$2(cljs.core.PersistentArrayMap.EMPTY,this__5135__auto____$1),self__.__meta),k__5136__auto__);
} else {
return (new schema.core.One(self__.schema,self__.optional_QMARK_,self__.name,self__.__meta,cljs.core.not_empty(cljs.core.dissoc.cljs$core$IFn$_invoke$arity$2(self__.__extmap,k__5136__auto__)),null));
}
});

schema.core.One.prototype.cljs$core$IAssociative$_assoc$arity$3 = (function (this__5133__auto__,k__5134__auto__,G__27394){
var self__ = this;
var this__5133__auto____$1 = this;
var pred__27398 = cljs.core.keyword_identical_QMARK_;
var expr__27399 = k__5134__auto__;
if(cljs.core.truth_((pred__27398.cljs$core$IFn$_invoke$arity$2 ? pred__27398.cljs$core$IFn$_invoke$arity$2(cljs.core.cst$kw$schema,expr__27399) : pred__27398.call(null,cljs.core.cst$kw$schema,expr__27399)))){
return (new schema.core.One(G__27394,self__.optional_QMARK_,self__.name,self__.__meta,self__.__extmap,null));
} else {
if(cljs.core.truth_((pred__27398.cljs$core$IFn$_invoke$arity$2 ? pred__27398.cljs$core$IFn$_invoke$arity$2(cljs.core.cst$kw$optional_QMARK_,expr__27399) : pred__27398.call(null,cljs.core.cst$kw$optional_QMARK_,expr__27399)))){
return (new schema.core.One(self__.schema,G__27394,self__.name,self__.__meta,self__.__extmap,null));
} else {
if(cljs.core.truth_((pred__27398.cljs$core$IFn$_invoke$arity$2 ? pred__27398.cljs$core$IFn$_invoke$arity$2(cljs.core.cst$kw$name,expr__27399) : pred__27398.call(null,cljs.core.cst$kw$name,expr__27399)))){
return (new schema.core.One(self__.schema,self__.optional_QMARK_,G__27394,self__.__meta,self__.__extmap,null));
} else {
return (new schema.core.One(self__.schema,self__.optional_QMARK_,self__.name,self__.__meta,cljs.core.assoc.cljs$core$IFn$_invoke$arity$3(self__.__extmap,k__5134__auto__,G__27394),null));
}
}
}
});

schema.core.One.prototype.cljs$core$ISeqable$_seq$arity$1 = (function (this__5138__auto__){
var self__ = this;
var this__5138__auto____$1 = this;
return cljs.core.seq(cljs.core.concat.cljs$core$IFn$_invoke$arity$2(new cljs.core.PersistentVector(null, 3, 5, cljs.core.PersistentVector.EMPTY_NODE, [(new cljs.core.PersistentVector(null,2,(5),cljs.core.PersistentVector.EMPTY_NODE,[cljs.core.cst$kw$schema,self__.schema],null)),(new cljs.core.PersistentVector(null,2,(5),cljs.core.PersistentVector.EMPTY_NODE,[cljs.core.cst$kw$optional_QMARK_,self__.optional_QMARK_],null)),(new cljs.core.PersistentVector(null,2,(5),cljs.core.PersistentVector.EMPTY_NODE,[cljs.core.cst$kw$name,self__.name],null))], null),self__.__extmap));
});

schema.core.One.prototype.cljs$core$IWithMeta$_with_meta$arity$2 = (function (this__5125__auto__,G__27394){
var self__ = this;
var this__5125__auto____$1 = this;
return (new schema.core.One(self__.schema,self__.optional_QMARK_,self__.name,G__27394,self__.__extmap,self__.__hash));
});

schema.core.One.prototype.cljs$core$ICollection$_conj$arity$2 = (function (this__5131__auto__,entry__5132__auto__){
var self__ = this;
var this__5131__auto____$1 = this;
if(cljs.core.vector_QMARK_(entry__5132__auto__)){
return cljs.core._assoc(this__5131__auto____$1,cljs.core._nth.cljs$core$IFn$_invoke$arity$2(entry__5132__auto__,(0)),cljs.core._nth.cljs$core$IFn$_invoke$arity$2(entry__5132__auto__,(1)));
} else {
return cljs.core.reduce.cljs$core$IFn$_invoke$arity$3(cljs.core._conj,this__5131__auto____$1,entry__5132__auto__);
}
});

schema.core.One.getBasis = (function (){
return new cljs.core.PersistentVector(null, 3, 5, cljs.core.PersistentVector.EMPTY_NODE, [cljs.core.cst$sym$schema,cljs.core.cst$sym$optional_QMARK_,cljs.core.cst$sym$name], null);
});

schema.core.One.cljs$lang$type = true;

schema.core.One.cljs$lang$ctorPrSeq = (function (this__5160__auto__){
return cljs.core._conj(cljs.core.List.EMPTY,"schema.core/One");
});

schema.core.One.cljs$lang$ctorPrWriter = (function (this__5160__auto__,writer__5161__auto__){
return cljs.core._write(writer__5161__auto__,"schema.core/One");
});

schema.core.__GT_One = (function schema$core$__GT_One(schema__$1,optional_QMARK_,name){
return (new schema.core.One(schema__$1,optional_QMARK_,name,null,null,null));
});

schema.core.map__GT_One = (function schema$core$map__GT_One(G__27396){
return (new schema.core.One(cljs.core.cst$kw$schema.cljs$core$IFn$_invoke$arity$1(G__27396),cljs.core.cst$kw$optional_QMARK_.cljs$core$IFn$_invoke$arity$1(G__27396),cljs.core.cst$kw$name.cljs$core$IFn$_invoke$arity$1(G__27396),null,cljs.core.dissoc.cljs$core$IFn$_invoke$arity$variadic(G__27396,cljs.core.cst$kw$schema,cljs.core.array_seq([cljs.core.cst$kw$optional_QMARK_,cljs.core.cst$kw$name], 0)),null));
});

/**
 * A single required element of a sequence (not repeated, the implicit default)
 */
schema.core.one = (function schema$core$one(schema__$1,name){
return (new schema.core.One(schema__$1,false,name,null,null,null));
});
/**
 * A single optional element of a sequence (not repeated, the implicit default)
 */
schema.core.optional = (function schema$core$optional(schema__$1,name){
return (new schema.core.One(schema__$1,true,name,null,null,null));
});
schema.core.parse_sequence_schema = (function schema$core$parse_sequence_schema(s){

var vec__27407 = cljs.core.split_with((function (p1__27402_SHARP_){
return ((p1__27402_SHARP_ instanceof schema.core.One)) && (cljs.core.not(cljs.core.cst$kw$optional_QMARK_.cljs$core$IFn$_invoke$arity$1(p1__27402_SHARP_)));
}),s);
var required = cljs.core.nth.cljs$core$IFn$_invoke$arity$3(vec__27407,(0),null);
var more = cljs.core.nth.cljs$core$IFn$_invoke$arity$3(vec__27407,(1),null);
var vec__27408 = cljs.core.split_with(((function (vec__27407,required,more){
return (function (p1__27403_SHARP_){
var and__4500__auto__ = (p1__27403_SHARP_ instanceof schema.core.One);
if(and__4500__auto__){
return cljs.core.cst$kw$optional_QMARK_.cljs$core$IFn$_invoke$arity$1(p1__27403_SHARP_);
} else {
return and__4500__auto__;
}
});})(vec__27407,required,more))
,more);
var optional = cljs.core.nth.cljs$core$IFn$_invoke$arity$3(vec__27408,(0),null);
var more__$1 = cljs.core.nth.cljs$core$IFn$_invoke$arity$3(vec__27408,(1),null);
if(((cljs.core.count(more__$1) <= (1))) && (cljs.core.every_QMARK_(((function (vec__27407,required,more,vec__27408,optional,more__$1){
return (function (p1__27404_SHARP_){
return !((p1__27404_SHARP_ instanceof schema.core.One));
});})(vec__27407,required,more,vec__27408,optional,more__$1))
,more__$1))){
} else {
throw (new Error(schema.utils.format_STAR_.cljs$core$IFn$_invoke$arity$variadic("Sequence schema %s does not match [one* optional* rest-schema?]",cljs.core.array_seq([s], 0))));
}

return new cljs.core.PersistentVector(null, 2, 5, cljs.core.PersistentVector.EMPTY_NODE, [cljs.core.concat.cljs$core$IFn$_invoke$arity$2(required,optional),cljs.core.first(more__$1)], null);
});
cljs.core.PersistentVector.prototype.schema$core$Schema$ = true;

cljs.core.PersistentVector.prototype.schema$core$Schema$walker$arity$1 = (function (this$){
var this$__$1 = this;
var vec__27409 = schema.core.parse_sequence_schema(this$__$1);
var singles = cljs.core.nth.cljs$core$IFn$_invoke$arity$3(vec__27409,(0),null);
var multi = cljs.core.nth.cljs$core$IFn$_invoke$arity$3(vec__27409,(1),null);
var single_walkers = cljs.core.vec((function (){var iter__5284__auto__ = ((function (vec__27409,singles,multi,this$__$1){
return (function schema$core$iter__27410(s__27411){
return (new cljs.core.LazySeq(null,((function (vec__27409,singles,multi,this$__$1){
return (function (){
var s__27411__$1 = s__27411;
while(true){
var temp__4425__auto__ = cljs.core.seq(s__27411__$1);
if(temp__4425__auto__){
var s__27411__$2 = temp__4425__auto__;
if(cljs.core.chunked_seq_QMARK_(s__27411__$2)){
var c__5282__auto__ = cljs.core.chunk_first(s__27411__$2);
var size__5283__auto__ = cljs.core.count(c__5282__auto__);
var b__27413 = cljs.core.chunk_buffer(size__5283__auto__);
if((function (){var i__27412 = (0);
while(true){
if((i__27412 < size__5283__auto__)){
var s = cljs.core._nth.cljs$core$IFn$_invoke$arity$2(c__5282__auto__,i__27412);
cljs.core.chunk_append(b__27413,new cljs.core.PersistentVector(null, 2, 5, cljs.core.PersistentVector.EMPTY_NODE, [s,(schema.core.subschema_walker.cljs$core$IFn$_invoke$arity$1 ? schema.core.subschema_walker.cljs$core$IFn$_invoke$arity$1(s.schema) : schema.core.subschema_walker.call(null,s.schema))], null));

var G__27452 = (i__27412 + (1));
i__27412 = G__27452;
continue;
} else {
return true;
}
break;
}
})()){
return cljs.core.chunk_cons(cljs.core.chunk(b__27413),schema$core$iter__27410(cljs.core.chunk_rest(s__27411__$2)));
} else {
return cljs.core.chunk_cons(cljs.core.chunk(b__27413),null);
}
} else {
var s = cljs.core.first(s__27411__$2);
return cljs.core.cons(new cljs.core.PersistentVector(null, 2, 5, cljs.core.PersistentVector.EMPTY_NODE, [s,(schema.core.subschema_walker.cljs$core$IFn$_invoke$arity$1 ? schema.core.subschema_walker.cljs$core$IFn$_invoke$arity$1(s.schema) : schema.core.subschema_walker.call(null,s.schema))], null),schema$core$iter__27410(cljs.core.rest(s__27411__$2)));
}
} else {
return null;
}
break;
}
});})(vec__27409,singles,multi,this$__$1))
,null,null));
});})(vec__27409,singles,multi,this$__$1))
;
return iter__5284__auto__(singles);
})());
var multi_walker = (cljs.core.truth_(multi)?(schema.core.subschema_walker.cljs$core$IFn$_invoke$arity$1 ? schema.core.subschema_walker.cljs$core$IFn$_invoke$arity$1(multi) : schema.core.subschema_walker.call(null,multi)):null);
var err_conj = schema.utils.result_builder(((function (vec__27409,singles,multi,single_walkers,multi_walker,this$__$1){
return (function (m){
return cljs.core.vec(cljs.core.repeat.cljs$core$IFn$_invoke$arity$2(cljs.core.count(m),null));
});})(vec__27409,singles,multi,single_walkers,multi_walker,this$__$1))
);
return ((function (vec__27409,singles,multi,single_walkers,multi_walker,err_conj,this$__$1){
return (function (x){
var or__4512__auto__ = ((((x == null)) || (cljs.core.sequential_QMARK_(x)))?null:schema.utils.error(schema.utils.__GT_ValidationError(this$__$1,x,(new cljs.core.Delay(((function (vec__27409,singles,multi,single_walkers,multi_walker,err_conj,this$__$1){
return (function (){
return cljs.core._conj(cljs.core._conj(cljs.core.List.EMPTY,schema.utils.value_name(x)),cljs.core.cst$sym$sequential_QMARK_);
});})(vec__27409,singles,multi,single_walkers,multi_walker,err_conj,this$__$1))
,null)),null)));
if(cljs.core.truth_(or__4512__auto__)){
return or__4512__auto__;
} else {
var single_walkers__$1 = single_walkers;
var x__$1 = x;
var out = cljs.core.PersistentVector.EMPTY;
while(true){
var temp__4423__auto__ = cljs.core.first(single_walkers__$1);
if(cljs.core.truth_(temp__4423__auto__)){
var vec__27416 = temp__4423__auto__;
var first_single = cljs.core.nth.cljs$core$IFn$_invoke$arity$3(vec__27416,(0),null);
var single_walker = cljs.core.nth.cljs$core$IFn$_invoke$arity$3(vec__27416,(1),null);
if(cljs.core.empty_QMARK_(x__$1)){
if(cljs.core.truth_(first_single.optional_QMARK_)){
return out;
} else {
var G__27427 = out;
var G__27428 = schema.utils.error(schema.utils.__GT_ValidationError(cljs.core.vec(cljs.core.map.cljs$core$IFn$_invoke$arity$2(cljs.core.first,single_walkers__$1)),null,(new cljs.core.Delay(((function (single_walkers__$1,x__$1,out,G__27427,vec__27416,first_single,single_walker,temp__4423__auto__,or__4512__auto__,vec__27409,singles,multi,single_walkers,multi_walker,err_conj,this$__$1){
return (function (){
return cljs.core.list_STAR_.cljs$core$IFn$_invoke$arity$2(cljs.core.cst$sym$present_QMARK_,(function (){var iter__5284__auto__ = ((function (single_walkers__$1,x__$1,out,G__27427,vec__27416,first_single,single_walker,temp__4423__auto__,or__4512__auto__,vec__27409,singles,multi,single_walkers,multi_walker,err_conj,this$__$1){
return (function schema$core$iter__27429(s__27430){
return (new cljs.core.LazySeq(null,((function (single_walkers__$1,x__$1,out,G__27427,vec__27416,first_single,single_walker,temp__4423__auto__,or__4512__auto__,vec__27409,singles,multi,single_walkers,multi_walker,err_conj,this$__$1){
return (function (){
var s__27430__$1 = s__27430;
while(true){
var temp__4425__auto__ = cljs.core.seq(s__27430__$1);
if(temp__4425__auto__){
var s__27430__$2 = temp__4425__auto__;
if(cljs.core.chunked_seq_QMARK_(s__27430__$2)){
var c__5282__auto__ = cljs.core.chunk_first(s__27430__$2);
var size__5283__auto__ = cljs.core.count(c__5282__auto__);
var b__27432 = cljs.core.chunk_buffer(size__5283__auto__);
if((function (){var i__27431 = (0);
while(true){
if((i__27431 < size__5283__auto__)){
var vec__27437 = cljs.core._nth.cljs$core$IFn$_invoke$arity$2(c__5282__auto__,i__27431);
var single = cljs.core.nth.cljs$core$IFn$_invoke$arity$3(vec__27437,(0),null);
if(cljs.core.not(single.optional_QMARK_)){
cljs.core.chunk_append(b__27432,single.name);

var G__27453 = (i__27431 + (1));
i__27431 = G__27453;
continue;
} else {
return null;
}
} else {
return true;
}
break;
}
})()){
return cljs.core.chunk_cons(cljs.core.chunk(b__27432),schema$core$iter__27429(cljs.core.chunk_rest(s__27430__$2)));
} else {
return cljs.core.chunk_cons(cljs.core.chunk(b__27432),null);
}
} else {
var vec__27438 = cljs.core.first(s__27430__$2);
var single = cljs.core.nth.cljs$core$IFn$_invoke$arity$3(vec__27438,(0),null);
if(cljs.core.not(single.optional_QMARK_)){
return cljs.core.cons(single.name,schema$core$iter__27429(cljs.core.rest(s__27430__$2)));
} else {
return null;
}
}
} else {
return null;
}
break;
}
});})(single_walkers__$1,x__$1,out,G__27427,vec__27416,first_single,single_walker,temp__4423__auto__,or__4512__auto__,vec__27409,singles,multi,single_walkers,multi_walker,err_conj,this$__$1))
,null,null));
});})(single_walkers__$1,x__$1,out,G__27427,vec__27416,first_single,single_walker,temp__4423__auto__,or__4512__auto__,vec__27409,singles,multi,single_walkers,multi_walker,err_conj,this$__$1))
;
return iter__5284__auto__(single_walkers__$1);
})());
});})(single_walkers__$1,x__$1,out,G__27427,vec__27416,first_single,single_walker,temp__4423__auto__,or__4512__auto__,vec__27409,singles,multi,single_walkers,multi_walker,err_conj,this$__$1))
,null)),null));
return (err_conj.cljs$core$IFn$_invoke$arity$2 ? err_conj.cljs$core$IFn$_invoke$arity$2(G__27427,G__27428) : err_conj.call(null,G__27427,G__27428));
}
} else {
var G__27454 = cljs.core.next(single_walkers__$1);
var G__27455 = cljs.core.rest(x__$1);
var G__27456 = (function (){var G__27440 = out;
var G__27441 = schema.utils.wrap_error_name(first_single.name,(function (){var G__27442 = cljs.core.first(x__$1);
return (single_walker.cljs$core$IFn$_invoke$arity$1 ? single_walker.cljs$core$IFn$_invoke$arity$1(G__27442) : single_walker.call(null,G__27442));
})());
return (err_conj.cljs$core$IFn$_invoke$arity$2 ? err_conj.cljs$core$IFn$_invoke$arity$2(G__27440,G__27441) : err_conj.call(null,G__27440,G__27441));
})();
single_walkers__$1 = G__27454;
x__$1 = G__27455;
out = G__27456;
continue;
}
} else {
if(cljs.core.truth_(multi)){
return cljs.core.reduce.cljs$core$IFn$_invoke$arity$3(err_conj,out,cljs.core.map.cljs$core$IFn$_invoke$arity$2(multi_walker,x__$1));
} else {
if(cljs.core.seq(x__$1)){
var G__27443 = out;
var G__27444 = schema.utils.error(schema.utils.__GT_ValidationError(null,x__$1,(new cljs.core.Delay(((function (single_walkers__$1,x__$1,out,G__27443,temp__4423__auto__,or__4512__auto__,vec__27409,singles,multi,single_walkers,multi_walker,err_conj,this$__$1){
return (function (){
return cljs.core._conj(cljs.core._conj(cljs.core.List.EMPTY,cljs.core.count(x__$1)),cljs.core.cst$sym$has_DASH_extra_DASH_elts_QMARK_);
});})(single_walkers__$1,x__$1,out,G__27443,temp__4423__auto__,or__4512__auto__,vec__27409,singles,multi,single_walkers,multi_walker,err_conj,this$__$1))
,null)),null));
return (err_conj.cljs$core$IFn$_invoke$arity$2 ? err_conj.cljs$core$IFn$_invoke$arity$2(G__27443,G__27444) : err_conj.call(null,G__27443,G__27444));
} else {
return out;

}
}
}
break;
}
}
});
;})(vec__27409,singles,multi,single_walkers,multi_walker,err_conj,this$__$1))
});

cljs.core.PersistentVector.prototype.schema$core$Schema$explain$arity$1 = (function (this$){
var this$__$1 = this;
var vec__27445 = schema.core.parse_sequence_schema(this$__$1);
var singles = cljs.core.nth.cljs$core$IFn$_invoke$arity$3(vec__27445,(0),null);
var multi = cljs.core.nth.cljs$core$IFn$_invoke$arity$3(vec__27445,(1),null);
return cljs.core.vec(cljs.core.concat.cljs$core$IFn$_invoke$arity$2((function (){var iter__5284__auto__ = ((function (vec__27445,singles,multi,this$__$1){
return (function schema$core$iter__27446(s__27447){
return (new cljs.core.LazySeq(null,((function (vec__27445,singles,multi,this$__$1){
return (function (){
var s__27447__$1 = s__27447;
while(true){
var temp__4425__auto__ = cljs.core.seq(s__27447__$1);
if(temp__4425__auto__){
var s__27447__$2 = temp__4425__auto__;
if(cljs.core.chunked_seq_QMARK_(s__27447__$2)){
var c__5282__auto__ = cljs.core.chunk_first(s__27447__$2);
var size__5283__auto__ = cljs.core.count(c__5282__auto__);
var b__27449 = cljs.core.chunk_buffer(size__5283__auto__);
if((function (){var i__27448 = (0);
while(true){
if((i__27448 < size__5283__auto__)){
var s = cljs.core._nth.cljs$core$IFn$_invoke$arity$2(c__5282__auto__,i__27448);
cljs.core.chunk_append(b__27449,cljs.core._conj(cljs.core._conj(cljs.core._conj(cljs.core.List.EMPTY,cljs.core.cst$kw$name.cljs$core$IFn$_invoke$arity$1(s)),schema.core.explain(cljs.core.cst$kw$schema.cljs$core$IFn$_invoke$arity$1(s))),(cljs.core.truth_(s.optional_QMARK_)?cljs.core.cst$sym$optional:cljs.core.cst$sym$one)));

var G__27457 = (i__27448 + (1));
i__27448 = G__27457;
continue;
} else {
return true;
}
break;
}
})()){
return cljs.core.chunk_cons(cljs.core.chunk(b__27449),schema$core$iter__27446(cljs.core.chunk_rest(s__27447__$2)));
} else {
return cljs.core.chunk_cons(cljs.core.chunk(b__27449),null);
}
} else {
var s = cljs.core.first(s__27447__$2);
return cljs.core.cons(cljs.core._conj(cljs.core._conj(cljs.core._conj(cljs.core.List.EMPTY,cljs.core.cst$kw$name.cljs$core$IFn$_invoke$arity$1(s)),schema.core.explain(cljs.core.cst$kw$schema.cljs$core$IFn$_invoke$arity$1(s))),(cljs.core.truth_(s.optional_QMARK_)?cljs.core.cst$sym$optional:cljs.core.cst$sym$one)),schema$core$iter__27446(cljs.core.rest(s__27447__$2)));
}
} else {
return null;
}
break;
}
});})(vec__27445,singles,multi,this$__$1))
,null,null));
});})(vec__27445,singles,multi,this$__$1))
;
return iter__5284__auto__(singles);
})(),(cljs.core.truth_(multi)?new cljs.core.PersistentVector(null, 1, 5, cljs.core.PersistentVector.EMPTY_NODE, [schema.core.explain(multi)], null):null)));
});
/**
 * A schema for a pair of schemas and their names
 */
schema.core.pair = (function schema$core$pair(first_schema,first_name,second_schema,second_name){
return new cljs.core.PersistentVector(null, 2, 5, cljs.core.PersistentVector.EMPTY_NODE, [schema.core.one(first_schema,first_name),schema.core.one(second_schema,second_name)], null);
});

/**
* @constructor
 * @implements {cljs.core.IRecord}
 * @implements {cljs.core.IEquiv}
 * @implements {cljs.core.IHash}
 * @implements {cljs.core.ICollection}
 * @implements {schema.core.Schema}
 * @implements {cljs.core.ICounted}
 * @implements {cljs.core.ISeqable}
 * @implements {cljs.core.IMeta}
 * @implements {cljs.core.ICloneable}
 * @implements {cljs.core.IPrintWithWriter}
 * @implements {cljs.core.IIterable}
 * @implements {cljs.core.IWithMeta}
 * @implements {cljs.core.IAssociative}
 * @implements {cljs.core.IMap}
 * @implements {cljs.core.ILookup}
*/
schema.core.Record = (function (klass,schema,__meta,__extmap,__hash){
this.klass = klass;
this.schema = schema;
this.__meta = __meta;
this.__extmap = __extmap;
this.__hash = __hash;
this.cljs$lang$protocol_mask$partition0$ = 2229667594;
this.cljs$lang$protocol_mask$partition1$ = 8192;
})
schema.core.Record.prototype.cljs$core$ILookup$_lookup$arity$2 = (function (this__5126__auto__,k__5127__auto__){
var self__ = this;
var this__5126__auto____$1 = this;
return cljs.core._lookup.cljs$core$IFn$_invoke$arity$3(this__5126__auto____$1,k__5127__auto__,null);
});

schema.core.Record.prototype.cljs$core$ILookup$_lookup$arity$3 = (function (this__5128__auto__,k27459,else__5129__auto__){
var self__ = this;
var this__5128__auto____$1 = this;
var G__27461 = (((k27459 instanceof cljs.core.Keyword))?k27459.fqn:null);
switch (G__27461) {
case "klass":
return self__.klass;

break;
case "schema":
return self__.schema;

break;
default:
return cljs.core.get.cljs$core$IFn$_invoke$arity$3(self__.__extmap,k27459,else__5129__auto__);

}
});

schema.core.Record.prototype.cljs$core$IPrintWithWriter$_pr_writer$arity$3 = (function (this__5140__auto__,writer__5141__auto__,opts__5142__auto__){
var self__ = this;
var this__5140__auto____$1 = this;
var pr_pair__5143__auto__ = ((function (this__5140__auto____$1){
return (function (keyval__5144__auto__){
return cljs.core.pr_sequential_writer(writer__5141__auto__,cljs.core.pr_writer,""," ","",opts__5142__auto__,keyval__5144__auto__);
});})(this__5140__auto____$1))
;
return cljs.core.pr_sequential_writer(writer__5141__auto__,pr_pair__5143__auto__,"#schema.core.Record{",", ","}",opts__5142__auto__,cljs.core.concat.cljs$core$IFn$_invoke$arity$2(new cljs.core.PersistentVector(null, 2, 5, cljs.core.PersistentVector.EMPTY_NODE, [(new cljs.core.PersistentVector(null,2,(5),cljs.core.PersistentVector.EMPTY_NODE,[cljs.core.cst$kw$klass,self__.klass],null)),(new cljs.core.PersistentVector(null,2,(5),cljs.core.PersistentVector.EMPTY_NODE,[cljs.core.cst$kw$schema,self__.schema],null))], null),self__.__extmap));
});

schema.core.Record.prototype.cljs$core$IIterable$ = true;

schema.core.Record.prototype.cljs$core$IIterable$_iterator$arity$1 = (function (G__27458){
var self__ = this;
var G__27458__$1 = this;
return (new cljs.core.RecordIter((0),G__27458__$1,2,new cljs.core.PersistentVector(null, 2, 5, cljs.core.PersistentVector.EMPTY_NODE, [cljs.core.cst$kw$klass,cljs.core.cst$kw$schema], null),cljs.core._iterator(self__.__extmap)));
});

schema.core.Record.prototype.cljs$core$IMeta$_meta$arity$1 = (function (this__5124__auto__){
var self__ = this;
var this__5124__auto____$1 = this;
return self__.__meta;
});

schema.core.Record.prototype.cljs$core$ICloneable$_clone$arity$1 = (function (this__5120__auto__){
var self__ = this;
var this__5120__auto____$1 = this;
return (new schema.core.Record(self__.klass,self__.schema,self__.__meta,self__.__extmap,self__.__hash));
});

schema.core.Record.prototype.cljs$core$ICounted$_count$arity$1 = (function (this__5130__auto__){
var self__ = this;
var this__5130__auto____$1 = this;
return (2 + cljs.core.count(self__.__extmap));
});

schema.core.Record.prototype.cljs$core$IHash$_hash$arity$1 = (function (this__5121__auto__){
var self__ = this;
var this__5121__auto____$1 = this;
var h__4947__auto__ = self__.__hash;
if(!((h__4947__auto__ == null))){
return h__4947__auto__;
} else {
var h__4947__auto____$1 = cljs.core.hash_imap(this__5121__auto____$1);
self__.__hash = h__4947__auto____$1;

return h__4947__auto____$1;
}
});

schema.core.Record.prototype.cljs$core$IEquiv$_equiv$arity$2 = (function (this__5122__auto__,other__5123__auto__){
var self__ = this;
var this__5122__auto____$1 = this;
if(cljs.core.truth_((function (){var and__4500__auto__ = other__5123__auto__;
if(cljs.core.truth_(and__4500__auto__)){
var and__4500__auto____$1 = (this__5122__auto____$1.constructor === other__5123__auto__.constructor);
if(and__4500__auto____$1){
return cljs.core.equiv_map(this__5122__auto____$1,other__5123__auto__);
} else {
return and__4500__auto____$1;
}
} else {
return and__4500__auto__;
}
})())){
return true;
} else {
return false;
}
});

schema.core.Record.prototype.cljs$core$IMap$_dissoc$arity$2 = (function (this__5135__auto__,k__5136__auto__){
var self__ = this;
var this__5135__auto____$1 = this;
if(cljs.core.contains_QMARK_(new cljs.core.PersistentHashSet(null, new cljs.core.PersistentArrayMap(null, 2, [cljs.core.cst$kw$schema,null,cljs.core.cst$kw$klass,null], null), null),k__5136__auto__)){
return cljs.core.dissoc.cljs$core$IFn$_invoke$arity$2(cljs.core.with_meta(cljs.core.into.cljs$core$IFn$_invoke$arity$2(cljs.core.PersistentArrayMap.EMPTY,this__5135__auto____$1),self__.__meta),k__5136__auto__);
} else {
return (new schema.core.Record(self__.klass,self__.schema,self__.__meta,cljs.core.not_empty(cljs.core.dissoc.cljs$core$IFn$_invoke$arity$2(self__.__extmap,k__5136__auto__)),null));
}
});

schema.core.Record.prototype.cljs$core$IAssociative$_assoc$arity$3 = (function (this__5133__auto__,k__5134__auto__,G__27458){
var self__ = this;
var this__5133__auto____$1 = this;
var pred__27462 = cljs.core.keyword_identical_QMARK_;
var expr__27463 = k__5134__auto__;
if(cljs.core.truth_((pred__27462.cljs$core$IFn$_invoke$arity$2 ? pred__27462.cljs$core$IFn$_invoke$arity$2(cljs.core.cst$kw$klass,expr__27463) : pred__27462.call(null,cljs.core.cst$kw$klass,expr__27463)))){
return (new schema.core.Record(G__27458,self__.schema,self__.__meta,self__.__extmap,null));
} else {
if(cljs.core.truth_((pred__27462.cljs$core$IFn$_invoke$arity$2 ? pred__27462.cljs$core$IFn$_invoke$arity$2(cljs.core.cst$kw$schema,expr__27463) : pred__27462.call(null,cljs.core.cst$kw$schema,expr__27463)))){
return (new schema.core.Record(self__.klass,G__27458,self__.__meta,self__.__extmap,null));
} else {
return (new schema.core.Record(self__.klass,self__.schema,self__.__meta,cljs.core.assoc.cljs$core$IFn$_invoke$arity$3(self__.__extmap,k__5134__auto__,G__27458),null));
}
}
});

schema.core.Record.prototype.cljs$core$ISeqable$_seq$arity$1 = (function (this__5138__auto__){
var self__ = this;
var this__5138__auto____$1 = this;
return cljs.core.seq(cljs.core.concat.cljs$core$IFn$_invoke$arity$2(new cljs.core.PersistentVector(null, 2, 5, cljs.core.PersistentVector.EMPTY_NODE, [(new cljs.core.PersistentVector(null,2,(5),cljs.core.PersistentVector.EMPTY_NODE,[cljs.core.cst$kw$klass,self__.klass],null)),(new cljs.core.PersistentVector(null,2,(5),cljs.core.PersistentVector.EMPTY_NODE,[cljs.core.cst$kw$schema,self__.schema],null))], null),self__.__extmap));
});

schema.core.Record.prototype.cljs$core$IWithMeta$_with_meta$arity$2 = (function (this__5125__auto__,G__27458){
var self__ = this;
var this__5125__auto____$1 = this;
return (new schema.core.Record(self__.klass,self__.schema,G__27458,self__.__extmap,self__.__hash));
});

schema.core.Record.prototype.cljs$core$ICollection$_conj$arity$2 = (function (this__5131__auto__,entry__5132__auto__){
var self__ = this;
var this__5131__auto____$1 = this;
if(cljs.core.vector_QMARK_(entry__5132__auto__)){
return cljs.core._assoc(this__5131__auto____$1,cljs.core._nth.cljs$core$IFn$_invoke$arity$2(entry__5132__auto__,(0)),cljs.core._nth.cljs$core$IFn$_invoke$arity$2(entry__5132__auto__,(1)));
} else {
return cljs.core.reduce.cljs$core$IFn$_invoke$arity$3(cljs.core._conj,this__5131__auto____$1,entry__5132__auto__);
}
});

schema.core.Record.prototype.schema$core$Schema$ = true;

schema.core.Record.prototype.schema$core$Schema$walker$arity$1 = (function (this$){
var self__ = this;
var this$__$1 = this;
var map_checker = (schema.core.subschema_walker.cljs$core$IFn$_invoke$arity$1 ? schema.core.subschema_walker.cljs$core$IFn$_invoke$arity$1(self__.schema) : schema.core.subschema_walker.call(null,self__.schema));
var pred_checker = (function (){var temp__4425__auto__ = cljs.core.cst$kw$extra_DASH_validator_DASH_fn.cljs$core$IFn$_invoke$arity$1(this$__$1);
if(cljs.core.truth_(temp__4425__auto__)){
var evf = temp__4425__auto__;
return (schema.core.subschema_walker.cljs$core$IFn$_invoke$arity$1 ? schema.core.subschema_walker.cljs$core$IFn$_invoke$arity$1(schema.core.pred.cljs$core$IFn$_invoke$arity$1(evf)) : schema.core.subschema_walker.call(null,schema.core.pred.cljs$core$IFn$_invoke$arity$1(evf)));
} else {
return null;
}
})();
return ((function (map_checker,pred_checker,this$__$1){
return (function (r){
var or__4512__auto__ = (((r instanceof self__.klass))?null:schema.utils.error(schema.utils.__GT_ValidationError(this$__$1,r,(new cljs.core.Delay(((function (map_checker,pred_checker,this$__$1){
return (function (){
return cljs.core._conj(cljs.core._conj(cljs.core._conj(cljs.core.List.EMPTY,schema.utils.value_name(r)),self__.klass),cljs.core.cst$sym$instance_QMARK_);
});})(map_checker,pred_checker,this$__$1))
,null)),null)));
if(cljs.core.truth_(or__4512__auto__)){
return or__4512__auto__;
} else {
var res = (map_checker.cljs$core$IFn$_invoke$arity$1 ? map_checker.cljs$core$IFn$_invoke$arity$1(r) : map_checker.call(null,r));
if(cljs.core.truth_(schema.utils.error_QMARK_(res))){
return res;
} else {
var pred_res = (cljs.core.truth_(pred_checker)?(pred_checker.cljs$core$IFn$_invoke$arity$1 ? pred_checker.cljs$core$IFn$_invoke$arity$1(r) : pred_checker.call(null,r)):null);
if(cljs.core.truth_(schema.utils.error_QMARK_(pred_res))){
return pred_res;
} else {
return res;
}
}
}
});
;})(map_checker,pred_checker,this$__$1))
});

schema.core.Record.prototype.schema$core$Schema$explain$arity$1 = (function (this$){
var self__ = this;
var this$__$1 = this;
return cljs.core._conj(cljs.core._conj(cljs.core._conj(cljs.core.List.EMPTY,schema.core.explain(self__.schema)),cljs.core.symbol.cljs$core$IFn$_invoke$arity$1(cljs.core.pr_str.cljs$core$IFn$_invoke$arity$variadic(cljs.core.array_seq([self__.klass], 0)))),cljs.core.cst$sym$record);
});

schema.core.Record.getBasis = (function (){
return new cljs.core.PersistentVector(null, 2, 5, cljs.core.PersistentVector.EMPTY_NODE, [cljs.core.cst$sym$klass,cljs.core.cst$sym$schema], null);
});

schema.core.Record.cljs$lang$type = true;

schema.core.Record.cljs$lang$ctorPrSeq = (function (this__5160__auto__){
return cljs.core._conj(cljs.core.List.EMPTY,"schema.core/Record");
});

schema.core.Record.cljs$lang$ctorPrWriter = (function (this__5160__auto__,writer__5161__auto__){
return cljs.core._write(writer__5161__auto__,"schema.core/Record");
});

schema.core.__GT_Record = (function schema$core$__GT_Record(klass,schema__$1){
return (new schema.core.Record(klass,schema__$1,null,null,null));
});

schema.core.map__GT_Record = (function schema$core$map__GT_Record(G__27460){
return (new schema.core.Record(cljs.core.cst$kw$klass.cljs$core$IFn$_invoke$arity$1(G__27460),cljs.core.cst$kw$schema.cljs$core$IFn$_invoke$arity$1(G__27460),null,cljs.core.dissoc.cljs$core$IFn$_invoke$arity$variadic(G__27460,cljs.core.cst$kw$klass,cljs.core.array_seq([cljs.core.cst$kw$schema], 0)),null));
});

/**
 * A Record instance of type klass, whose elements match map schema 'schema'.
 */
schema.core.record = (function schema$core$record(klass,schema__$1){
if(cljs.core.map_QMARK_(schema__$1)){
} else {
throw (new Error(schema.utils.format_STAR_.cljs$core$IFn$_invoke$arity$variadic("Expected map, got %s",cljs.core.array_seq([schema.utils.type_of(schema__$1)], 0))));
}

return (new schema.core.Record(klass,schema__$1,null,null,null));
});
schema.core.explain_input_schema = (function schema$core$explain_input_schema(input_schema){
var vec__27469 = cljs.core.split_with((function (p1__27466_SHARP_){
return (p1__27466_SHARP_ instanceof schema.core.One);
}),input_schema);
var required = cljs.core.nth.cljs$core$IFn$_invoke$arity$3(vec__27469,(0),null);
var more = cljs.core.nth.cljs$core$IFn$_invoke$arity$3(vec__27469,(1),null);
return cljs.core.concat.cljs$core$IFn$_invoke$arity$2(cljs.core.map.cljs$core$IFn$_invoke$arity$2(((function (vec__27469,required,more){
return (function (p1__27467_SHARP_){
return schema.core.explain(p1__27467_SHARP_.schema);
});})(vec__27469,required,more))
,required),((cljs.core.seq(more))?new cljs.core.PersistentVector(null, 2, 5, cljs.core.PersistentVector.EMPTY_NODE, [cljs.core.cst$sym$_AMPERSAND_,cljs.core.mapv.cljs$core$IFn$_invoke$arity$2(schema.core.explain,more)], null):null));
});

/**
* @constructor
 * @implements {cljs.core.IRecord}
 * @implements {cljs.core.IEquiv}
 * @implements {cljs.core.IHash}
 * @implements {cljs.core.ICollection}
 * @implements {schema.core.Schema}
 * @implements {cljs.core.ICounted}
 * @implements {cljs.core.ISeqable}
 * @implements {cljs.core.IMeta}
 * @implements {cljs.core.ICloneable}
 * @implements {cljs.core.IPrintWithWriter}
 * @implements {cljs.core.IIterable}
 * @implements {cljs.core.IWithMeta}
 * @implements {cljs.core.IAssociative}
 * @implements {cljs.core.IMap}
 * @implements {cljs.core.ILookup}
*/
schema.core.FnSchema = (function (output_schema,input_schemas,__meta,__extmap,__hash){
this.output_schema = output_schema;
this.input_schemas = input_schemas;
this.__meta = __meta;
this.__extmap = __extmap;
this.__hash = __hash;
this.cljs$lang$protocol_mask$partition0$ = 2229667594;
this.cljs$lang$protocol_mask$partition1$ = 8192;
})
schema.core.FnSchema.prototype.cljs$core$ILookup$_lookup$arity$2 = (function (this__5126__auto__,k__5127__auto__){
var self__ = this;
var this__5126__auto____$1 = this;
return cljs.core._lookup.cljs$core$IFn$_invoke$arity$3(this__5126__auto____$1,k__5127__auto__,null);
});

schema.core.FnSchema.prototype.cljs$core$ILookup$_lookup$arity$3 = (function (this__5128__auto__,k27471,else__5129__auto__){
var self__ = this;
var this__5128__auto____$1 = this;
var G__27473 = (((k27471 instanceof cljs.core.Keyword))?k27471.fqn:null);
switch (G__27473) {
case "output-schema":
return self__.output_schema;

break;
case "input-schemas":
return self__.input_schemas;

break;
default:
return cljs.core.get.cljs$core$IFn$_invoke$arity$3(self__.__extmap,k27471,else__5129__auto__);

}
});

schema.core.FnSchema.prototype.cljs$core$IPrintWithWriter$_pr_writer$arity$3 = (function (this__5140__auto__,writer__5141__auto__,opts__5142__auto__){
var self__ = this;
var this__5140__auto____$1 = this;
var pr_pair__5143__auto__ = ((function (this__5140__auto____$1){
return (function (keyval__5144__auto__){
return cljs.core.pr_sequential_writer(writer__5141__auto__,cljs.core.pr_writer,""," ","",opts__5142__auto__,keyval__5144__auto__);
});})(this__5140__auto____$1))
;
return cljs.core.pr_sequential_writer(writer__5141__auto__,pr_pair__5143__auto__,"#schema.core.FnSchema{",", ","}",opts__5142__auto__,cljs.core.concat.cljs$core$IFn$_invoke$arity$2(new cljs.core.PersistentVector(null, 2, 5, cljs.core.PersistentVector.EMPTY_NODE, [(new cljs.core.PersistentVector(null,2,(5),cljs.core.PersistentVector.EMPTY_NODE,[cljs.core.cst$kw$output_DASH_schema,self__.output_schema],null)),(new cljs.core.PersistentVector(null,2,(5),cljs.core.PersistentVector.EMPTY_NODE,[cljs.core.cst$kw$input_DASH_schemas,self__.input_schemas],null))], null),self__.__extmap));
});

schema.core.FnSchema.prototype.cljs$core$IIterable$ = true;

schema.core.FnSchema.prototype.cljs$core$IIterable$_iterator$arity$1 = (function (G__27470){
var self__ = this;
var G__27470__$1 = this;
return (new cljs.core.RecordIter((0),G__27470__$1,2,new cljs.core.PersistentVector(null, 2, 5, cljs.core.PersistentVector.EMPTY_NODE, [cljs.core.cst$kw$output_DASH_schema,cljs.core.cst$kw$input_DASH_schemas], null),cljs.core._iterator(self__.__extmap)));
});

schema.core.FnSchema.prototype.cljs$core$IMeta$_meta$arity$1 = (function (this__5124__auto__){
var self__ = this;
var this__5124__auto____$1 = this;
return self__.__meta;
});

schema.core.FnSchema.prototype.cljs$core$ICloneable$_clone$arity$1 = (function (this__5120__auto__){
var self__ = this;
var this__5120__auto____$1 = this;
return (new schema.core.FnSchema(self__.output_schema,self__.input_schemas,self__.__meta,self__.__extmap,self__.__hash));
});

schema.core.FnSchema.prototype.cljs$core$ICounted$_count$arity$1 = (function (this__5130__auto__){
var self__ = this;
var this__5130__auto____$1 = this;
return (2 + cljs.core.count(self__.__extmap));
});

schema.core.FnSchema.prototype.cljs$core$IHash$_hash$arity$1 = (function (this__5121__auto__){
var self__ = this;
var this__5121__auto____$1 = this;
var h__4947__auto__ = self__.__hash;
if(!((h__4947__auto__ == null))){
return h__4947__auto__;
} else {
var h__4947__auto____$1 = cljs.core.hash_imap(this__5121__auto____$1);
self__.__hash = h__4947__auto____$1;

return h__4947__auto____$1;
}
});

schema.core.FnSchema.prototype.cljs$core$IEquiv$_equiv$arity$2 = (function (this__5122__auto__,other__5123__auto__){
var self__ = this;
var this__5122__auto____$1 = this;
if(cljs.core.truth_((function (){var and__4500__auto__ = other__5123__auto__;
if(cljs.core.truth_(and__4500__auto__)){
var and__4500__auto____$1 = (this__5122__auto____$1.constructor === other__5123__auto__.constructor);
if(and__4500__auto____$1){
return cljs.core.equiv_map(this__5122__auto____$1,other__5123__auto__);
} else {
return and__4500__auto____$1;
}
} else {
return and__4500__auto__;
}
})())){
return true;
} else {
return false;
}
});

schema.core.FnSchema.prototype.cljs$core$IMap$_dissoc$arity$2 = (function (this__5135__auto__,k__5136__auto__){
var self__ = this;
var this__5135__auto____$1 = this;
if(cljs.core.contains_QMARK_(new cljs.core.PersistentHashSet(null, new cljs.core.PersistentArrayMap(null, 2, [cljs.core.cst$kw$output_DASH_schema,null,cljs.core.cst$kw$input_DASH_schemas,null], null), null),k__5136__auto__)){
return cljs.core.dissoc.cljs$core$IFn$_invoke$arity$2(cljs.core.with_meta(cljs.core.into.cljs$core$IFn$_invoke$arity$2(cljs.core.PersistentArrayMap.EMPTY,this__5135__auto____$1),self__.__meta),k__5136__auto__);
} else {
return (new schema.core.FnSchema(self__.output_schema,self__.input_schemas,self__.__meta,cljs.core.not_empty(cljs.core.dissoc.cljs$core$IFn$_invoke$arity$2(self__.__extmap,k__5136__auto__)),null));
}
});

schema.core.FnSchema.prototype.cljs$core$IAssociative$_assoc$arity$3 = (function (this__5133__auto__,k__5134__auto__,G__27470){
var self__ = this;
var this__5133__auto____$1 = this;
var pred__27474 = cljs.core.keyword_identical_QMARK_;
var expr__27475 = k__5134__auto__;
if(cljs.core.truth_((pred__27474.cljs$core$IFn$_invoke$arity$2 ? pred__27474.cljs$core$IFn$_invoke$arity$2(cljs.core.cst$kw$output_DASH_schema,expr__27475) : pred__27474.call(null,cljs.core.cst$kw$output_DASH_schema,expr__27475)))){
return (new schema.core.FnSchema(G__27470,self__.input_schemas,self__.__meta,self__.__extmap,null));
} else {
if(cljs.core.truth_((pred__27474.cljs$core$IFn$_invoke$arity$2 ? pred__27474.cljs$core$IFn$_invoke$arity$2(cljs.core.cst$kw$input_DASH_schemas,expr__27475) : pred__27474.call(null,cljs.core.cst$kw$input_DASH_schemas,expr__27475)))){
return (new schema.core.FnSchema(self__.output_schema,G__27470,self__.__meta,self__.__extmap,null));
} else {
return (new schema.core.FnSchema(self__.output_schema,self__.input_schemas,self__.__meta,cljs.core.assoc.cljs$core$IFn$_invoke$arity$3(self__.__extmap,k__5134__auto__,G__27470),null));
}
}
});

schema.core.FnSchema.prototype.cljs$core$ISeqable$_seq$arity$1 = (function (this__5138__auto__){
var self__ = this;
var this__5138__auto____$1 = this;
return cljs.core.seq(cljs.core.concat.cljs$core$IFn$_invoke$arity$2(new cljs.core.PersistentVector(null, 2, 5, cljs.core.PersistentVector.EMPTY_NODE, [(new cljs.core.PersistentVector(null,2,(5),cljs.core.PersistentVector.EMPTY_NODE,[cljs.core.cst$kw$output_DASH_schema,self__.output_schema],null)),(new cljs.core.PersistentVector(null,2,(5),cljs.core.PersistentVector.EMPTY_NODE,[cljs.core.cst$kw$input_DASH_schemas,self__.input_schemas],null))], null),self__.__extmap));
});

schema.core.FnSchema.prototype.cljs$core$IWithMeta$_with_meta$arity$2 = (function (this__5125__auto__,G__27470){
var self__ = this;
var this__5125__auto____$1 = this;
return (new schema.core.FnSchema(self__.output_schema,self__.input_schemas,G__27470,self__.__extmap,self__.__hash));
});

schema.core.FnSchema.prototype.cljs$core$ICollection$_conj$arity$2 = (function (this__5131__auto__,entry__5132__auto__){
var self__ = this;
var this__5131__auto____$1 = this;
if(cljs.core.vector_QMARK_(entry__5132__auto__)){
return cljs.core._assoc(this__5131__auto____$1,cljs.core._nth.cljs$core$IFn$_invoke$arity$2(entry__5132__auto__,(0)),cljs.core._nth.cljs$core$IFn$_invoke$arity$2(entry__5132__auto__,(1)));
} else {
return cljs.core.reduce.cljs$core$IFn$_invoke$arity$3(cljs.core._conj,this__5131__auto____$1,entry__5132__auto__);
}
});

schema.core.FnSchema.prototype.schema$core$Schema$ = true;

schema.core.FnSchema.prototype.schema$core$Schema$walker$arity$1 = (function (this$){
var self__ = this;
var this$__$1 = this;
return ((function (this$__$1){
return (function (x){
if(cljs.core.ifn_QMARK_(x)){
return x;
} else {
return schema.utils.error(schema.utils.__GT_ValidationError(this$__$1,x,(new cljs.core.Delay(((function (this$__$1){
return (function (){
return cljs.core._conj(cljs.core._conj(cljs.core.List.EMPTY,schema.utils.value_name(x)),cljs.core.cst$sym$ifn_QMARK_);
});})(this$__$1))
,null)),null));
}
});
;})(this$__$1))
});

schema.core.FnSchema.prototype.schema$core$Schema$explain$arity$1 = (function (this$){
var self__ = this;
var this$__$1 = this;
if((cljs.core.count(self__.input_schemas) > (1))){
return cljs.core.list_STAR_.cljs$core$IFn$_invoke$arity$3(cljs.core.cst$sym$_EQ__GT__STAR_,schema.core.explain(self__.output_schema),cljs.core.map.cljs$core$IFn$_invoke$arity$2(schema.core.explain_input_schema,self__.input_schemas));
} else {
return cljs.core.list_STAR_.cljs$core$IFn$_invoke$arity$3(cljs.core.cst$sym$_EQ__GT_,schema.core.explain(self__.output_schema),schema.core.explain_input_schema(cljs.core.first(self__.input_schemas)));
}
});

schema.core.FnSchema.getBasis = (function (){
return new cljs.core.PersistentVector(null, 2, 5, cljs.core.PersistentVector.EMPTY_NODE, [cljs.core.cst$sym$output_DASH_schema,cljs.core.cst$sym$input_DASH_schemas], null);
});

schema.core.FnSchema.cljs$lang$type = true;

schema.core.FnSchema.cljs$lang$ctorPrSeq = (function (this__5160__auto__){
return cljs.core._conj(cljs.core.List.EMPTY,"schema.core/FnSchema");
});

schema.core.FnSchema.cljs$lang$ctorPrWriter = (function (this__5160__auto__,writer__5161__auto__){
return cljs.core._write(writer__5161__auto__,"schema.core/FnSchema");
});

schema.core.__GT_FnSchema = (function schema$core$__GT_FnSchema(output_schema,input_schemas){
return (new schema.core.FnSchema(output_schema,input_schemas,null,null,null));
});

schema.core.map__GT_FnSchema = (function schema$core$map__GT_FnSchema(G__27472){
return (new schema.core.FnSchema(cljs.core.cst$kw$output_DASH_schema.cljs$core$IFn$_invoke$arity$1(G__27472),cljs.core.cst$kw$input_DASH_schemas.cljs$core$IFn$_invoke$arity$1(G__27472),null,cljs.core.dissoc.cljs$core$IFn$_invoke$arity$variadic(G__27472,cljs.core.cst$kw$output_DASH_schema,cljs.core.array_seq([cljs.core.cst$kw$input_DASH_schemas], 0)),null));
});

schema.core.arity = (function schema$core$arity(input_schema){
if(cljs.core.seq(input_schema)){
if((cljs.core.last(input_schema) instanceof schema.core.One)){
return cljs.core.count(input_schema);
} else {
return Number.MAX_VALUE;
}
} else {
return (0);
}
});
/**
 * A function outputting a value in output schema, whose argument vector must match one of
 * input-schemas, each of which should be a sequence schema.
 * Currently function schemas are purely descriptive; they validate against any function,
 * regardless of actual input and output types.
 */
schema.core.make_fn_schema = (function schema$core$make_fn_schema(output_schema,input_schemas){
if(cljs.core.seq(input_schemas)){
} else {
throw (new Error(schema.utils.format_STAR_("Function must have at least one input schema")));
}

if(cljs.core.every_QMARK_(cljs.core.vector_QMARK_,input_schemas)){
} else {
throw (new Error(schema.utils.format_STAR_("Each arity must be a vector.")));
}

if(cljs.core.truth_(cljs.core.apply.cljs$core$IFn$_invoke$arity$2(cljs.core.distinct_QMARK_,cljs.core.map.cljs$core$IFn$_invoke$arity$2(schema.core.arity,input_schemas)))){
} else {
throw (new Error(schema.utils.format_STAR_("Arities must be distinct")));
}

return (new schema.core.FnSchema(output_schema,cljs.core.sort_by.cljs$core$IFn$_invoke$arity$2(schema.core.arity,input_schemas),null,null,null));
});
schema.core.schema_with_name = (function schema$core$schema_with_name(schema__$1,name){

return cljs.core.vary_meta.cljs$core$IFn$_invoke$arity$4(schema__$1,cljs.core.assoc,cljs.core.cst$kw$name,name);
});
schema.core.schema_name = (function schema$core$schema_name(schema__$1){

return cljs.core.cst$kw$name.cljs$core$IFn$_invoke$arity$1(cljs.core.meta(schema__$1));
});
/**
 * Get the current global schema validation setting.
 */
schema.core.fn_validation_QMARK_ = (function schema$core$fn_validation_QMARK_(){
return schema.utils.use_fn_validation.get_cell();
});
/**
 * Globally turn on schema validation for all s/fn and s/defn instances.
 */
schema.core.set_fn_validation_BANG_ = (function schema$core$set_fn_validation_BANG_(on_QMARK_){
return schema.utils.use_fn_validation.set_cell(on_QMARK_);
});
/**
 * Attach the schema to fn f at runtime, extractable by fn-schema.
 */
schema.core.schematize_fn = (function schema$core$schematize_fn(f,schema__$1){
return cljs.core.vary_meta.cljs$core$IFn$_invoke$arity$4(f,cljs.core.assoc,cljs.core.cst$kw$schema,schema__$1);
});
/**
 * Produce the schema for a function defined with s/fn or s/defn.
 */
schema.core.fn_schema = (function schema$core$fn_schema(f){
if(cljs.core.fn_QMARK_(f)){
} else {
throw (new Error(schema.utils.format_STAR_.cljs$core$IFn$_invoke$arity$variadic("Non-function %s",cljs.core.array_seq([schema.utils.type_of(f)], 0))));
}

var or__4512__auto__ = schema.utils.class_schema(schema.utils.fn_schema_bearer(f));
if(cljs.core.truth_(or__4512__auto__)){
return or__4512__auto__;
} else {
var m__8321__auto__ = cljs.core.meta(f);
var k__8322__auto__ = cljs.core.cst$kw$schema;
var temp__4423__auto__ = cljs.core.find(m__8321__auto__,k__8322__auto__);
if(cljs.core.truth_(temp__4423__auto__)){
var pair__8323__auto__ = temp__4423__auto__;
return cljs.core.val(pair__8323__auto__);
} else {
throw (new Error(schema.utils.format_STAR_.cljs$core$IFn$_invoke$arity$variadic("Key %s not found in %s",cljs.core.array_seq([k__8322__auto__,m__8321__auto__], 0))));
}
}
});
